(*

HiggsMassSquared.m by D.Ochoa on 18 Oct 2017 

3-loop order Higgs boson squared pole mass of O(alpha_t alpha_3^2) x
O((v^2/M_S^2)^0) expressed in terms of MSSM DR-bar parameters in the
MSSM.

*** List of symbols ***

k = 1/(4 Pi)^2      : Loop factor 
M                   : Renormalization scale
MZ                  : Z boson pole mass
z2                  : Zeta[2]
z3                  : Zeta[3]

MSSM DR parameters

mt = yt vu                      : Top quark mass, 

a3 = k^(1/2) g3^2               : Strong gauge coupling "fine structure constant"
at = k^(1/2) yt^2 Sin[beta]^2   : Top Yukawa coupling "fine structure constant"

beta   : ArcTan of the ratio of the vacuum expectation values of the Higgs fields

m3     : Soft-SUSY breaking gluino mass parameter
mQ3    : Soft-SUSY breaking doublet stop mass parameter
mU3    : Soft-SUSY breaking singlet stop mass parameter
msq    : Soft-SUSY breaking squark mass parameter
Xt     : Soft-SUSY breaking stop mixing

MSSM MS bar threshold corrections

dlatas2        : Threshold correction of the SM self-interaction coupling of O(at a3^2)

*)

Mh2=( 
MZ^2 Cos[2 beta]^2+at mt^2 (Sqrt[k] ((12 Xt^4)/(mQ3^2-mU3^2)^2+(6+(12 Xt^2)/(mQ3^2-mU3^2)+(-((6 mQ3^2)/(mQ3^2-mU3^2)^3)-(6 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[mQ3^2/M^2]+12 Log[M^2/mt^2]+(6-(12 Xt^2)/(mQ3^2-mU3^2)+((6 mQ3^2)/(mQ3^2-mU3^2)^3+(6 mU3^2)/(mQ3^2-mU3^2)^3) Xt^4) Log[mU3^2/M^2])+a3 k (-64+(32 m3^2)/mQ3^2+(32 m3^2)/mU3^2-(64 m3^2 Xt^2)/(mQ3^2 mU3^2)-(512 m3 Xt^3)/(mQ3^2-mU3^2)^2+(192/(mQ3^2-mU3^2)^2+(32 m3^2)/(mQ3^2 (mQ3^2-mU3^2)^2)+(32 m3^2)/(mU3^2 (mQ3^2-mU3^2)^2)) Xt^4+(-((16 (3 m3^4-2 m3^2 mQ3^2+mQ3^4))/(m3^2-mQ3^2)^2)+(128 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))-(32 (3 mQ3^2-mU3^2) Xt^2)/(mQ3^2-mU3^2)^2+(64 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(32 (m3^2 mQ3^2+2 mQ3^4-m3^2 mU3^2+2 mQ3^2 mU3^2) Xt^4)/(mQ3^2-mU3^2)^4) Log[mQ3^2/M^2]^2+96 Log[M^2/mt^2]^2+((32 (3 m3^4-2 m3^2 mU3^2+mU3^4))/(m3^2-mU3^2)^2-(64 (2 m3^3-m3 mU3^2) Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(64 (3 m3^2-2 mU3^2) Xt^2)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(128 (m3 mQ3^2+3 m3 mU3^2) Xt^3)/(-mQ3^2+mU3^2)^3-(32 (2 m3^2+3 mQ3^2+7 mU3^2) Xt^4)/(-mQ3^2+mU3^2)^3) Log[mU3^2/M^2]+(-((16 (3 m3^4-2 m3^2 mU3^2+mU3^4))/(m3^2-mU3^2)^2)+(128 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))-(32 (-mQ3^2+3 mU3^2) Xt^2)/(-mQ3^2+mU3^2)^2+(64 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(-mQ3^2+mU3^2)^3+(32 (-m3^2 mQ3^2+m3^2 mU3^2+2 mQ3^2 mU3^2+2 mU3^4) Xt^4)/(-mQ3^2+mU3^2)^4) Log[mU3^2/M^2]^2+Log[M^2/mt^2] (-((32 (2 m3^4-m3^2 mQ3^2-m3^2 mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2)))+((32 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(128 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[m3^2/M^2]+(-((64 m3^2 mQ3^2)/(m3^2-mQ3^2)^2)+(32 mQ3^4)/(m3^2-mQ3^2)^2+(128 m3 mQ3^2 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))) Log[mQ3^2/M^2]+(-((64 m3^2 mU3^2)/(m3^2-mU3^2)^2)+(32 mU3^4)/(m3^2-mU3^2)^2+(128 m3 mU3^2 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))) Log[mU3^2/M^2])+Log[m3^2/M^2] (-((32 (m3^10 mQ3^2-2 m3^8 mQ3^4+m3^6 mQ3^6+m3^10 mU3^2+2 m3^8 mQ3^2 mU3^2-2 m3^6 mQ3^4 mU3^2+m3^4 mQ3^6 mU3^2-2 m3^8 mU3^4-2 m3^6 mQ3^2 mU3^4+m3^6 mU3^6+m3^4 mQ3^2 mU3^6))/(mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2))+(64 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))+(64 m3^4 (-m3^2+mQ3^2+mU3^2) Xt^2)/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))-(32 (m3^2 mQ3^2+m3^2 mU3^2) Xt^4)/(mQ3^2 mU3^2 (mQ3^2-mU3^2)^2)+((32 m3^4)/(m3^2-mQ3^2)^2-(128 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))-(256 m3^3 Xt^3)/(mQ3^2-mU3^2)^3) Log[mQ3^2/M^2]+((32 m3^4)/(m3^2-mU3^2)^2-(128 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))-(256 m3^3 Xt^3)/(-mQ3^2+mU3^2)^3) Log[mU3^2/M^2])+Log[mQ3^2/M^2] ((32 (3 m3^4-2 m3^2 mQ3^2+mQ3^4))/(m3^2-mQ3^2)^2-(64 (2 m3^3-m3 mQ3^2) Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(64 (3 m3^2-2 mQ3^2) Xt^2)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(128 (3 m3 mQ3^2+m3 mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(32 (2 m3^2+7 mQ3^2+3 mU3^2) Xt^4)/(mQ3^2-mU3^2)^3+((64 (mQ3^2+mU3^2) Xt^2)/(mQ3^2-mU3^2)^2-(64 (mQ3^4+2 mQ3^2 mU3^2+mU3^4) Xt^4)/(mQ3^2-mU3^2)^4) Log[mU3^2/M^2])+(-((64 m3^4)/(-m3^2+mQ3^2)^2)+(256 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(128 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(32 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mQ3^2]+(-((64 m3^4)/(m3^2-mU3^2)^2)+(256 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(128 m3 (-2 m3^2+mQ3^2+mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(32 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mU3^2])+a3^2 k^(3/2) (dlatas2-(128 m3^4)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))+(128 mQ3^2 mU3^2)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(16 (m3^4-mQ3^2 mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2))-(48 (m3^4-mQ3^2 mU3^2)^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(16 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) (-2 m3^6 mQ3^2+2 m3^4 mQ3^4-2 m3^6 mU3^2+9 m3^4 mQ3^2 mU3^2-6 m3^2 mQ3^4 mU3^2+2 m3^4 mU3^4-6 m3^2 mQ3^2 mU3^4+3 mQ3^4 mU3^4))/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+(4 (291 m3^8-282 m3^6 mQ3^2+93 m3^4 mQ3^4-18 m3^2 mQ3^6-360 m3^6 msq^2+300 m3^4 mQ3^2 msq^2-180 m3^2 mQ3^4 msq^2-282 m3^6 mU3^2-112 m3^4 mQ3^2 mU3^2+208 m3^2 mQ3^4 mU3^2-6 mQ3^6 mU3^2+300 m3^4 msq^2 mU3^2+240 m3^2 mQ3^2 msq^2 mU3^2-60 mQ3^4 msq^2 mU3^2+93 m3^4 mU3^4+208 m3^2 mQ3^2 mU3^4-169 mQ3^4 mU3^4-180 m3^2 msq^2 mU3^4-60 mQ3^2 msq^2 mU3^4-18 m3^2 mU3^6-6 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(3776/9)+(2816 z2^2)/15+256/9 Log[2]^2 (6 z2-Log[2]^2)+320 z3-2048/3 PolyLog[4,1/2]-(64 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(64 m3^2 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2) mU3^2 (m3^2-mU3^2))-(512 m3 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(8 (-44 m3^10 mQ3^2+56 m3^8 mQ3^4-12 m3^6 mQ3^6-44 m3^10 mU3^2-455 m3^8 mQ3^2 mU3^2+558 m3^6 mQ3^4 mU3^2-173 m3^4 mQ3^6 mU3^2+18 m3^2 mQ3^8 mU3^2+360 m3^6 mQ3^2 msq^2 mU3^2-300 m3^4 mQ3^4 msq^2 mU3^2+180 m3^2 mQ3^6 msq^2 mU3^2+56 m3^8 mU3^4+558 m3^6 mQ3^2 mU3^4-168 m3^4 mQ3^4 mU3^4-236 m3^2 mQ3^6 mU3^4+6 mQ3^8 mU3^4-300 m3^4 mQ3^2 msq^2 mU3^4-240 m3^2 mQ3^4 msq^2 mU3^4+60 mQ3^6 msq^2 mU3^4-12 m3^6 mU3^6-173 m3^4 mQ3^2 mU3^6-236 m3^2 mQ3^4 mU3^6+277 mQ3^6 mU3^6+180 m3^2 mQ3^2 msq^2 mU3^6+60 mQ3^4 msq^2 mU3^6+18 m3^2 mQ3^2 mU3^8+6 mQ3^4 mU3^8) Xt^4)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(128 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(-((4 (128 m3^16-544 m3^14 mQ3^2+490 m3^12 mQ3^4-69 m3^10 mQ3^6-96 m3^8 mQ3^8+6 m3^6 mQ3^10+38 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-352 m3^14 mU3^2+1900 m3^12 mQ3^2 mU3^2-1841 m3^10 mQ3^4 mU3^2+213 m3^8 mQ3^6 mU3^2+634 m3^6 mQ3^8 mU3^2-374 m3^4 mQ3^10 mU3^2+47 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+298 m3^12 mU3^4-2539 m3^10 mQ3^2 mU3^4+3011 m3^8 mQ3^4 mU3^4-1170 m3^6 mQ3^6 mU3^4-116 m3^4 mQ3^8 mU3^4+181 m3^2 mQ3^10 mU3^4-mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-31 m3^10 mU3^6+1383 m3^8 mQ3^2 mU3^6-1874 m3^6 mQ3^4 mU3^6+882 m3^4 mQ3^6 mU3^6-71 m3^2 mQ3^8 mU3^6-65 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-31 m3^8 mU3^8-284 m3^6 mQ3^2 mU3^8+466 m3^4 mQ3^4 mU3^8-276 m3^2 mQ3^6 mU3^8+69 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(32 (22 m3^11 mQ3^2-206 m3^9 mQ3^4+196 m3^7 mQ3^6-65 m3^5 mQ3^8+8 m3^3 mQ3^10-3 m3 mQ3^12+60 m3^7 mQ3^4 msq^2-60 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-22 m3^11 mU3^2+145 m3^9 mQ3^2 mU3^2+246 m3^7 mQ3^4 mU3^2-325 m3^5 mQ3^6 mU3^2+90 m3^3 mQ3^8 mU3^2+10 m3 mQ3^10 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+45 m3^9 mU3^4-352 m3^7 mQ3^2 mU3^4+91 m3^5 mQ3^4 mU3^4+142 m3^3 mQ3^6 mU3^4-70 m3 mQ3^8 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4-60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-26 m3^7 mU3^6+203 m3^5 mQ3^2 mU3^6-176 m3^3 mQ3^4 mU3^6+47 m3 mQ3^6 mU3^6-60 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(8 (128 m3^16-1056 m3^14 mQ3^2+1658 m3^12 mQ3^4-605 m3^10 mQ3^6-224 m3^8 mQ3^8+78 m3^6 mQ3^10+38 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-344 m3^14 mU3^2+3292 m3^12 mQ3^2 mU3^2-5273 m3^10 mQ3^4 mU3^2+1821 m3^8 mQ3^6 mU3^2+1018 m3^6 mQ3^8 mU3^2-590 m3^4 mQ3^10 mU3^2+47 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+274 m3^12 mU3^4-3643 m3^10 mQ3^2 mU3^4+6299 m3^8 mQ3^4 mU3^4-2778 m3^6 mQ3^6 mU3^4-500 m3^4 mQ3^8 mU3^4+397 m3^2 mQ3^10 mU3^4-mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-7 m3^10 mU3^6+1463 m3^8 mQ3^2 mU3^6-2826 m3^6 mQ3^4 mU3^6+1418 m3^4 mQ3^6 mU3^6+57 m3^2 mQ3^8 mU3^6-137 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-39 m3^8 mU3^8-140 m3^6 mQ3^2 mU3^8+394 m3^4 mQ3^4 mU3^8-276 m3^2 mQ3^6 mU3^8+69 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(64 (2 m3^13-21 m3^11 mQ3^2-61 m3^9 mQ3^4+179 m3^7 mQ3^6-88 m3^5 mQ3^8+8 m3^3 mQ3^10-3 m3 mQ3^12+60 m3^7 mQ3^4 msq^2-60 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-23 m3^11 mU3^2+146 m3^9 mQ3^2 mU3^2-35 m3^7 mQ3^4 mU3^2-282 m3^5 mQ3^6 mU3^2+136 m3^3 mQ3^8 mU3^2+10 m3 mQ3^10 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+41 m3^9 mU3^4-225 m3^7 mQ3^2 mU3^4+218 m3^5 mQ3^4 mU3^4+107 m3^3 mQ3^6 mU3^4-93 m3 mQ3^8 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4-60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-23 m3^7 mU3^6+118 m3^5 mQ3^2 mU3^6-167 m3^3 mQ3^4 mU3^6+56 m3 mQ3^6 mU3^6-60 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(4 (1160 m3^16 mQ3^2-4284 m3^14 mQ3^4+4730 m3^12 mQ3^6-1157 m3^10 mQ3^8-584 m3^8 mQ3^10+114 m3^6 mQ3^12+38 m3^4 mQ3^14-9 m3^2 mQ3^16+60 m3^12 mQ3^4 msq^2+120 m3^10 mQ3^6 msq^2-180 m3^8 mQ3^8 msq^2-90 m3^10 mQ3^4 msq^4+60 m3^8 mQ3^6 msq^4+30 m3^6 mQ3^8 msq^4+120 m3^16 mU3^2-4328 m3^14 mQ3^2 mU3^2+15098 m3^12 mQ3^4 mU3^2-16422 m3^10 mQ3^6 mU3^2+3885 m3^8 mQ3^8 mU3^2+2248 m3^6 mQ3^10 mU3^2-660 m3^4 mQ3^12 mU3^2+38 m3^2 mQ3^14 mU3^2-3 mQ3^16 mU3^2-180 m3^10 mQ3^4 msq^2 mU3^2-360 m3^8 mQ3^6 msq^2 mU3^2+540 m3^6 mQ3^8 msq^2 mU3^2+270 m3^8 mQ3^4 msq^4 mU3^2-180 m3^6 mQ3^6 msq^4 mU3^2-90 m3^4 mQ3^8 msq^4 mU3^2-316 m3^14 mU3^4+6078 m3^12 mQ3^2 mU3^4-19808 m3^10 mQ3^4 mU3^4+21176 m3^8 mQ3^6 mU3^4-5276 m3^6 mQ3^8 mU3^4-2386 m3^4 mQ3^10 mU3^4+552 m3^2 mQ3^12 mU3^4-4 mQ3^14 mU3^4-60 m3^12 msq^2 mU3^4-120 m3^10 mQ3^2 msq^2 mU3^4+360 m3^8 mQ3^4 msq^2 mU3^4+360 m3^6 mQ3^6 msq^2 mU3^4-540 m3^4 mQ3^8 msq^2 mU3^4+90 m3^10 msq^4 mU3^4-60 m3^8 mQ3^2 msq^4 mU3^4-300 m3^6 mQ3^4 msq^4 mU3^4+180 m3^4 mQ3^6 msq^4 mU3^4+90 m3^2 mQ3^8 msq^4 mU3^4+238 m3^12 mU3^6-3938 m3^10 mQ3^2 mU3^6+11830 m3^8 mQ3^4 mU3^6-12228 m3^6 mQ3^6 mU3^6+3258 m3^4 mQ3^8 mU3^6+1030 m3^2 mQ3^10 mU3^6-174 mQ3^12 mU3^6+180 m3^10 msq^2 mU3^6+360 m3^8 mQ3^2 msq^2 mU3^6-600 m3^6 mQ3^4 msq^2 mU3^6-120 m3^4 mQ3^6 msq^2 mU3^6+180 m3^2 mQ3^8 msq^2 mU3^6-270 m3^8 msq^4 mU3^6+180 m3^6 mQ3^2 msq^4 mU3^6+180 m3^4 mQ3^4 msq^4 mU3^6-60 m3^2 mQ3^6 msq^4 mU3^6-30 mQ3^8 msq^4 mU3^6+13 m3^10 mU3^8+1016 m3^8 mQ3^2 mU3^8-2878 m3^6 mQ3^4 mU3^8+2708 m3^4 mQ3^6 mU3^8-743 m3^2 mQ3^8 mU3^8-140 mQ3^10 mU3^8-180 m3^8 msq^2 mU3^8-360 m3^6 mQ3^2 msq^2 mU3^8+540 m3^4 mQ3^4 msq^2 mU3^8+270 m3^6 msq^4 mU3^8-180 m3^4 mQ3^2 msq^4 mU3^8-90 m3^2 mQ3^4 msq^4 mU3^8-43 m3^8 mU3^10-60 m3^6 mQ3^2 mU3^10+210 m3^4 mQ3^4 mU3^10-132 m3^2 mQ3^6 mU3^10+33 mQ3^8 mU3^10+60 m3^6 msq^2 mU3^10+120 m3^4 mQ3^2 msq^2 mU3^10-180 m3^2 mQ3^4 msq^2 mU3^10-90 m3^4 msq^4 mU3^10+60 m3^2 mQ3^2 msq^4 mU3^10+30 mQ3^4 msq^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(32 (14 m3^11 mQ3^4+86 m3^9 mQ3^6-224 m3^7 mQ3^8+113 m3^5 mQ3^10-8 m3^3 mQ3^12+3 m3 mQ3^14-60 m3^7 mQ3^6 msq^2+60 m3^5 mQ3^8 msq^2+30 m3^7 mQ3^4 msq^4-30 m3^5 mQ3^6 msq^4-32 m3^11 mQ3^2 mU3^2+117 m3^9 mQ3^4 mU3^2-362 m3^7 mQ3^6 mU3^2+510 m3^5 mQ3^8 mU3^2-194 m3^3 mQ3^10 mU3^2-7 m3 mQ3^12 mU3^2+120 m3^5 mQ3^6 msq^2 mU3^2-120 m3^3 mQ3^8 msq^2 mU3^2-60 m3^5 mQ3^4 msq^4 mU3^2+60 m3^3 mQ3^6 msq^4 mU3^2+18 m3^11 mU3^4-6 m3^9 mQ3^2 mU3^4-182 m3^7 mQ3^4 mU3^4+450 m3^5 mQ3^6 mU3^4-388 m3^3 mQ3^8 mU3^4+108 m3 mQ3^10 mU3^4+60 m3^7 mQ3^2 msq^2 mU3^4-60 m3^5 mQ3^4 msq^2 mU3^4-60 m3^3 mQ3^6 msq^2 mU3^4+60 m3 mQ3^8 msq^2 mU3^4-30 m3^7 msq^4 mU3^4+30 m3^5 mQ3^2 msq^4 mU3^4+30 m3^3 mQ3^4 msq^4 mU3^4-30 m3 mQ3^6 msq^4 mU3^4-37 m3^9 mU3^6+106 m3^7 mQ3^2 mU3^6-30 m3^5 mQ3^4 mU3^6-158 m3^3 mQ3^6 mU3^6+87 m3 mQ3^8 mU3^6-120 m3^5 mQ3^2 msq^2 mU3^6+120 m3^3 mQ3^4 msq^2 mU3^6+60 m3^5 msq^4 mU3^6-60 m3^3 mQ3^2 msq^4 mU3^6+22 m3^7 mU3^8-83 m3^5 mQ3^2 mU3^8+108 m3^3 mQ3^4 mU3^8-31 m3 mQ3^6 mU3^8+60 m3^3 mQ3^2 msq^2 mU3^8-60 m3 mQ3^4 msq^2 mU3^8-30 m3^3 msq^4 mU3^8+30 m3 mQ3^2 msq^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^5)+(1280 m3^2 mQ3^4 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^5)) Log[mQ3^2/M^2]^3+(24 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))+24 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-1/(mQ3^2-mU3^2)^2 48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^4-(48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^5)/(mQ3^2-mU3^2)^2) Log[msq^2/M^2]^2+736 Log[M^2/mt^2]^3+((4 (128 m3^14-252 m3^12 mQ3^2+20 m3^10 mQ3^4+115 m3^8 mQ3^6-20 m3^6 mQ3^8+9 m3^4 mQ3^10-60 m3^10 mQ3^2 msq^2+120 m3^8 mQ3^4 msq^2-60 m3^6 mQ3^6 msq^2+90 m3^8 mQ3^2 msq^4-180 m3^6 mQ3^4 msq^4+90 m3^4 mQ3^6 msq^4-1028 m3^12 mU3^2+3132 m3^10 mQ3^2 mU3^2-2847 m3^8 mQ3^4 mU3^2+764 m3^6 mQ3^6 mU3^2-39 m3^4 mQ3^8 mU3^2-6 m3^2 mQ3^10 mU3^2+60 m3^10 msq^2 mU3^2-420 m3^8 mQ3^2 msq^2 mU3^2+660 m3^6 mQ3^4 msq^2 mU3^2-300 m3^4 mQ3^6 msq^2 mU3^2-90 m3^8 msq^4 mU3^2+120 m3^6 mQ3^2 msq^4 mU3^2+30 m3^4 mQ3^4 msq^4 mU3^2-60 m3^2 mQ3^6 msq^4 mU3^2+1328 m3^10 mU3^4-4635 m3^8 mQ3^2 mU3^4+4812 m3^6 mQ3^4 mU3^4-1436 m3^4 mQ3^6 mU3^4+54 m3^2 mQ3^8 mU3^4-3 mQ3^10 mU3^4+300 m3^8 msq^2 mU3^4-300 m3^6 mQ3^2 msq^2 mU3^4-300 m3^4 mQ3^4 msq^2 mU3^4+300 m3^2 mQ3^6 msq^2 mU3^4+60 m3^6 msq^4 mU3^4-150 m3^4 mQ3^2 msq^4 mU3^4+120 m3^2 mQ3^4 msq^4 mU3^4-30 mQ3^6 msq^4 mU3^4-313 m3^8 mU3^6+2044 m3^6 mQ3^2 mU3^6-2968 m3^4 mQ3^4 mU3^6+992 m3^2 mQ3^6 mU3^6+5 mQ3^8 mU3^6-300 m3^6 msq^2 mU3^6+660 m3^4 mQ3^2 msq^2 mU3^6-420 m3^2 mQ3^4 msq^2 mU3^6+60 mQ3^6 msq^2 mU3^6+30 m3^4 msq^4 mU3^6-60 m3^2 mQ3^2 msq^4 mU3^6+30 mQ3^4 msq^4 mU3^6-560 m3^6 mU3^8+815 m3^4 mQ3^2 mU3^8+284 m3^2 mQ3^4 mU3^8-299 mQ3^6 mU3^8-60 m3^4 msq^2 mU3^8+120 m3^2 mQ3^2 msq^2 mU3^8-60 mQ3^4 msq^2 mU3^8+291 m3^4 mU3^10-702 m3^2 mQ3^2 mU3^10+291 mQ3^4 mU3^10+18 m3^2 mU3^12+6 mQ3^2 mU3^12))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2))-(32 (-32 m3^9 mQ3^2+m3^7 mQ3^4+2 m3^5 mQ3^6-3 m3^3 mQ3^8+30 m3^5 mQ3^2 msq^4-30 m3^3 mQ3^4 msq^4+32 m3^9 mU3^2+380 m3^7 mQ3^2 mU3^2-311 m3^5 mQ3^4 mU3^2+14 m3^3 mQ3^6 mU3^2+3 m3 mQ3^8 mU3^2-120 m3^5 mQ3^2 msq^2 mU3^2+120 m3^3 mQ3^4 msq^2 mU3^2-30 m3^5 msq^4 mU3^2+30 m3 mQ3^4 msq^4 mU3^2-397 m3^7 mU3^4-122 m3^5 mQ3^2 mU3^4+367 m3^3 mQ3^4 mU3^4-16 m3 mQ3^6 mU3^4+120 m3^5 msq^2 mU3^4-120 m3 mQ3^4 msq^2 mU3^4+30 m3^3 msq^4 mU3^4-30 m3 mQ3^2 msq^4 mU3^4+479 m3^5 mU3^6-226 m3^3 mQ3^2 mU3^6-137 m3 mQ3^4 mU3^6-120 m3^3 msq^2 mU3^6+120 m3 mQ3^2 msq^2 mU3^6-200 m3^3 mU3^8+160 m3 mQ3^2 mU3^8+6 m3 mU3^10) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)-(16 (-24 m3^10 mQ3^2+65 m3^8 mQ3^4-41 m3^6 mQ3^6+492 m3^10 mU3^2-1462 m3^8 mQ3^2 mU3^2+1346 m3^6 mQ3^4 mU3^2-373 m3^4 mQ3^6 mU3^2+9 m3^2 mQ3^8 mU3^2+90 m3^6 mQ3^2 msq^2 mU3^2-180 m3^4 mQ3^4 msq^2 mU3^2+90 m3^2 mQ3^6 msq^2 mU3^2-383 m3^8 mU3^4+1473 m3^6 mQ3^2 mU3^4-1612 m3^4 mQ3^4 mU3^4+471 m3^2 mQ3^6 mU3^4+3 mQ3^8 mU3^4-90 m3^6 msq^2 mU3^4+210 m3^4 mQ3^2 msq^2 mU3^4-150 m3^2 mQ3^4 msq^2 mU3^4+30 mQ3^6 msq^2 mU3^4-226 m3^6 mU3^6+137 m3^4 mQ3^2 mU3^6+338 m3^2 mQ3^4 mU3^6-177 mQ3^6 mU3^6-30 m3^4 msq^2 mU3^6+60 m3^2 mQ3^2 msq^2 mU3^6-30 mQ3^4 msq^2 mU3^6+192 m3^4 mU3^8-391 m3^2 mQ3^2 mU3^8+151 mQ3^4 mU3^8+9 m3^2 mU3^10+3 mQ3^2 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)+(128 (11 m3^9+7 m3^7 mQ3^2-14 m3^5 mQ3^4-37 m3^7 mU3^2+150 m3^5 mQ3^2 mU3^2-117 m3^3 mQ3^4 mU3^2+3 m3 mQ3^6 mU3^2-30 m3^3 mQ3^2 msq^2 mU3^2+30 m3 mQ3^4 msq^2 mU3^2-48 m3^5 mU3^4-47 m3^3 mQ3^2 mU3^4+82 m3 mQ3^4 mU3^4+30 m3^3 msq^2 mU3^4-30 m3 mQ3^2 msq^2 mU3^4+66 m3^3 mU3^6-53 m3 mQ3^2 mU3^6-3 m3 mU3^8) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(8 (-176 m3^14 mQ3^2+322 m3^12 mQ3^4-147 m3^10 mQ3^6-10 m3^8 mQ3^8+20 m3^6 mQ3^10-9 m3^4 mQ3^12+60 m3^10 mQ3^4 msq^2-120 m3^8 mQ3^6 msq^2+60 m3^6 mQ3^8 msq^2-90 m3^8 mQ3^4 msq^4+180 m3^6 mQ3^6 msq^4-90 m3^4 mQ3^8 msq^4+176 m3^14 mU3^2-120 m3^12 mQ3^2 mU3^2+201 m3^10 mQ3^4 mU3^2-474 m3^8 mQ3^6 mU3^2+178 m3^6 mQ3^8 mU3^2+21 m3^4 mQ3^10 mU3^2+6 m3^2 mQ3^12 mU3^2-120 m3^10 mQ3^2 msq^2 mU3^2+270 m3^8 mQ3^4 msq^2 mU3^2-180 m3^6 mQ3^6 msq^2 mU3^2+30 m3^4 mQ3^8 msq^2 mU3^2+180 m3^8 mQ3^2 msq^4 mU3^2-300 m3^6 mQ3^4 msq^4 mU3^2+60 m3^4 mQ3^6 msq^4 mU3^2+60 m3^2 mQ3^8 msq^4 mU3^2-2586 m3^12 mU3^4+6267 m3^10 mQ3^2 mU3^4-6104 m3^8 mQ3^4 mU3^4+3202 m3^6 mQ3^6 mU3^4-692 m3^4 mQ3^8 mU3^4-42 m3^2 mQ3^10 mU3^4+3 mQ3^12 mU3^4+60 m3^10 msq^2 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+420 m3^6 mQ3^4 msq^2 mU3^4-120 m3^2 mQ3^8 msq^2 mU3^4-90 m3^8 msq^4 mU3^4+60 m3^6 mQ3^2 msq^4 mU3^4+180 m3^4 mQ3^4 msq^4 mU3^4-180 m3^2 mQ3^6 msq^4 mU3^4+30 mQ3^8 msq^4 mU3^4+5679 m3^10 mU3^6-13622 m3^8 mQ3^2 mU3^6+11390 m3^6 mQ3^4 mU3^6-4230 m3^4 mQ3^6 mU3^6+722 m3^2 mQ3^8 mU3^6+mQ3^10 mU3^6+210 m3^8 msq^2 mU3^6-60 m3^6 mQ3^2 msq^2 mU3^6-480 m3^4 mQ3^4 msq^2 mU3^6+300 m3^2 mQ3^6 msq^2 mU3^6+30 mQ3^8 msq^2 mU3^6+60 m3^6 msq^4 mU3^6-180 m3^4 mQ3^2 msq^4 mU3^6+180 m3^2 mQ3^4 msq^4 mU3^6-60 mQ3^6 msq^4 mU3^6-4030 m3^8 mU3^8+9482 m3^6 mQ3^2 mU3^8-7123 m3^4 mQ3^4 mU3^8+1861 m3^2 mQ3^6 mU3^8-190 mQ3^8 mU3^8-240 m3^6 msq^2 mU3^8+480 m3^4 mQ3^2 msq^2 mU3^8-240 m3^2 mQ3^4 msq^2 mU3^8+30 m3^4 msq^4 mU3^8-60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8+368 m3^6 mU3^10-1079 m3^4 mQ3^2 mU3^10+1015 m3^2 mQ3^4 mU3^10-244 mQ3^6 mU3^10-30 m3^4 msq^2 mU3^10+60 m3^2 mQ3^2 msq^2 mU3^10-30 mQ3^4 msq^2 mU3^10+392 m3^4 mU3^12-787 m3^2 mQ3^2 mU3^12+347 mQ3^4 mU3^12+9 m3^2 mU3^14+3 mQ3^2 mU3^14) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)^4)-(64 (34 m3^9 mQ3^2-35 m3^7 mQ3^4-2 m3^5 mQ3^6+3 m3^3 mQ3^8-30 m3^5 mQ3^2 msq^4+30 m3^3 mQ3^4 msq^4+30 m3^9 mU3^2-6 m3^7 mQ3^2 mU3^2-5 m3^5 mQ3^4 mU3^2-5 m3^3 mQ3^6 mU3^2-3 m3 mQ3^8 mU3^2+30 m3^5 mQ3^2 msq^2 mU3^2-30 m3^3 mQ3^4 msq^2 mU3^2+30 m3^5 msq^4 mU3^2-30 m3 mQ3^4 msq^4 mU3^2+179 m3^7 mU3^4-306 m3^5 mQ3^2 mU3^4+106 m3^3 mQ3^4 mU3^4+7 m3 mQ3^6 mU3^4-90 m3^5 msq^2 mU3^4+60 m3^3 mQ3^2 msq^2 mU3^4+30 m3 mQ3^4 msq^2 mU3^4-30 m3^3 msq^4 mU3^4+30 m3 mQ3^2 msq^4 mU3^4-437 m3^5 mU3^6+511 m3^3 mQ3^2 mU3^6-82 m3 mQ3^4 mU3^6+90 m3^3 msq^2 mU3^6-90 m3 mQ3^2 msq^2 mU3^6+215 m3^3 mU3^8-201 m3 mQ3^2 mU3^8-3 m3 mU3^10) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(2560 m3^2 mU3^4 Xt^6)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)+(-(1/(3 (m3^2-mU3^2)^3))8 (27 m3^6-60 m3^4 msq^2+90 m3^2 msq^4-65 m3^4 mU3^2-180 m3^2 msq^2 mU3^2+30 msq^4 mU3^2+57 m3^2 mU3^4-19 mU3^6)+(64 (9 m3^5+30 m3 msq^4-14 m3^3 mU3^2-60 m3 msq^2 mU3^2+5 m3 mU3^4) Xt)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(-((16 (60 m3^4 msq^2-90 m3^2 msq^4+2 m3^4 mU3^2+180 m3^2 msq^2 mU3^2-30 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6))/(3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(48 (m3^4 (mQ3^2-3 mU3^2)+2 mU3^4 (mQ3^2-2 mU3^2)+m3^2 (-4 mQ3^2 mU3^2+8 mU3^4)))/((m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Xt^2+(-((64 (-60 m3 msq^4+m3^3 mU3^2+120 m3 msq^2 mU3^2-m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2))+(96 m3 (2 m3^4+3 mQ3^2 mU3^2-mU3^4-m3^2 (mQ3^2+3 mU3^2)))/((m3^2-mU3^2) (-mQ3^2+mU3^2)^3)) Xt^3+((8 (mQ3^2+mU3^2) (60 m3^4 msq^2-90 m3^2 msq^4+2 m3^4 mU3^2+180 m3^2 msq^2 mU3^2-30 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6))/(3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(24 (8 m3^4 mQ3^2 mU3^2-mQ3^4 mU3^4+4 mQ3^2 mU3^6+5 mU3^8-2 m3^6 (mQ3^2-mU3^2)+2 m3^2 mU3^2 (mQ3^4-5 mQ3^2 mU3^2-4 mU3^4)))/((m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Xt^4+(320 (mQ3^2+mU3^2) (-6 m3 msq^4+m3^3 mU3^2+12 m3 msq^2 mU3^2-m3 mU3^4) Xt^5)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) Log[msq^2/M^2]) Log[mU3^2/M^2]^2+(-((4 (128 m3^10+38 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-550 m3^8 mU3^2+276 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+494 m3^6 mU3^4-429 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-59 m3^4 mU3^6+234 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6-136 m3^2 mU3^8-63 mQ3^2 mU3^8+67 mU3^10))/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)))+(32 (-20 m3^7 mQ3^2+m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^2 msq^4+20 m3^7 mU3^2+185 m3^5 mQ3^2 mU3^2-11 m3^3 mQ3^4 mU3^2-3 m3 mQ3^6 mU3^2-60 m3^3 mQ3^2 msq^2 mU3^2-30 m3^3 msq^4 mU3^2-30 m3 mQ3^2 msq^4 mU3^2-202 m3^5 mU3^4-149 m3^3 mQ3^2 mU3^4+10 m3 mQ3^4 mU3^4+60 m3^3 msq^2 mU3^4+60 m3 mQ3^2 msq^2 mU3^4+30 m3 msq^4 mU3^4+189 m3^3 mU3^6+32 m3 mQ3^2 mU3^6-60 m3 msq^2 mU3^6-55 m3 mU3^8) Xt)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)-(8 (128 m3^10+42 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-1062 m3^8 mU3^2+140 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+1670 m3^6 mU3^4-361 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-615 m3^4 mU3^6+234 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6-248 m3^2 mU3^8-63 mQ3^2 mU3^8+135 mU3^10) Xt^2)/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)^2)+(32 (2 m3^9-37 m3^7 mQ3^2+2 m3^5 mQ3^4+6 m3^3 mQ3^6+60 m3^3 mQ3^2 msq^4-35 m3^7 mU3^2+205 m3^5 mQ3^2 mU3^2-22 m3^3 mQ3^4 mU3^2-6 m3 mQ3^6 mU3^2-120 m3^3 mQ3^2 msq^2 mU3^2-60 m3^3 msq^4 mU3^2-60 m3 mQ3^2 msq^4 mU3^2-131 m3^5 mU3^4-281 m3^3 mQ3^2 mU3^4+20 m3 mQ3^4 mU3^4+120 m3^3 msq^2 mU3^4+120 m3 mQ3^2 msq^2 mU3^4+60 m3 msq^4 mU3^4+353 m3^3 mU3^6+81 m3 mQ3^2 mU3^6-120 m3 msq^2 mU3^6-157 m3 mU3^8) Xt^3)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^3)+(4 (124 m3^10 mQ3^2+44 m3^8 mQ3^4-2 m3^6 mQ3^6+9 m3^4 mQ3^8-60 m3^6 mQ3^4 msq^2+90 m3^4 mQ3^4 msq^4+1156 m3^10 mU3^2-872 m3^8 mQ3^2 mU3^2+66 m3^6 mQ3^4 mU3^2-24 m3^4 mQ3^6 mU3^2-6 m3^2 mQ3^8 mU3^2-120 m3^4 mQ3^4 msq^2 mU3^2-60 m3^2 mQ3^4 msq^4 mU3^2-4276 m3^8 mU3^4+2314 m3^6 mQ3^2 mU3^4-222 m3^4 mQ3^4 mU3^4+30 m3^2 mQ3^6 mU3^4-3 mQ3^8 mU3^4+60 m3^6 msq^2 mU3^4+180 m3^2 mQ3^4 msq^2 mU3^4-90 m3^4 msq^4 mU3^4-30 mQ3^4 msq^4 mU3^4+4726 m3^6 mU3^6-2344 m3^4 mQ3^2 mU3^6+134 m3^2 mQ3^4 mU3^6-4 mQ3^6 mU3^6+120 m3^4 msq^2 mU3^6+60 m3^2 msq^4 mU3^6-1163 m3^4 mU3^8+638 m3^2 mQ3^2 mU3^8-30 mQ3^4 mU3^8-180 m3^2 msq^2 mU3^8+30 msq^4 mU3^8-604 m3^2 mU3^10+140 mQ3^2 mU3^10+169 mU3^12) Xt^4)/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)^4)+(32 (18 m3^7 mQ3^4-m3^5 mQ3^6-3 m3^3 mQ3^8-30 m3^3 mQ3^4 msq^4-32 m3^7 mQ3^2 mU3^2-70 m3^5 mQ3^4 mU3^2+8 m3^3 mQ3^6 mU3^2+3 m3 mQ3^8 mU3^2+60 m3^3 mQ3^4 msq^2 mU3^2+30 m3 mQ3^4 msq^4 mU3^2+14 m3^7 mU3^4+145 m3^5 mQ3^2 mU3^4+94 m3^3 mQ3^4 mU3^4-7 m3 mQ3^6 mU3^4-60 m3 mQ3^4 msq^2 mU3^4+30 m3^3 msq^4 mU3^4+86 m3^5 mU3^6-200 m3^3 mQ3^2 mU3^6-26 m3 mQ3^4 mU3^6-60 m3^3 msq^2 mU3^6-30 m3 msq^4 mU3^6-219 m3^3 mU3^8+87 m3 mQ3^2 mU3^8+60 m3 msq^2 mU3^8+103 m3 mU3^10) Xt^5)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^5)+(1280 m3^2 mU3^4 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^5)) Log[mU3^2/M^2]^3+Log[M^2/mt^2]^2 (-((64 (3 m3^4+m3^2 mQ3^2+m3^2 mU3^2-5 mQ3^2 mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2)))+(192+(256 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(1024 m3^3 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[m3^2/M^2]+((32 (m3^4-18 m3^2 mQ3^2+9 mQ3^4))/(m3^2-mQ3^2)^2+(1024 m3 mQ3^2 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))) Log[mQ3^2/M^2]+144 Log[msq^2/M^2]+((16 (m3^4-34 m3^2 mU3^2+17 mU3^4))/(m3^2-mU3^2)^2+(1024 m3 mU3^2 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))) Log[mU3^2/M^2])+Log[m3^2/M^2]^2 (-((8 (88 m3^18 mQ3^2-288 m3^16 mQ3^4+368 m3^14 mQ3^6-224 m3^12 mQ3^8+56 m3^10 mQ3^10+88 m3^18 mU3^2-302 m3^16 mQ3^2 mU3^2+838 m3^14 mQ3^4 mU3^2-797 m3^12 mQ3^6 mU3^2+280 m3^10 mQ3^8 mU3^2-39 m3^8 mQ3^10 mU3^2-288 m3^16 mU3^4+838 m3^14 mQ3^2 mU3^4-2636 m3^12 mQ3^4 mU3^4+2588 m3^10 mQ3^6 mU3^4-908 m3^8 mQ3^8 mU3^4+134 m3^6 mQ3^10 mU3^4+368 m3^14 mU3^6-797 m3^12 mQ3^2 mU3^6+2588 m3^10 mQ3^4 mU3^6-2048 m3^8 mQ3^6 mU3^6+272 m3^6 mQ3^8 mU3^6+25 m3^4 mQ3^10 mU3^6-224 m3^12 mU3^8+280 m3^10 mQ3^2 mU3^8-908 m3^8 mQ3^4 mU3^8+272 m3^6 mQ3^6 mU3^8+452 m3^4 mQ3^8 mU3^8-144 m3^2 mQ3^10 mU3^8+56 m3^10 mU3^10-39 m3^8 mQ3^2 mU3^10+134 m3^6 mQ3^4 mU3^10+25 m3^4 mQ3^6 mU3^10-144 m3^2 mQ3^8 mU3^10+36 mQ3^10 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4))+(64 (16 m3^13 mQ3^2-32 m3^11 mQ3^4+16 m3^9 mQ3^6+16 m3^13 mU3^2+31 m3^11 mQ3^2 mU3^2+18 m3^9 mQ3^4 mU3^2-25 m3^7 mQ3^6 mU3^2-32 m3^11 mU3^4+18 m3^9 mQ3^2 mU3^4-196 m3^7 mQ3^4 mU3^4+130 m3^5 mQ3^6 mU3^4+16 m3^9 mU3^6-25 m3^7 mQ3^2 mU3^6+130 m3^5 mQ3^4 mU3^6-81 m3^3 mQ3^6 mU3^6) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)+(-((1792 m3^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(64 m3^4 (-m3^2+mQ3^2+mU3^2) (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))) Xt^2-(2048 m3^7 (m3^2-mQ3^2-mU3^2) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)-(16 (44 m3^18 mQ3^2-144 m3^16 mQ3^4+184 m3^14 mQ3^6-112 m3^12 mQ3^8+28 m3^10 mQ3^10+44 m3^18 mU3^2-2 m3^16 mQ3^2 mU3^2-294 m3^14 mQ3^4 mU3^2+277 m3^12 mQ3^6 mU3^2+4 m3^10 mQ3^8 mU3^2-25 m3^8 mQ3^10 mU3^2-144 m3^16 mU3^4-294 m3^14 mQ3^2 mU3^4+1932 m3^12 mQ3^4 mU3^4-2012 m3^10 mQ3^6 mU3^4+572 m3^8 mQ3^8 mU3^4-70 m3^6 mQ3^10 mU3^4+184 m3^14 mU3^6+277 m3^12 mQ3^2 mU3^6-2012 m3^10 mQ3^4 mU3^6+1712 m3^8 mQ3^6 mU3^6-88 m3^6 mQ3^8 mU3^6-49 m3^4 mQ3^10 mU3^6-112 m3^12 mU3^8+4 m3^10 mQ3^2 mU3^8+572 m3^8 mQ3^4 mU3^8-88 m3^6 mQ3^6 mU3^8-548 m3^4 mQ3^8 mU3^8+156 m3^2 mQ3^10 mU3^8+28 m3^10 mU3^10-25 m3^8 mQ3^2 mU3^10-70 m3^6 mQ3^4 mU3^10-49 m3^4 mQ3^6 mU3^10+156 m3^2 mQ3^8 mU3^10-36 mQ3^10 mU3^10) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)+(128 (8 m3^13 mQ3^2-16 m3^11 mQ3^4+8 m3^9 mQ3^6+8 m3^13 mU3^2+7 m3^11 mQ3^2 mU3^2-54 m3^9 mQ3^4 mU3^2+47 m3^7 mQ3^6 mU3^2-16 m3^11 mU3^4-54 m3^9 mQ3^2 mU3^4+188 m3^7 mQ3^4 mU3^4-134 m3^5 mQ3^6 mU3^4+8 m3^9 mU3^6+47 m3^7 mQ3^2 mU3^6-134 m3^5 mQ3^4 mU3^6+87 m3^3 mQ3^6 mU3^6) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(2560 m3^6 Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(-((8 (198 m3^16-814 m3^14 mQ3^2+877 m3^12 mQ3^4-344 m3^10 mQ3^6+55 m3^8 mQ3^8-638 m3^14 mU3^2+2700 m3^12 mQ3^2 mU3^2-2796 m3^10 mQ3^4 mU3^2+1004 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2+565 m3^12 mU3^4-2604 m3^10 mQ3^2 mU3^4+2240 m3^8 mQ3^4 mU3^4-368 m3^6 mQ3^6 mU3^4-m3^4 mQ3^8 mU3^4-120 m3^10 mU3^6+812 m3^8 mQ3^2 mU3^6-272 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6-m3^8 mU3^8-110 m3^6 mQ3^2 mU3^8-25 m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4))-(64 (44 m3^13-79 m3^11 mQ3^2+122 m3^9 mQ3^4-63 m3^7 mQ3^6-77 m3^11 mU3^2+72 m3^9 mQ3^2 mU3^2-209 m3^7 mQ3^4 mU3^2+142 m3^5 mQ3^6 mU3^2+22 m3^9 mU3^4+101 m3^7 mQ3^2 mU3^4+36 m3^5 mQ3^4 mU3^4-87 m3^3 mQ3^6 mU3^4+19 m3^7 mU3^6-118 m3^5 mQ3^2 mU3^6+75 m3^3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(16 (30 m3^16-334 m3^14 mQ3^2+517 m3^12 mQ3^4-264 m3^10 mQ3^6+55 m3^8 mQ3^8-238 m3^14 mU3^2+1484 m3^12 mQ3^2 mU3^2-1948 m3^10 mQ3^4 mU3^2+844 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2+325 m3^12 mU3^4-1660 m3^10 mQ3^2 mU3^4+1648 m3^8 mQ3^4 mU3^4-288 m3^6 mQ3^6 mU3^4-m3^4 mQ3^8 mU3^4-168 m3^10 mU3^6+652 m3^8 mQ3^2 mU3^6-192 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+55 m3^8 mU3^8-158 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))-(128 (44 m3^15-100 m3^13 mQ3^2+45 m3^11 mQ3^4+66 m3^9 mQ3^6-63 m3^7 mQ3^8-100 m3^13 mU3^2+282 m3^11 mQ3^2 mU3^2-250 m3^9 mQ3^4 mU3^2-42 m3^7 mQ3^6 mU3^2+142 m3^5 mQ3^8 mU3^2+45 m3^11 mU3^4-250 m3^9 mQ3^2 mU3^4+422 m3^7 mQ3^4 mU3^4-178 m3^5 mQ3^6 mU3^4-87 m3^3 mQ3^8 mU3^4+66 m3^9 mU3^6-42 m3^7 mQ3^2 mU3^6-178 m3^5 mQ3^4 mU3^6+186 m3^3 mQ3^6 mU3^6-63 m3^7 mU3^8+142 m3^5 mQ3^2 mU3^8-87 m3^3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(8 (1024 m3^18-2786 m3^16 mQ3^2+1826 m3^14 mQ3^4+549 m3^12 mQ3^6-664 m3^10 mQ3^8+55 m3^8 mQ3^10-2786 m3^16 mU3^2+8004 m3^14 mQ3^2 mU3^2-5903 m3^12 mQ3^4 mU3^2-868 m3^10 mQ3^6 mU3^2+1699 m3^8 mQ3^8 mU3^2-158 m3^6 mQ3^10 mU3^2+1826 m3^14 mU3^4-5903 m3^12 mQ3^2 mU3^4+5224 m3^10 mQ3^4 mU3^4-292 m3^8 mQ3^6 mU3^4-846 m3^6 mQ3^8 mU3^4-m3^4 mQ3^10 mU3^4+549 m3^12 mU3^6-868 m3^10 mQ3^2 mU3^6-292 m3^8 mQ3^4 mU3^6+928 m3^6 mQ3^6 mU3^6-453 m3^4 mQ3^8 mU3^6+144 m3^2 mQ3^10 mU3^6-664 m3^10 mU3^8+1699 m3^8 mQ3^2 mU3^8-846 m3^6 mQ3^4 mU3^8-453 m3^4 mQ3^6 mU3^8+288 m3^2 mQ3^8 mU3^8-36 mQ3^10 mU3^8+55 m3^8 mU3^10-158 m3^6 mQ3^2 mU3^10-m3^4 mQ3^4 mU3^10+144 m3^2 mQ3^6 mU3^10-36 mQ3^8 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3)-(64 (39 m3^11 mQ3^2-94 m3^9 mQ3^4+63 m3^7 mQ3^6+39 m3^11 mU3^2-188 m3^9 mQ3^2 mU3^2+283 m3^7 mQ3^4 mU3^2-142 m3^5 mQ3^6 mU3^2-94 m3^9 mU3^4+283 m3^7 mQ3^2 mU3^4-284 m3^5 mQ3^4 mU3^4+87 m3^3 mQ3^6 mU3^4+63 m3^7 mU3^6-142 m3^5 mQ3^2 mU3^6+87 m3^3 mQ3^4 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(1280 m3^6 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[mQ3^2/M^2]+(-((8 (198 m3^16-638 m3^14 mQ3^2+565 m3^12 mQ3^4-120 m3^10 mQ3^6-m3^8 mQ3^8-814 m3^14 mU3^2+2700 m3^12 mQ3^2 mU3^2-2604 m3^10 mQ3^4 mU3^2+812 m3^8 mQ3^6 mU3^2-110 m3^6 mQ3^8 mU3^2+877 m3^12 mU3^4-2796 m3^10 mQ3^2 mU3^4+2240 m3^8 mQ3^4 mU3^4-272 m3^6 mQ3^6 mU3^4-25 m3^4 mQ3^8 mU3^4-344 m3^10 mU3^6+1004 m3^8 mQ3^2 mU3^6-368 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+55 m3^8 mU3^8-158 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4))+(64 (44 m3^13-77 m3^11 mQ3^2+22 m3^9 mQ3^4+19 m3^7 mQ3^6-79 m3^11 mU3^2+72 m3^9 mQ3^2 mU3^2+101 m3^7 mQ3^4 mU3^2-118 m3^5 mQ3^6 mU3^2+122 m3^9 mU3^4-209 m3^7 mQ3^2 mU3^4+36 m3^5 mQ3^4 mU3^4+75 m3^3 mQ3^6 mU3^4-63 m3^7 mU3^6+142 m3^5 mQ3^2 mU3^6-87 m3^3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(16 (30 m3^16-238 m3^14 mQ3^2+325 m3^12 mQ3^4-168 m3^10 mQ3^6+55 m3^8 mQ3^8-334 m3^14 mU3^2+1484 m3^12 mQ3^2 mU3^2-1660 m3^10 mQ3^4 mU3^2+652 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2+517 m3^12 mU3^4-1948 m3^10 mQ3^2 mU3^4+1648 m3^8 mQ3^4 mU3^4-192 m3^6 mQ3^6 mU3^4-m3^4 mQ3^8 mU3^4-264 m3^10 mU3^6+844 m3^8 mQ3^2 mU3^6-288 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+55 m3^8 mU3^8-158 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2))+(128 (44 m3^15-100 m3^13 mQ3^2+45 m3^11 mQ3^4+66 m3^9 mQ3^6-63 m3^7 mQ3^8-100 m3^13 mU3^2+282 m3^11 mQ3^2 mU3^2-250 m3^9 mQ3^4 mU3^2-42 m3^7 mQ3^6 mU3^2+142 m3^5 mQ3^8 mU3^2+45 m3^11 mU3^4-250 m3^9 mQ3^2 mU3^4+422 m3^7 mQ3^4 mU3^4-178 m3^5 mQ3^6 mU3^4-87 m3^3 mQ3^8 mU3^4+66 m3^9 mU3^6-42 m3^7 mQ3^2 mU3^6-178 m3^5 mQ3^4 mU3^6+186 m3^3 mQ3^6 mU3^6-63 m3^7 mU3^8+142 m3^5 mQ3^2 mU3^8-87 m3^3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(8 (1024 m3^18-2786 m3^16 mQ3^2+1826 m3^14 mQ3^4+549 m3^12 mQ3^6-664 m3^10 mQ3^8+55 m3^8 mQ3^10-2786 m3^16 mU3^2+8004 m3^14 mQ3^2 mU3^2-5903 m3^12 mQ3^4 mU3^2-868 m3^10 mQ3^6 mU3^2+1699 m3^8 mQ3^8 mU3^2-158 m3^6 mQ3^10 mU3^2+1826 m3^14 mU3^4-5903 m3^12 mQ3^2 mU3^4+5224 m3^10 mQ3^4 mU3^4-292 m3^8 mQ3^6 mU3^4-846 m3^6 mQ3^8 mU3^4-m3^4 mQ3^10 mU3^4+549 m3^12 mU3^6-868 m3^10 mQ3^2 mU3^6-292 m3^8 mQ3^4 mU3^6+928 m3^6 mQ3^6 mU3^6-453 m3^4 mQ3^8 mU3^6+144 m3^2 mQ3^10 mU3^6-664 m3^10 mU3^8+1699 m3^8 mQ3^2 mU3^8-846 m3^6 mQ3^4 mU3^8-453 m3^4 mQ3^6 mU3^8+288 m3^2 mQ3^8 mU3^8-36 mQ3^10 mU3^8+55 m3^8 mU3^10-158 m3^6 mQ3^2 mU3^10-m3^4 mQ3^4 mU3^10+144 m3^2 mQ3^6 mU3^10-36 mQ3^8 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3)+(64 (39 m3^11 mQ3^2-94 m3^9 mQ3^4+63 m3^7 mQ3^6+39 m3^11 mU3^2-188 m3^9 mQ3^2 mU3^2+283 m3^7 mQ3^4 mU3^2-142 m3^5 mQ3^6 mU3^2-94 m3^9 mU3^4+283 m3^7 mQ3^2 mU3^4-284 m3^5 mQ3^4 mU3^4+87 m3^3 mQ3^6 mU3^4+63 m3^7 mU3^6-142 m3^5 mQ3^2 mU3^6+87 m3^3 mQ3^4 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(1280 m3^6 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[mU3^2/M^2])+Log[mQ3^2/M^2]^2 ((4 (128 m3^16-1016 m3^14 mQ3^2+1308 m3^12 mQ3^4-307 m3^10 mQ3^6-528 m3^8 mQ3^8+232 m3^6 mQ3^10+56 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+300 m3^10 mQ3^4 msq^2-300 m3^8 mQ3^6 msq^2-60 m3^6 mQ3^8 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-392 m3^14 mU3^2+4132 m3^12 mQ3^2 mU3^2-5861 m3^10 mQ3^4 mU3^2+2119 m3^8 mQ3^6 mU3^2+1686 m3^6 mQ3^8 mU3^2-1172 m3^4 mQ3^10 mU3^2+35 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-480 m3^10 mQ3^2 msq^2 mU3^2-600 m3^8 mQ3^4 msq^2 mU3^2+960 m3^6 mQ3^6 msq^2 mU3^2+180 m3^4 mQ3^8 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+320 m3^12 mU3^4-6037 m3^10 mQ3^2 mU3^4+9653 m3^8 mQ3^4 mU3^4-5278 m3^6 mQ3^6 mU3^4-446 m3^4 mQ3^8 mU3^4+979 m3^2 mQ3^10 mU3^4-7 mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+1080 m3^8 mQ3^2 msq^2 mU3^4-1080 m3^4 mQ3^6 msq^2 mU3^4-180 m3^2 mQ3^8 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4+45 m3^10 mU3^6+3565 m3^8 mQ3^2 mU3^6-6262 m3^6 mQ3^4 mU3^6+4086 m3^4 mQ3^6 mU3^6-599 m3^2 mQ3^8 mU3^6-291 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-960 m3^6 mQ3^2 msq^2 mU3^6+600 m3^4 mQ3^4 msq^2 mU3^6+480 m3^2 mQ3^6 msq^2 mU3^6+60 mQ3^8 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-89 m3^8 mU3^8-746 m3^6 mQ3^2 mU3^8+1426 m3^4 mQ3^4 mU3^8-1034 m3^2 mQ3^6 mU3^8+307 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+300 m3^4 mQ3^2 msq^2 mU3^8-300 m3^2 mQ3^4 msq^2 mU3^8-60 mQ3^6 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8+18 m3^4 mQ3^2 mU3^10-12 m3^2 mQ3^4 mU3^10-6 mQ3^6 mU3^10))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(32 (34 m3^11 mQ3^2-402 m3^9 mQ3^4+488 m3^7 mQ3^6-211 m3^5 mQ3^8+14 m3^3 mQ3^10-3 m3 mQ3^12+120 m3^7 mQ3^4 msq^2-120 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-34 m3^11 mU3^2+349 m3^9 mQ3^2 mU3^2+266 m3^7 mQ3^4 mU3^2-687 m3^5 mQ3^6 mU3^2+342 m3^3 mQ3^8 mU3^2+4 m3 mQ3^10 mU3^2-120 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+240 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+37 m3^9 mU3^4-684 m3^7 mQ3^2 mU3^4+477 m3^5 mQ3^4 mU3^4+106 m3^3 mQ3^6 mU3^4-176 m3 mQ3^8 mU3^4+240 m3^5 mQ3^2 msq^2 mU3^4-120 m3^3 mQ3^4 msq^2 mU3^4-120 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-6 m3^7 mU3^6+325 m3^5 mQ3^2 mU3^6-392 m3^3 mQ3^4 mU3^6+153 m3 mQ3^6 mU3^6-120 m3^3 mQ3^2 msq^2 mU3^6+120 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6-6 m3^3 mQ3^2 mU3^8+6 m3 mQ3^4 mU3^8) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(16 (486 m3^10 mQ3^2-367 m3^8 mQ3^4-240 m3^6 mQ3^6+196 m3^4 mQ3^8+9 m3^2 mQ3^10-90 m3^6 mQ3^4 msq^2-30 m3^4 mQ3^6 msq^2-18 m3^10 mU3^2-1466 m3^8 mQ3^2 mU3^2+1455 m3^6 mQ3^4 mU3^2+161 m3^4 mQ3^6 mU3^2-399 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2+90 m3^6 mQ3^2 msq^2 mU3^2+210 m3^4 mQ3^4 msq^2 mU3^2+60 m3^2 mQ3^6 msq^2 mU3^2+53 m3^8 mU3^4+1372 m3^6 mQ3^2 mU3^4-1624 m3^4 mQ3^4 mU3^4+332 m3^2 mQ3^6 mU3^4+155 mQ3^8 mU3^4-180 m3^4 mQ3^2 msq^2 mU3^4-150 m3^2 mQ3^4 msq^2 mU3^4-30 mQ3^6 msq^2 mU3^4-35 m3^6 mU3^6-389 m3^4 mQ3^2 mU3^6+485 m3^2 mQ3^4 mU3^6-181 mQ3^6 mU3^6+90 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6+9 m3^2 mQ3^2 mU3^8+3 mQ3^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(64 (22 m3^9-77 m3^7 mQ3^2-90 m3^5 mQ3^4+129 m3^3 mQ3^6-6 m3 mQ3^8+60 m3^3 mQ3^4 msq^2+13 m3^7 mU3^2+305 m3^5 mQ3^2 mU3^2-101 m3^3 mQ3^4 mU3^2-103 m3 mQ3^6 mU3^2-60 m3^3 mQ3^2 msq^2 mU3^2-60 m3 mQ3^4 msq^2 mU3^2-27 m3^5 mU3^4-236 m3^3 mQ3^2 mU3^4+165 m3 mQ3^4 mU3^4+60 m3 mQ3^2 msq^2 mU3^4+6 m3 mQ3^2 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(8 (180 m3^16 mQ3^2-2588 m3^14 mQ3^4+5659 m3^12 mQ3^6-3988 m3^10 mQ3^8+364 m3^8 mQ3^10+343 m3^6 mQ3^12+47 m3^4 mQ3^14-9 m3^2 mQ3^16+60 m3^12 mQ3^4 msq^2+210 m3^10 mQ3^6 msq^2-240 m3^8 mQ3^8 msq^2-30 m3^6 mQ3^10 msq^2-90 m3^10 mQ3^4 msq^4+60 m3^8 mQ3^6 msq^4+30 m3^6 mQ3^8 msq^4-180 m3^16 mU3^2-300 m3^14 mQ3^2 mU3^2+8831 m3^12 mQ3^4 mU3^2-19173 m3^10 mQ3^6 mU3^2+13130 m3^8 mQ3^8 mU3^2-969 m3^6 mQ3^10 mU3^2-1426 m3^4 mQ3^12 mU3^2+50 m3^2 mQ3^14 mU3^2-3 mQ3^16 mU3^2-120 m3^12 mQ3^2 msq^2 mU3^2-420 m3^10 mQ3^4 msq^2 mU3^2-270 m3^8 mQ3^6 msq^2 mU3^2+720 m3^6 mQ3^8 msq^2 mU3^2+90 m3^4 mQ3^10 msq^2 mU3^2+180 m3^10 mQ3^2 msq^4 mU3^2+150 m3^8 mQ3^4 msq^4 mU3^2-240 m3^6 mQ3^6 msq^4 mU3^2-90 m3^4 mQ3^8 msq^4 mU3^2+504 m3^14 mU3^4+369 m3^12 mQ3^2 mU3^4-12473 m3^10 mQ3^4 mU3^4+25400 m3^8 mQ3^6 mU3^4-17072 m3^6 mQ3^8 mU3^4+2250 m3^4 mQ3^10 mU3^4+1103 m3^2 mQ3^12 mU3^4-mQ3^14 mU3^4+60 m3^12 msq^2 mU3^4+390 m3^10 mQ3^2 msq^2 mU3^4+780 m3^8 mQ3^4 msq^2 mU3^4-420 m3^6 mQ3^6 msq^2 mU3^4-720 m3^4 mQ3^8 msq^2 mU3^4-90 m3^2 mQ3^10 msq^2 mU3^4-90 m3^10 msq^4 mU3^4-480 m3^8 mQ3^2 msq^4 mU3^4+120 m3^6 mQ3^4 msq^4 mU3^4+360 m3^4 mQ3^6 msq^4 mU3^4+90 m3^2 mQ3^8 msq^4 mU3^4-475 m3^12 mU3^6-743 m3^10 mQ3^2 mU3^6+9104 m3^8 mQ3^4 mU3^6-15392 m3^6 mQ3^6 mU3^6+9007 m3^4 mQ3^8 mU3^6-1225 m3^2 mQ3^10 mU3^6-356 mQ3^12 mU3^6-180 m3^10 msq^2 mU3^6-450 m3^8 mQ3^2 msq^2 mU3^6-420 m3^6 mQ3^4 msq^2 mU3^6+780 m3^4 mQ3^6 msq^2 mU3^6+240 m3^2 mQ3^8 msq^2 mU3^6+30 mQ3^10 msq^2 mU3^6+270 m3^8 msq^4 mU3^6+360 m3^6 mQ3^2 msq^4 mU3^6-360 m3^4 mQ3^4 msq^4 mU3^6-240 m3^2 mQ3^6 msq^4 mU3^6-30 mQ3^8 msq^4 mU3^6+137 m3^10 mU3^8+880 m3^8 mQ3^2 mU3^8-3971 m3^6 mQ3^4 mU3^8+4850 m3^4 mQ3^6 mU3^8-2108 m3^2 mQ3^8 mU3^8+252 mQ3^10 mU3^8+180 m3^8 msq^2 mU3^8+210 m3^6 mQ3^2 msq^2 mU3^8-120 m3^4 mQ3^4 msq^2 mU3^8-270 m3^2 mQ3^6 msq^2 mU3^8-270 m3^6 msq^4 mU3^8+210 m3^2 mQ3^4 msq^4 mU3^8+60 mQ3^6 msq^4 mU3^8+2 m3^8 mU3^10-299 m3^6 mQ3^2 mU3^10+767 m3^4 mQ3^4 mU3^10-669 m3^2 mQ3^6 mU3^10+191 mQ3^8 mU3^10-60 m3^6 msq^2 mU3^10-30 m3^4 mQ3^2 msq^2 mU3^10+120 m3^2 mQ3^4 msq^2 mU3^10-30 mQ3^6 msq^2 mU3^10+90 m3^4 msq^4 mU3^10-60 m3^2 mQ3^2 msq^4 mU3^10-30 mQ3^4 msq^4 mU3^10+9 m3^4 mQ3^2 mU3^12-6 m3^2 mQ3^4 mU3^12-3 mQ3^6 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(64 (30 m3^11 mQ3^2+179 m3^9 mQ3^4-442 m3^7 mQ3^6+225 m3^5 mQ3^8-11 m3^3 mQ3^10+3 m3 mQ3^12-90 m3^7 mQ3^4 msq^2+90 m3^5 mQ3^6 msq^2+30 m3^7 mQ3^2 msq^4-30 m3^5 mQ3^4 msq^4+34 m3^11 mU3^2-36 m3^9 mQ3^2 mU3^2-470 m3^7 mQ3^4 mU3^2+923 m3^5 mQ3^6 mU3^2-396 m3^3 mQ3^8 mU3^2-7 m3 mQ3^10 mU3^2+30 m3^7 mQ3^2 msq^2 mU3^2+150 m3^5 mQ3^4 msq^2 mU3^2-180 m3^3 mQ3^6 msq^2 mU3^2-30 m3^7 msq^4 mU3^2-30 m3^5 mQ3^2 msq^4 mU3^2+60 m3^3 mQ3^4 msq^4 mU3^2-69 m3^9 mU3^4-14 m3^7 mQ3^2 mU3^4+427 m3^5 mQ3^4 mU3^4-608 m3^3 mQ3^6 mU3^4+216 m3 mQ3^8 mU3^4-60 m3^5 mQ3^2 msq^2 mU3^4-30 m3^3 mQ3^4 msq^2 mU3^4+90 m3 mQ3^6 msq^2 mU3^4+60 m3^5 msq^4 mU3^4-30 m3^3 mQ3^2 msq^4 mU3^4-30 m3 mQ3^4 msq^4 mU3^4+38 m3^7 mU3^6+5 m3^5 mQ3^2 mU3^6-94 m3^3 mQ3^4 mU3^6+67 m3 mQ3^6 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6-30 m3 mQ3^4 msq^2 mU3^6-30 m3^3 msq^4 mU3^6+30 m3 mQ3^2 msq^4 mU3^6-3 m3^3 mQ3^2 mU3^8+3 m3 mQ3^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)-(2560 m3^2 mQ3^4 Xt^6)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)+(-(1/(3 (m3^2-mQ3^2)^3))8 (27 m3^6-65 m3^4 mQ3^2+57 m3^2 mQ3^4-19 mQ3^6-60 m3^4 msq^2-180 m3^2 mQ3^2 msq^2+90 m3^2 msq^4+30 mQ3^2 msq^4)+(64 (9 m3^5-14 m3^3 mQ3^2+5 m3 mQ3^4-60 m3 mQ3^2 msq^2+30 m3 msq^4) Xt)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+((16 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+60 m3^4 msq^2+180 m3^2 mQ3^2 msq^2-90 m3^2 msq^4-30 mQ3^2 msq^4))/(3 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2))-(48 (4 mQ3^6-2 mQ3^4 mU3^2+m3^4 (3 mQ3^2-mU3^2)+m3^2 (-8 mQ3^4+4 mQ3^2 mU3^2)))/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)) Xt^2+(-((64 (m3^3 mQ3^2-m3 mQ3^4+120 m3 mQ3^2 msq^2-60 m3 msq^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2))+(96 m3 (2 m3^4-mQ3^4+3 mQ3^2 mU3^2-m3^2 (3 mQ3^2+mU3^2)))/((m3^2-mQ3^2) (mQ3^2-mU3^2)^3)) Xt^3+(-((8 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+60 m3^4 msq^2+180 m3^2 mQ3^2 msq^2-90 m3^2 msq^4-30 mQ3^2 msq^4) (mQ3^2+mU3^2))/(3 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^3))+(24 (5 mQ3^8+8 m3^4 mQ3^2 mU3^2+4 mQ3^6 mU3^2-mQ3^4 mU3^4+2 m3^6 (mQ3^2-mU3^2)-2 m3^2 (4 mQ3^6+5 mQ3^4 mU3^2-mQ3^2 mU3^4)))/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) Xt^4+(320 (m3^3 mQ3^2-m3 mQ3^4+12 m3 mQ3^2 msq^2-6 m3 msq^4) (mQ3^2+mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) Log[msq^2/M^2]+(-((4 (128 m3^16-550 m3^14 mQ3^2+670 m3^12 mQ3^4-247 m3^10 mQ3^6-64 m3^8 mQ3^8+100 m3^6 mQ3^10-38 m3^4 mQ3^12+9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-346 m3^14 mU3^2+1542 m3^12 mQ3^2 mU3^2-1619 m3^10 mQ3^4 mU3^2+199 m3^8 mQ3^6 mU3^2+324 m3^6 mQ3^8 mU3^2-88 m3^4 mQ3^10 mU3^2-47 m3^2 mQ3^12 mU3^2+3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+476 m3^12 mU3^4-2307 m3^10 mQ3^2 mU3^4+2585 m3^8 mQ3^4 mU3^4-386 m3^6 mQ3^6 mU3^4-638 m3^4 mQ3^8 mU3^4+317 m3^2 mQ3^10 mU3^4+mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-307 m3^10 mU3^6+1699 m3^8 mQ3^2 mU3^6-2306 m3^6 mQ3^4 mU3^6+1002 m3^4 mQ3^6 mU3^6-19 m3^2 mQ3^8 mU3^6-101 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6+61 m3^8 mU3^8-420 m3^6 mQ3^2 mU3^8+658 m3^4 mQ3^4 mU3^8-388 m3^2 mQ3^6 mU3^8+97 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(32 (20 m3^11 mQ3^2-91 m3^9 mQ3^4+122 m3^7 mQ3^6-30 m3^5 mQ3^8-8 m3^3 mQ3^10+3 m3 mQ3^12+60 m3^7 mQ3^4 msq^2-60 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-20 m3^11 mU3^2+51 m3^9 mQ3^2 mU3^2+65 m3^7 mQ3^4 mU3^2-219 m3^5 mQ3^6 mU3^2+85 m3^3 mQ3^8 mU3^2-10 m3 mQ3^10 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+56 m3^9 mU3^4-244 m3^7 mQ3^2 mU3^4+278 m3^5 mQ3^4 mU3^4-42 m3^3 mQ3^6 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4-60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-7 m3^7 mU3^6+67 m3^5 mQ3^2 mU3^6-99 m3^3 mQ3^4 mU3^6+23 m3 mQ3^6 mU3^6-60 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(8 (128 m3^16-554 m3^14 mQ3^2+446 m3^12 mQ3^4+265 m3^10 mQ3^6-272 m3^8 mQ3^8-92 m3^6 mQ3^10+114 m3^4 mQ3^12-27 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-842 m3^14 mU3^2+3562 m3^12 mQ3^2 mU3^2-4347 m3^10 mQ3^4 mU3^2+1295 m3^8 mQ3^6 mU3^2+688 m3^6 mQ3^8 mU3^2-520 m3^4 mQ3^10 mU3^2+141 m3^2 mQ3^12 mU3^2-9 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+1188 m3^12 mU3^4-4987 m3^10 mQ3^2 mU3^4+6721 m3^8 mQ3^4 mU3^4-3662 m3^6 mQ3^6 mU3^4+950 m3^4 mQ3^8 mU3^4-159 m3^2 mQ3^10 mU3^4-3 mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4-375 m3^10 mU3^6+1523 m3^8 mQ3^2 mU3^6-1766 m3^6 mQ3^4 mU3^6+670 m3^4 mQ3^6 mU3^6-123 m3^2 mQ3^8 mU3^6+39 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-87 m3^8 mU3^8+324 m3^6 mQ3^2 mU3^8-538 m3^4 mQ3^4 mU3^8+412 m3^2 mQ3^6 mU3^8-103 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(32 (2 m3^13-43 m3^11 mQ3^2+243 m3^9 mQ3^4-405 m3^7 mQ3^6+201 m3^5 mQ3^8-48 m3^3 mQ3^10+18 m3 mQ3^12-120 m3^7 mQ3^4 msq^2+120 m3^5 mQ3^6 msq^2+60 m3^7 mQ3^2 msq^4-60 m3^5 mQ3^4 msq^4-33 m3^11 mU3^2+403 m3^9 mQ3^2 mU3^2-1127 m3^7 mQ3^4 mU3^2+1201 m3^5 mQ3^6 mU3^2-288 m3^3 mQ3^8 mU3^2-60 m3 mQ3^10 mU3^2+120 m3^7 mQ3^2 msq^2 mU3^2+120 m3^5 mQ3^4 msq^2 mU3^2-240 m3^3 mQ3^6 msq^2 mU3^2-60 m3^7 msq^4 mU3^2-60 m3^5 mQ3^2 msq^4 mU3^2+120 m3^3 mQ3^4 msq^4 mU3^2-40 m3^9 mU3^4-233 m3^7 mQ3^2 mU3^4+959 m3^5 mQ3^4 mU3^4-1143 m3^3 mQ3^6 mU3^4+361 m3 mQ3^8 mU3^4-240 m3^5 mQ3^2 msq^2 mU3^4+120 m3^3 mQ3^4 msq^2 mU3^4+120 m3 mQ3^6 msq^2 mU3^4+120 m3^5 msq^4 mU3^4-60 m3^3 mQ3^2 msq^4 mU3^4-60 m3 mQ3^4 msq^4 mU3^4+61 m3^7 mU3^6-155 m3^5 mQ3^2 mU3^6+123 m3^3 mQ3^4 mU3^6+3 m3 mQ3^6 mU3^6+120 m3^3 mQ3^2 msq^2 mU3^6-120 m3 mQ3^4 msq^2 mU3^6-60 m3^3 msq^4 mU3^6+60 m3 mQ3^2 msq^4 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(4 (124 m3^16 mQ3^2-532 m3^14 mQ3^4-210 m3^12 mQ3^6+1597 m3^10 mQ3^8-1036 m3^8 mQ3^10-22 m3^6 mQ3^12+114 m3^4 mQ3^14-27 m3^2 mQ3^16+60 m3^12 mQ3^4 msq^2+120 m3^10 mQ3^6 msq^2-180 m3^8 mQ3^8 msq^2-90 m3^10 mQ3^4 msq^4+60 m3^8 mQ3^6 msq^4+30 m3^6 mQ3^8 msq^4+1156 m3^16 mU3^2-5356 m3^14 mQ3^2 mU3^2+9070 m3^12 mQ3^4 mU3^2-3338 m3^10 mQ3^6 mU3^2-5097 m3^8 mQ3^8 mU3^2+3988 m3^6 mQ3^10 mU3^2-552 m3^4 mQ3^12 mU3^2+114 m3^2 mQ3^14 mU3^2-9 mQ3^16 mU3^2-180 m3^10 mQ3^4 msq^2 mU3^2-360 m3^8 mQ3^6 msq^2 mU3^2+540 m3^6 mQ3^8 msq^2 mU3^2+270 m3^8 mQ3^4 msq^4 mU3^2-180 m3^6 mQ3^6 msq^4 mU3^2-90 m3^4 mQ3^8 msq^4 mU3^2-3024 m3^14 mU3^4+14710 m3^12 mQ3^2 mU3^4-25608 m3^10 mQ3^4 mU3^4+15072 m3^8 mQ3^6 mU3^4+2920 m3^6 mQ3^8 mU3^4-4138 m3^4 mQ3^10 mU3^4+96 m3^2 mQ3^12 mU3^4-12 mQ3^14 mU3^4-60 m3^12 msq^2 mU3^4-120 m3^10 mQ3^2 msq^2 mU3^4+360 m3^8 mQ3^4 msq^2 mU3^4+360 m3^6 mQ3^6 msq^2 mU3^4-540 m3^4 mQ3^8 msq^2 mU3^4+90 m3^10 msq^4 mU3^4-60 m3^8 mQ3^2 msq^4 mU3^4-300 m3^6 mQ3^4 msq^4 mU3^4+180 m3^4 mQ3^6 msq^4 mU3^4+90 m3^2 mQ3^8 msq^4 mU3^4+2462 m3^12 mU3^6-13106 m3^10 mQ3^2 mU3^6+24682 m3^8 mQ3^4 mU3^6-18228 m3^6 mQ3^6 mU3^6+2474 m3^4 mQ3^8 mU3^6+1734 m3^2 mQ3^10 mU3^6-2 mQ3^12 mU3^6+180 m3^10 msq^2 mU3^6+360 m3^8 mQ3^2 msq^2 mU3^6-600 m3^6 mQ3^4 msq^2 mU3^6-120 m3^4 mQ3^6 msq^2 mU3^6+180 m3^2 mQ3^8 msq^2 mU3^6-270 m3^8 msq^4 mU3^6+180 m3^6 mQ3^2 msq^4 mU3^6+180 m3^4 mQ3^4 msq^4 mU3^6-60 m3^2 mQ3^6 msq^4 mU3^6-30 mQ3^8 msq^4 mU3^6-521 m3^10 mU3^8+3160 m3^8 mQ3^2 mU3^8-6526 m3^6 mQ3^4 mU3^8+5636 m3^4 mQ3^6 mU3^8-1633 m3^2 mQ3^8 mU3^8-140 mQ3^10 mU3^8-180 m3^8 msq^2 mU3^8-360 m3^6 mQ3^2 msq^2 mU3^8+540 m3^4 mQ3^4 msq^2 mU3^8+270 m3^6 msq^4 mU3^8-180 m3^4 mQ3^2 msq^4 mU3^8-90 m3^2 mQ3^4 msq^4 mU3^8-61 m3^8 mU3^10+348 m3^6 mQ3^2 mU3^10-702 m3^4 mQ3^4 mU3^10+564 m3^2 mQ3^6 mU3^10-141 mQ3^8 mU3^10+60 m3^6 msq^2 mU3^10+120 m3^4 mQ3^2 msq^2 mU3^10-180 m3^2 mQ3^4 msq^2 mU3^10-90 m3^4 msq^4 mU3^10+60 m3^2 mQ3^2 msq^4 mU3^10+30 mQ3^4 msq^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(32 (18 m3^11 mQ3^4-149 m3^9 mQ3^6+264 m3^7 mQ3^8-132 m3^5 mQ3^10+24 m3^3 mQ3^12-9 m3 mQ3^14+60 m3^7 mQ3^6 msq^2-60 m3^5 mQ3^8 msq^2-30 m3^7 mQ3^4 msq^4+30 m3^5 mQ3^6 msq^4-32 m3^11 mQ3^2 mU3^2-182 m3^9 mQ3^4 mU3^2+753 m3^7 mQ3^6 mU3^2-823 m3^5 mQ3^8 mU3^2+231 m3^3 mQ3^10 mU3^2+21 m3 mQ3^12 mU3^2-120 m3^5 mQ3^6 msq^2 mU3^2+120 m3^3 mQ3^8 msq^2 mU3^2+60 m3^5 mQ3^4 msq^4 mU3^2-60 m3^3 mQ3^6 msq^4 mU3^2+14 m3^11 mU3^4-155 m3^9 mQ3^2 mU3^4+719 m3^7 mQ3^4 mU3^4-1267 m3^5 mQ3^6 mU3^4+879 m3^3 mQ3^8 mU3^4-190 m3 mQ3^10 mU3^4-60 m3^7 mQ3^2 msq^2 mU3^4+60 m3^5 mQ3^4 msq^2 mU3^4+60 m3^3 mQ3^6 msq^2 mU3^4-60 m3 mQ3^8 msq^2 mU3^4+30 m3^7 msq^4 mU3^4-30 m3^5 mQ3^2 msq^4 mU3^4-30 m3^3 mQ3^4 msq^4 mU3^4+30 m3 mQ3^6 msq^4 mU3^4+6 m3^9 mU3^6+207 m3^7 mQ3^2 mU3^6-681 m3^5 mQ3^4 mU3^6+761 m3^3 mQ3^6 mU3^6-261 m3 mQ3^8 mU3^6+120 m3^5 mQ3^2 msq^2 mU3^6-120 m3^3 mQ3^4 msq^2 mU3^6-60 m3^5 msq^4 mU3^6+60 m3^3 mQ3^2 msq^4 mU3^6-23 m3^7 mU3^8+23 m3^5 mQ3^2 mU3^8+25 m3^3 mQ3^4 mU3^8-41 m3 mQ3^6 mU3^8-60 m3^3 mQ3^2 msq^2 mU3^8+60 m3 mQ3^4 msq^2 mU3^8+30 m3^3 msq^4 mU3^8-30 m3 mQ3^2 msq^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^5)-(1280 m3^2 (mQ3^2+mU3^2) (m3^2 mQ3^4+2 m3^2 mQ3^2 mU3^2-3 mQ3^4 mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^5)) Log[mU3^2/M^2])+((8 (128 m3^14-334 m3^12 mQ3^2+368 m3^10 mQ3^4-89 m3^8 mQ3^6+11 m3^6 mQ3^8-29 m3^4 mQ3^10+9 m3^2 mQ3^12-434 m3^12 mU3^2+1072 m3^10 mQ3^2 mU3^2-1145 m3^8 mQ3^4 mU3^2+100 m3^6 mQ3^6 mU3^2+192 m3^4 mQ3^8 mU3^2-44 m3^2 mQ3^10 mU3^2+3 mQ3^12 mU3^2+480 m3^10 mU3^4-1177 m3^8 mQ3^2 mU3^4+1498 m3^6 mQ3^4 mU3^4-376 m3^4 mQ3^6 mU3^4-42 m3^2 mQ3^8 mU3^4+mQ3^10 mU3^4-149 m3^8 mU3^6+324 m3^6 mQ3^2 mU3^6-598 m3^4 mQ3^4 mU3^6+148 m3^2 mQ3^6 mU3^6+19 mQ3^8 mU3^6-13 m3^6 mU3^8+43 m3^4 mQ3^2 mU3^8+57 m3^2 mQ3^4 mU3^8-23 mQ3^6 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+((256 m3^3 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+(64 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt+(128 m3 (2 m3^2-mQ3^2-mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(16 (172 m3^12-372 m3^10 mQ3^2+138 m3^8 mQ3^4+3 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-468 m3^10 mU3^2+1084 m3^8 mQ3^2 mU3^2-313 m3^6 mQ3^4 mU3^2-149 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+398 m3^8 mU3^4-1151 m3^6 mQ3^2 mU3^4+411 m3^4 mQ3^4 mU3^4+55 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-59 m3^6 mU3^6+409 m3^4 mQ3^2 mU3^6-129 m3^2 mQ3^4 mU3^6-29 mQ3^6 mU3^6-31 m3^4 mU3^8-30 m3^2 mQ3^2 mU3^8+13 mQ3^4 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(128 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4) Xt^5)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+((80 (m3^2-msq^2) (2 m3^10-8 m3^8 mQ3^2+3 m3^6 mQ3^4-m3^4 mQ3^6+6 m3^8 msq^2-8 m3^6 mQ3^2 msq^2+9 m3^4 mQ3^4 msq^2-3 m3^2 mQ3^6 msq^2-8 m3^8 mU3^2+30 m3^6 mQ3^2 mU3^2-15 m3^4 mQ3^4 mU3^2+5 m3^2 mQ3^6 mU3^2-8 m3^6 msq^2 mU3^2-6 m3^4 mQ3^2 msq^2 mU3^2+3 m3^2 mQ3^4 msq^2 mU3^2-mQ3^6 msq^2 mU3^2+3 m3^6 mU3^4-15 m3^4 mQ3^2 mU3^4+9 m3^4 msq^2 mU3^4+3 m3^2 mQ3^2 msq^2 mU3^4-m3^4 mU3^6+5 m3^2 mQ3^2 mU3^6-3 m3^2 msq^2 mU3^6-mQ3^2 msq^2 mU3^6))/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(160 (m3^2-msq^2) (2 m3^10-8 m3^8 mQ3^2+3 m3^6 mQ3^4-m3^4 mQ3^6+6 m3^8 msq^2-8 m3^6 mQ3^2 msq^2+9 m3^4 mQ3^4 msq^2-3 m3^2 mQ3^6 msq^2-8 m3^8 mU3^2+30 m3^6 mQ3^2 mU3^2-15 m3^4 mQ3^4 mU3^2+5 m3^2 mQ3^6 mU3^2-8 m3^6 msq^2 mU3^2-6 m3^4 mQ3^2 msq^2 mU3^2+3 m3^2 mQ3^4 msq^2 mU3^2-mQ3^6 msq^2 mU3^2+3 m3^6 mU3^4-15 m3^4 mQ3^2 mU3^4+9 m3^4 msq^2 mU3^4+3 m3^2 mQ3^2 msq^2 mU3^4-m3^4 mU3^6+5 m3^2 mQ3^2 mU3^6-3 m3^2 msq^2 mU3^6-mQ3^2 msq^2 mU3^6) Xt^4)/((m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-m3^2/msq^2]+(-((80 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2))/(m3^2-mQ3^2)^3)+(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(160 (mQ3^2-msq^2) (2 m3^4+3 m3^2 mQ3^2-mQ3^4-3 m3^2 msq^2-mQ3^2 msq^2) Xt^4)/((m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^2)-(1280 m3 (mQ3^2-msq^2)^2 Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-msq^2/mQ3^2]+(-((8 (128 m3^14-434 m3^12 mQ3^2+480 m3^10 mQ3^4-149 m3^8 mQ3^6-13 m3^6 mQ3^8-334 m3^12 mU3^2+1072 m3^10 mQ3^2 mU3^2-1177 m3^8 mQ3^4 mU3^2+324 m3^6 mQ3^6 mU3^2+43 m3^4 mQ3^8 mU3^2+368 m3^10 mU3^4-1145 m3^8 mQ3^2 mU3^4+1498 m3^6 mQ3^4 mU3^4-598 m3^4 mQ3^6 mU3^4+57 m3^2 mQ3^8 mU3^4-89 m3^8 mU3^6+100 m3^6 mQ3^2 mU3^6-376 m3^4 mQ3^4 mU3^6+148 m3^2 mQ3^6 mU3^6-23 mQ3^8 mU3^6+11 m3^6 mU3^8+192 m3^4 mQ3^2 mU3^8-42 m3^2 mQ3^4 mU3^8+19 mQ3^6 mU3^8-29 m3^4 mU3^10-44 m3^2 mQ3^2 mU3^10+mQ3^4 mU3^10+9 m3^2 mU3^12+3 mQ3^2 mU3^12))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+((256 m3^3 (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(64 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) Xt+(128 m3 (-2 m3^2+mQ3^2+mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2) Xt^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+(16 (172 m3^12-468 m3^10 mQ3^2+398 m3^8 mQ3^4-59 m3^6 mQ3^6-31 m3^4 mQ3^8-372 m3^10 mU3^2+1084 m3^8 mQ3^2 mU3^2-1151 m3^6 mQ3^4 mU3^2+409 m3^4 mQ3^6 mU3^2-30 m3^2 mQ3^8 mU3^2+138 m3^8 mU3^4-313 m3^6 mQ3^2 mU3^4+411 m3^4 mQ3^4 mU3^4-129 m3^2 mQ3^6 mU3^4+13 mQ3^8 mU3^4+3 m3^6 mU3^6-149 m3^4 mQ3^2 mU3^6+55 m3^2 mQ3^4 mU3^6-29 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(128 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2]+Log[msq^2/M^2] ((16 (9 m3^10 mQ3^2-18 m3^8 mQ3^4+9 m3^6 mQ3^6+9 m3^10 mU3^2-114 m3^8 mQ3^2 mU3^2+196 m3^6 mQ3^4 mU3^2-91 m3^4 mQ3^6 mU3^2-90 m3^6 mQ3^2 msq^2 mU3^2+75 m3^4 mQ3^4 msq^2 mU3^2-45 m3^2 mQ3^6 msq^2 mU3^2-18 m3^8 mU3^4+196 m3^6 mQ3^2 mU3^4-328 m3^4 mQ3^4 mU3^4+150 m3^2 mQ3^6 mU3^4+75 m3^4 mQ3^2 msq^2 mU3^4+60 m3^2 mQ3^4 msq^2 mU3^4-15 mQ3^6 msq^2 mU3^4+9 m3^6 mU3^6-91 m3^4 mQ3^2 mU3^6+150 m3^2 mQ3^4 mU3^6-68 mQ3^6 mU3^6-45 m3^2 mQ3^2 msq^2 mU3^6-15 mQ3^4 msq^2 mU3^6))/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+(640 m3 msq^2 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))-(96 m3^2 Xt^2)/(mQ3^2 mU3^2)-(768 m3 Xt^3)/(mQ3^2-mU3^2)^2+(16 (9 m3^10 mQ3^2-18 m3^8 mQ3^4+9 m3^6 mQ3^6+9 m3^10 mU3^2+138 m3^8 mQ3^2 mU3^2-293 m3^6 mQ3^4 mU3^2+146 m3^4 mQ3^6 mU3^2+180 m3^6 mQ3^2 msq^2 mU3^2-150 m3^4 mQ3^4 msq^2 mU3^2+90 m3^2 mQ3^6 msq^2 mU3^2-18 m3^8 mU3^4-293 m3^6 mQ3^2 mU3^4+620 m3^4 mQ3^4 mU3^4-309 m3^2 mQ3^6 mU3^4-150 m3^4 mQ3^2 msq^2 mU3^4-120 m3^2 mQ3^4 msq^2 mU3^4+30 mQ3^6 msq^2 mU3^4+9 m3^6 mU3^6+146 m3^4 mQ3^2 mU3^6-309 m3^2 mQ3^4 mU3^6+154 mQ3^6 mU3^6+90 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6) Xt^4)/(3 mQ3^2 (-m3^2+mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(1280 m3 msq^2 Xt^5)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)+(-((96 m3^4)/(-m3^2+mQ3^2)^2)+(384 m3^3 Xt)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(192 m3 (2 m3^2-mQ3^2-mU3^2) Xt^3)/(mQ3^2-mU3^2)^3-(48 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mQ3^2]+(-((96 m3^4)/(m3^2-mU3^2)^2)+(384 m3^3 Xt)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(192 m3 (-2 m3^2+mQ3^2+mU3^2) Xt^3)/(mQ3^2-mU3^2)^3+(48 (-2 m3^2+mQ3^2+mU3^2) Xt^4)/(mQ3^2-mU3^2)^3) PolyLog[2,1-m3^2/mU3^2])+Log[m3^2/M^2] ((8 (132 m3^14 mQ3^2-300 m3^12 mQ3^4+236 m3^10 mQ3^6-68 m3^8 mQ3^8+132 m3^14 mU3^2-1336 m3^12 mQ3^2 mU3^2+2738 m3^10 mQ3^4 mU3^2-2016 m3^8 mQ3^6 mU3^2+519 m3^6 mQ3^8 mU3^2-9 m3^4 mQ3^10 mU3^2+180 m3^10 mQ3^2 msq^2 mU3^2-240 m3^8 mQ3^4 msq^2 mU3^2+270 m3^6 mQ3^6 msq^2 mU3^2-90 m3^4 mQ3^8 msq^2 mU3^2-300 m3^12 mU3^4+2738 m3^10 mQ3^2 mU3^4-5432 m3^8 mQ3^4 mU3^4+3785 m3^6 mQ3^6 mU3^4-860 m3^4 mQ3^8 mU3^4-3 m3^2 mQ3^10 mU3^4-240 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4+90 m3^4 mQ3^6 msq^2 mU3^4-30 m3^2 mQ3^8 msq^2 mU3^4+236 m3^10 mU3^6-2016 m3^8 mQ3^2 mU3^6+3785 m3^6 mQ3^4 mU3^6-2470 m3^4 mQ3^6 mU3^6+513 m3^2 mQ3^8 mU3^6+270 m3^6 mQ3^2 msq^2 mU3^6+90 m3^4 mQ3^4 msq^2 mU3^6-68 m3^8 mU3^8+519 m3^6 mQ3^2 mU3^8-860 m3^4 mQ3^4 mU3^8+513 m3^2 mQ3^6 mU3^8-96 mQ3^8 mU3^8-90 m3^4 mQ3^2 msq^2 mU3^8-30 m3^2 mQ3^4 msq^2 mU3^8-9 m3^4 mQ3^2 mU3^10-3 m3^2 mQ3^4 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3)-(64 (16 m3^9 mQ3^2-16 m3^7 mQ3^4+16 m3^9 mU3^2-128 m3^7 mQ3^2 mU3^2+112 m3^5 mQ3^4 mU3^2-3 m3^3 mQ3^6 mU3^2+60 m3^5 mQ3^2 msq^2 mU3^2-30 m3^3 mQ3^4 msq^2 mU3^2-16 m3^7 mU3^4+112 m3^5 mQ3^2 mU3^4-90 m3^3 mQ3^4 mU3^4-30 m3^3 mQ3^2 msq^2 mU3^4-3 m3^3 mQ3^2 mU3^6) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)-(64 (33 m3^10-42 m3^8 mQ3^2+17 m3^6 mQ3^4-42 m3^8 mU3^2+25 m3^6 mQ3^2 mU3^2-7 m3^4 mQ3^4 mU3^2+17 m3^6 mU3^4-7 m3^4 mQ3^2 mU3^4+6 m3^2 mQ3^4 mU3^4) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2)+((2048 m3^5)/(3 mQ3^2 (m3^2-mQ3^2) mU3^2 (m3^2-mU3^2))-(512 m3 (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(mQ3^2-mU3^2)^2) Xt^3+(16 (66 m3^14 mQ3^2-150 m3^12 mQ3^4+118 m3^10 mQ3^6-34 m3^8 mQ3^8+66 m3^14 mU3^2+1876 m3^12 mQ3^2 mU3^2-4190 m3^10 mQ3^4 mU3^2+2756 m3^8 mQ3^6 mU3^2-465 m3^6 mQ3^8 mU3^2+9 m3^4 mQ3^10 mU3^2-180 m3^10 mQ3^2 msq^2 mU3^2+240 m3^8 mQ3^4 msq^2 mU3^2-270 m3^6 mQ3^6 msq^2 mU3^2+90 m3^4 mQ3^8 msq^2 mU3^2-150 m3^12 mU3^4-4190 m3^10 mQ3^2 mU3^4+8964 m3^8 mQ3^4 mU3^4-5645 m3^6 mQ3^6 mU3^4+850 m3^4 mQ3^8 mU3^4+3 m3^2 mQ3^10 mU3^4+240 m3^8 mQ3^2 msq^2 mU3^4+180 m3^6 mQ3^4 msq^2 mU3^4-90 m3^4 mQ3^6 msq^2 mU3^4+30 m3^2 mQ3^8 msq^2 mU3^4+118 m3^10 mU3^6+2756 m3^8 mQ3^2 mU3^6-5645 m3^6 mQ3^4 mU3^6+3522 m3^4 mQ3^6 mU3^6-559 m3^2 mQ3^8 mU3^6-270 m3^6 mQ3^2 msq^2 mU3^6-90 m3^4 mQ3^4 msq^2 mU3^6-34 m3^8 mU3^8-465 m3^6 mQ3^2 mU3^8+850 m3^4 mQ3^4 mU3^8-559 m3^2 mQ3^6 mU3^8+120 mQ3^8 mU3^8+90 m3^4 mQ3^2 msq^2 mU3^8+30 m3^2 mQ3^4 msq^2 mU3^8+9 m3^4 mQ3^2 mU3^10+3 m3^2 mQ3^4 mU3^10) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(128 (8 m3^9 mQ3^2-8 m3^7 mQ3^4+8 m3^9 mU3^2+123 m3^7 mQ3^2 mU3^2-115 m3^5 mQ3^4 mU3^2+3 m3^3 mQ3^6 mU3^2-60 m3^5 mQ3^2 msq^2 mU3^2+30 m3^3 mQ3^4 msq^2 mU3^2-8 m3^7 mU3^4-115 m3^5 mQ3^2 mU3^4+101 m3^3 mQ3^4 mU3^4+30 m3^3 mQ3^2 msq^2 mU3^4+3 m3^3 mQ3^2 mU3^6) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+((16 (64 m3^16-244 m3^14 mQ3^2+238 m3^12 mQ3^4-169 m3^10 mQ3^6+110 m3^8 mQ3^8-27 m3^6 mQ3^10-140 m3^14 mU3^2+632 m3^12 mQ3^2 mU3^2-427 m3^10 mQ3^4 mU3^2+159 m3^8 mQ3^6 mU3^2-157 m3^6 mQ3^8 mU3^2+45 m3^4 mQ3^10 mU3^2+90 m3^12 mU3^4-703 m3^10 mQ3^2 mU3^4+397 m3^8 mQ3^4 mU3^4-17 m3^6 mQ3^6 mU3^4+101 m3^4 mQ3^8 mU3^4-36 m3^2 mQ3^10 mU3^4+19 m3^10 mU3^6+321 m3^8 mQ3^2 mU3^6-147 m3^6 mQ3^4 mU3^6-81 m3^4 mQ3^6 mU3^6-12 m3^2 mQ3^8 mU3^6+12 mQ3^10 mU3^6-27 m3^8 mU3^8-36 m3^6 mQ3^2 mU3^8-m3^4 mQ3^4 mU3^8+48 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(32 (16 m3^11+199 m3^9 mQ3^2-132 m3^7 mQ3^4+13 m3^5 mQ3^6-15 m3^9 mU3^2-433 m3^7 mQ3^2 mU3^2+267 m3^5 mQ3^4 mU3^2-11 m3^3 mQ3^6 mU3^2+25 m3^7 mU3^4+172 m3^5 mQ3^2 mU3^4-89 m3^3 mQ3^4 mU3^4-12 m3 mQ3^6 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(32 (64 m3^16-502 m3^14 mQ3^2+954 m3^12 mQ3^4-759 m3^10 mQ3^6+302 m3^8 mQ3^8-55 m3^6 mQ3^10-54 m3^14 mU3^2+962 m3^12 mQ3^2 mU3^2-1897 m3^10 mQ3^4 mU3^2+1401 m3^8 mQ3^6 mU3^2-525 m3^6 mQ3^8 mU3^2+97 m3^4 mQ3^10 mU3^2-72 m3^12 mU3^4-545 m3^10 mQ3^2 mU3^4+1199 m3^8 mQ3^4 mU3^4-811 m3^6 mQ3^6 mU3^4+325 m3^4 mQ3^8 mU3^4-72 m3^2 mQ3^10 mU3^4+69 m3^10 mU3^6+79 m3^8 mQ3^2 mU3^6-189 m3^6 mQ3^4 mU3^6+61 m3^4 mQ3^6 mU3^6-60 m3^2 mQ3^8 mU3^6+24 mQ3^10 mU3^6-m3^8 mU3^8-24 m3^6 mQ3^2 mU3^8-7 m3^4 mQ3^4 mU3^8+48 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(64 (40 m3^13-114 m3^11 mQ3^2+233 m3^9 mQ3^4-250 m3^7 mQ3^6+59 m3^5 mQ3^8-2 m3^11 mU3^2-56 m3^9 mQ3^2 mU3^2-55 m3^7 mQ3^4 mU3^2+264 m3^5 mQ3^6 mU3^2-55 m3^3 mQ3^8 mU3^2-69 m3^9 mU3^4+220 m3^7 mQ3^2 mU3^4-201 m3^5 mQ3^4 mU3^4-40 m3^3 mQ3^6 mU3^4-6 m3 mQ3^8 mU3^4+45 m3^7 mU3^6-90 m3^5 mQ3^2 mU3^6+59 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(16 (788 m3^16 mQ3^2-2556 m3^14 mQ3^4+2618 m3^12 mQ3^6-835 m3^10 mQ3^8+58 m3^8 mQ3^10-69 m3^6 mQ3^12-148 m3^16 mU3^2-1796 m3^14 mQ3^2 mU3^2+7366 m3^12 mQ3^4 mU3^2-8472 m3^10 mQ3^6 mU3^2+3031 m3^8 mQ3^8 mU3^2-116 m3^6 mQ3^10 mU3^2+123 m3^4 mQ3^12 mU3^2+336 m3^14 mU3^4+786 m3^12 mQ3^2 mU3^4-6232 m3^10 mQ3^4 mU3^4+8712 m3^8 mQ3^6 mU3^4-3742 m3^6 mQ3^8 mU3^4+238 m3^4 mQ3^10 mU3^4-90 m3^2 mQ3^12 mU3^4-194 m3^12 mU3^6+452 m3^10 mQ3^2 mU3^6+936 m3^8 mQ3^4 mU3^6-2648 m3^6 mQ3^6 mU3^6+1612 m3^4 mQ3^8 mU3^6-180 m3^2 mQ3^10 mU3^6+30 mQ3^12 mU3^6-m3^10 mU3^8-190 m3^8 mQ3^2 mU3^8+479 m3^6 mQ3^4 mU3^8-234 m3^4 mQ3^6 mU3^8-90 m3^2 mQ3^8 mU3^8+24 mQ3^10 mU3^8+13 m3^8 mU3^10-64 m3^6 mQ3^2 mU3^10+37 m3^4 mQ3^4 mU3^10+24 m3^2 mQ3^6 mU3^10-6 mQ3^8 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(32 (36 m3^11 mQ3^2-159 m3^9 mQ3^4+216 m3^7 mQ3^6-61 m3^5 mQ3^8+100 m3^11 mU3^2-392 m3^9 mQ3^2 mU3^2+585 m3^7 mQ3^4 mU3^2-384 m3^5 mQ3^6 mU3^2+59 m3^3 mQ3^8 mU3^2-169 m3^9 mU3^4+508 m3^7 mQ3^2 mU3^4-535 m3^5 mQ3^4 mU3^4+152 m3^3 mQ3^6 mU3^4+12 m3 mQ3^8 mU3^4+75 m3^7 mU3^6-148 m3^5 mQ3^2 mU3^6+93 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)-(2560 m3^4 mQ3^2 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^4)) Log[mQ3^2/M^2]^2+(-((16 (9 m3^14 mQ3^2-27 m3^12 mQ3^4+27 m3^10 mQ3^6-9 m3^8 mQ3^8+9 m3^14 mU3^2+20 m3^12 mQ3^2 mU3^2-109 m3^10 mQ3^4 mU3^2+30 m3^8 mQ3^6 mU3^2-10 m3^6 mQ3^8 mU3^2-27 m3^12 mU3^4-109 m3^10 mQ3^2 mU3^4+474 m3^8 mQ3^4 mU3^4-237 m3^6 mQ3^6 mU3^4+79 m3^4 mQ3^8 mU3^4+27 m3^10 mU3^6+30 m3^8 mQ3^2 mU3^6-237 m3^6 mQ3^4 mU3^6-9 m3^8 mU3^8-10 m3^6 mQ3^2 mU3^8+79 m3^4 mQ3^4 mU3^8))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3))-(32 (m3^7+59 m3^5 mQ3^2+59 m3^5 mU3^2-119 m3^3 mQ3^2 mU3^2) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(96 m3^4 (-m3^2+mQ3^2+mU3^2) Xt^2)/(mQ3^2 (-m3^2+mQ3^2) mU3^2 (m3^2-mU3^2))-(16 (9 m3^14 mQ3^2-27 m3^12 mQ3^4+27 m3^10 mQ3^6-9 m3^8 mQ3^8+9 m3^14 mU3^2-94 m3^12 mQ3^2 mU3^2+308 m3^10 mQ3^4 mU3^2-150 m3^8 mQ3^6 mU3^2+47 m3^6 mQ3^8 mU3^2-27 m3^12 mU3^4+308 m3^10 mQ3^2 mU3^4-1002 m3^8 mQ3^4 mU3^4+528 m3^6 mQ3^6 mU3^4-167 m3^4 mQ3^8 mU3^4+27 m3^10 mU3^6-150 m3^8 mQ3^2 mU3^6+528 m3^6 mQ3^4 mU3^6-54 m3^4 mQ3^6 mU3^6+9 m3^2 mQ3^8 mU3^6-9 m3^8 mU3^8+47 m3^6 mQ3^2 mU3^8-167 m3^4 mQ3^4 mU3^8+9 m3^2 mQ3^6 mU3^8) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(640 (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Log[msq^2/M^2]+(-((8 (2 m3^16 mQ3^4-6 m3^14 mQ3^6+6 m3^12 mQ3^8-2 m3^10 mQ3^10-128 m3^16 mQ3^2 mU3^2-760 m3^14 mQ3^4 mU3^2+2188 m3^12 mQ3^6 mU3^2-1902 m3^10 mQ3^8 mU3^2+547 m3^8 mQ3^10 mU3^2-9 m3^6 mQ3^12 mU3^2+180 m3^12 mQ3^4 msq^2 mU3^2-240 m3^10 mQ3^6 msq^2 mU3^2+270 m3^8 mQ3^8 msq^2 mU3^2-90 m3^6 mQ3^10 msq^2 mU3^2-2 m3^16 mU3^4+1462 m3^14 mQ3^2 mU3^4-610 m3^12 mQ3^4 mU3^4-3080 m3^10 mQ3^6 mU3^4+3528 m3^8 mQ3^8 mU3^4-972 m3^6 mQ3^10 mU3^4+6 m3^4 mQ3^12 mU3^4-180 m3^12 mQ3^2 msq^2 mU3^4-180 m3^10 mQ3^4 msq^2 mU3^4-210 m3^8 mQ3^6 msq^2 mU3^4-90 m3^6 mQ3^8 msq^2 mU3^4+60 m3^4 mQ3^10 msq^2 mU3^4+72 m3^14 mU3^6-3332 m3^12 mQ3^2 mU3^6+3634 m3^10 mQ3^4 mU3^6+1602 m3^8 mQ3^6 mU3^6-3698 m3^6 mQ3^8 mU3^6+1019 m3^4 mQ3^10 mU3^6+3 m3^2 mQ3^12 mU3^6+420 m3^10 mQ3^2 msq^2 mU3^6+450 m3^8 mQ3^4 msq^2 mU3^6+450 m3^6 mQ3^6 msq^2 mU3^6-150 m3^4 mQ3^8 msq^2 mU3^6+30 m3^2 mQ3^10 msq^2 mU3^6-172 m3^12 mU3^8+3774 m3^10 mQ3^2 mU3^8-5506 m3^8 mQ3^4 mU3^8+1682 m3^6 mQ3^6 mU3^8+1522 m3^4 mQ3^8 mU3^8-540 m3^2 mQ3^10 mU3^8-510 m3^8 mQ3^2 msq^2 mU3^8-630 m3^6 mQ3^4 msq^2 mU3^8-30 m3^4 mQ3^6 msq^2 mU3^8-30 m3^2 mQ3^8 msq^2 mU3^8+136 m3^10 mU3^10-2057 m3^8 mQ3^2 mU3^10+3339 m3^6 mQ3^4 mU3^10-2026 m3^4 mQ3^6 mU3^10+84 m3^2 mQ3^8 mU3^10+84 mQ3^10 mU3^10+360 m3^6 mQ3^2 msq^2 mU3^10+210 m3^4 mQ3^4 msq^2 mU3^10+30 m3^2 mQ3^6 msq^2 mU3^10-34 m3^8 mU3^12+426 m3^6 mQ3^2 mU3^12-640 m3^4 mQ3^4 mU3^12+456 m3^2 mQ3^6 mU3^12-84 mQ3^8 mU3^12-90 m3^4 mQ3^2 msq^2 mU3^12-30 m3^2 mQ3^4 msq^2 mU3^12-9 m3^4 mQ3^2 mU3^14-3 m3^2 mQ3^4 mU3^14))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))+(64 (4 m3^11 mQ3^2-130 m3^9 mQ3^4+123 m3^7 mQ3^6-3 m3^5 mQ3^8+60 m3^7 mQ3^4 msq^2-30 m3^5 mQ3^6 msq^2+16 m3^11 mU3^2+216 m3^9 mQ3^2 mU3^2-106 m3^7 mQ3^4 mU3^2-68 m3^5 mQ3^6 mU3^2+3 m3^3 mQ3^8 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2-32 m3^9 mU3^4-223 m3^7 mQ3^2 mU3^4+111 m3^5 mQ3^4 mU3^4+37 m3^3 mQ3^6 mU3^4+90 m3^5 mQ3^2 msq^2 mU3^4+16 m3^7 mU3^6+86 m3^5 mQ3^2 mU3^6-35 m3^3 mQ3^4 mU3^6-12 m3 mQ3^6 mU3^6-30 m3^3 mQ3^2 msq^2 mU3^6-3 m3^3 mQ3^2 mU3^8) Xt)/(3 mQ3^2 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(16 (2 m3^14 mQ3^2-6 m3^12 mQ3^4+6 m3^10 mQ3^6-2 m3^8 mQ3^8-2 m3^14 mU3^2-1472 m3^12 mQ3^2 mU3^2+3280 m3^10 mQ3^4 mU3^2-2436 m3^8 mQ3^6 mU3^2+587 m3^6 mQ3^8 mU3^2-9 m3^4 mQ3^10 mU3^2+180 m3^10 mQ3^2 msq^2 mU3^2-240 m3^8 mQ3^4 msq^2 mU3^2+270 m3^6 mQ3^6 msq^2 mU3^2-90 m3^4 mQ3^8 msq^2 mU3^2+70 m3^12 mU3^4+2660 m3^10 mQ3^2 mU3^4-5696 m3^8 mQ3^4 mU3^4+3987 m3^6 mQ3^6 mU3^4-850 m3^4 mQ3^8 mU3^4-3 m3^2 mQ3^10 mU3^4-240 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4+90 m3^4 mQ3^6 msq^2 mU3^4-30 m3^2 mQ3^8 msq^2 mU3^4-102 m3^10 mU3^6-1692 m3^8 mQ3^2 mU3^6+3519 m3^6 mQ3^4 mU3^6-2430 m3^4 mQ3^6 mU3^6+513 m3^2 mQ3^8 mU3^6+270 m3^6 mQ3^2 msq^2 mU3^6+90 m3^4 mQ3^4 msq^2 mU3^6+34 m3^8 mU3^8+395 m3^6 mQ3^2 mU3^8-734 m3^4 mQ3^4 mU3^8+489 m3^2 mQ3^6 mU3^8-96 mQ3^8 mU3^8-90 m3^4 mQ3^2 msq^2 mU3^8-30 m3^2 mQ3^4 msq^2 mU3^8-9 m3^4 mQ3^2 mU3^10-3 m3^2 mQ3^4 mU3^10) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 (22 m3^11 mQ3^2+10 m3^9 mQ3^4-193 m3^7 mQ3^6+153 m3^5 mQ3^8-3 m3^3 mQ3^10+60 m3^5 mQ3^6 msq^2-30 m3^3 mQ3^8 msq^2+6 m3^9 mQ3^2 mU3^2+238 m3^7 mQ3^4 mU3^2-85 m3^5 mQ3^6 mU3^2-91 m3^3 mQ3^8 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2+16 m3^9 mU3^4-217 m3^7 mQ3^2 mU3^4-17 m3^5 mQ3^4 mU3^4+106 m3^3 mQ3^6 mU3^4+6 m3 mQ3^8 mU3^4+60 m3^5 mQ3^2 msq^2 mU3^4+30 m3^3 mQ3^4 msq^2 mU3^4-16 m3^7 mU3^6+165 m3^5 mQ3^2 mU3^6-115 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6-30 m3^3 mQ3^2 msq^2 mU3^6-3 m3^3 mQ3^2 mU3^8) Xt^3)/(3 mQ3^2 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(8 (2 m3^16 mQ3^4-6 m3^14 mQ3^6+6 m3^12 mQ3^8-2 m3^10 mQ3^10+80 m3^16 mQ3^2 mU3^2-2152 m3^14 mQ3^4 mU3^2+4406 m3^12 mQ3^6 mU3^2-2996 m3^10 mQ3^8 mU3^2+643 m3^8 mQ3^10 mU3^2-9 m3^6 mQ3^12 mU3^2+180 m3^12 mQ3^4 msq^2 mU3^2-240 m3^10 mQ3^6 msq^2 mU3^2+270 m3^8 mQ3^8 msq^2 mU3^2-90 m3^6 mQ3^10 msq^2 mU3^2-2 m3^16 mU3^4-3178 m3^14 mQ3^2 mU3^4+12962 m3^12 mQ3^4 mU3^4-17588 m3^10 mQ3^6 mU3^4+9676 m3^8 mQ3^8 mU3^4-1828 m3^6 mQ3^10 mU3^4+6 m3^4 mQ3^12 mU3^4+180 m3^12 mQ3^2 msq^2 mU3^4-660 m3^10 mQ3^4 msq^2 mU3^4+330 m3^8 mQ3^6 msq^2 mU3^4-270 m3^6 mQ3^8 msq^2 mU3^4+60 m3^4 mQ3^10 msq^2 mU3^4+72 m3^14 mU3^6+9110 m3^12 mQ3^2 mU3^6-25584 m3^10 mQ3^4 mU3^6+25638 m3^8 mQ3^6 mU3^6-10862 m3^6 mQ3^8 mU3^6+1723 m3^4 mQ3^10 mU3^6+3 m3^2 mQ3^12 mU3^6-420 m3^10 mQ3^2 msq^2 mU3^6+570 m3^8 mQ3^4 msq^2 mU3^6+90 m3^6 mQ3^6 msq^2 mU3^6-30 m3^4 mQ3^8 msq^2 mU3^6+30 m3^2 mQ3^10 msq^2 mU3^6-172 m3^12 mU3^8-9926 m3^10 mQ3^2 mU3^8+22990 m3^8 mQ3^4 mU3^8-18070 m3^6 mQ3^6 mU3^8+5474 m3^4 mQ3^8 mU3^8-616 m3^2 mQ3^10 mU3^8+510 m3^8 mQ3^2 msq^2 mU3^8-90 m3^6 mQ3^4 msq^2 mU3^8-210 m3^4 mQ3^6 msq^2 mU3^8+30 m3^2 mQ3^8 msq^2 mU3^8+136 m3^10 mU3^10+4623 m3^8 mQ3^2 mU3^10-9553 m3^6 mQ3^4 mU3^10+6600 m3^4 mQ3^6 mU3^10-1602 m3^2 mQ3^8 mU3^10+96 mQ3^10 mU3^10-360 m3^6 mQ3^2 msq^2 mU3^10-30 m3^4 mQ3^4 msq^2 mU3^10+30 m3^2 mQ3^6 msq^2 mU3^10-34 m3^8 mU3^12-702 m3^6 mQ3^2 mU3^12+1268 m3^4 mQ3^4 mU3^12-884 m3^2 mQ3^6 mU3^12+240 mQ3^8 mU3^12+90 m3^4 mQ3^2 msq^2 mU3^12+30 m3^2 mQ3^4 msq^2 mU3^12+9 m3^4 mQ3^2 mU3^14+3 m3^2 mQ3^4 mU3^14) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^3)+(64 (52 m3^11 mQ3^2-276 m3^9 mQ3^4+214 m3^7 mQ3^6-3 m3^5 mQ3^8+60 m3^7 mQ3^4 msq^2-30 m3^5 mQ3^6 msq^2+16 m3^11 mU3^2-362 m3^9 mQ3^2 mU3^2+770 m3^7 mQ3^4 mU3^2-407 m3^5 mQ3^6 mU3^2+3 m3^3 mQ3^8 mU3^2+60 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2-32 m3^9 mU3^4+558 m3^7 mQ3^2 mU3^4-737 m3^5 mQ3^4 mU3^4+213 m3^3 mQ3^6 mU3^4-90 m3^5 mQ3^2 msq^2 mU3^4+60 m3^3 mQ3^4 msq^2 mU3^4+16 m3^7 mU3^6-219 m3^5 mQ3^2 mU3^6+179 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6+3 m3^3 mQ3^2 mU3^8) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(5120 m3^4 mU3^2 Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+((16 (29 m3^12-127 m3^10 mQ3^2+57 m3^8 mQ3^4-19 m3^6 mQ3^6-109 m3^10 mU3^2+447 m3^8 mQ3^2 mU3^2-237 m3^6 mQ3^4 mU3^2+79 m3^4 mQ3^6 mU3^2+30 m3^8 mU3^4-210 m3^6 mQ3^2 mU3^4-10 m3^6 mU3^6+70 m3^4 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(64 (9 m3^9-13 m3^7 mQ3^2+34 m3^5 mQ3^4-14 m3^7 mU3^2+18 m3^5 mQ3^2 mU3^2-64 m3^3 mQ3^4 mU3^2-25 m3^5 mU3^4+55 m3^3 mQ3^2 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(320 (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(-((64 (m3^7+59 m3^5 mQ3^2+59 m3^5 mU3^2-119 m3^3 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)))+(192 m3^3 (2 m3^4-mQ3^4+4 mQ3^2 mU3^2-mU3^4-2 m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^3+(160 (mQ3^2+mU3^2) (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(320 (mQ3^2+mU3^2) (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[msq^2/M^2]) Log[mU3^2/M^2]+(-((16 (64 m3^16-139 m3^14 mQ3^2+87 m3^12 mQ3^4+22 m3^10 mQ3^6-28 m3^8 mQ3^8-245 m3^14 mU3^2+633 m3^12 mQ3^2 mU3^2-700 m3^10 mQ3^4 mU3^2+316 m3^8 mQ3^6 mU3^2-34 m3^6 mQ3^8 mU3^2+240 m3^12 mU3^4-432 m3^10 mQ3^2 mU3^4+400 m3^8 mQ3^4 mU3^4-146 m3^6 mQ3^6 mU3^4-2 m3^4 mQ3^8 mU3^4-170 m3^10 mU3^6+162 m3^8 mQ3^2 mU3^6-20 m3^6 mQ3^4 mU3^6-80 m3^4 mQ3^6 mU3^6+48 m3^2 mQ3^8 mU3^6+110 m3^8 mU3^8-157 m3^6 mQ3^2 mU3^8+101 m3^4 mQ3^4 mU3^8-12 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8-27 m3^6 mU3^10+45 m3^4 mQ3^2 mU3^10-36 m3^2 mQ3^4 mU3^10+12 mQ3^6 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))-(32 (18 m3^11-19 m3^9 mQ3^2+27 m3^7 mQ3^4+195 m3^9 mU3^2-425 m3^7 mQ3^2 mU3^2+168 m3^5 mQ3^4 mU3^2-130 m3^7 mU3^4+263 m3^5 mQ3^2 mU3^4-87 m3^3 mQ3^4 mU3^4+13 m3^5 mU3^6-11 m3^3 mQ3^2 mU3^6-12 m3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(32 (64 m3^16-54 m3^14 mQ3^2-72 m3^12 mQ3^4+69 m3^10 mQ3^6-m3^8 mQ3^8-502 m3^14 mU3^2+962 m3^12 mQ3^2 mU3^2-545 m3^10 mQ3^4 mU3^2+79 m3^8 mQ3^6 mU3^2-24 m3^6 mQ3^8 mU3^2+954 m3^12 mU3^4-1897 m3^10 mQ3^2 mU3^4+1199 m3^8 mQ3^4 mU3^4-189 m3^6 mQ3^6 mU3^4-7 m3^4 mQ3^8 mU3^4-759 m3^10 mU3^6+1401 m3^8 mQ3^2 mU3^6-811 m3^6 mQ3^4 mU3^6+61 m3^4 mQ3^6 mU3^6+48 m3^2 mQ3^8 mU3^6+302 m3^8 mU3^8-525 m3^6 mQ3^2 mU3^8+325 m3^4 mQ3^4 mU3^8-60 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8-55 m3^6 mU3^10+97 m3^4 mQ3^2 mU3^10-72 m3^2 mQ3^4 mU3^10+24 mQ3^6 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)-(64 (42 m3^13-6 m3^11 mQ3^2-67 m3^9 mQ3^4+45 m3^7 mQ3^6-120 m3^11 mU3^2-44 m3^9 mQ3^2 mU3^2+214 m3^7 mQ3^4 mU3^2-90 m3^5 mQ3^6 mU3^2+239 m3^9 mU3^4-67 m3^7 mQ3^2 mU3^4-195 m3^5 mQ3^4 mU3^4+59 m3^3 mQ3^6 mU3^4-252 m3^7 mU3^6+268 m3^5 mQ3^2 mU3^6-42 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6+59 m3^5 mU3^8-55 m3^3 mQ3^2 mU3^8-6 m3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(16 (148 m3^16 mQ3^2-336 m3^14 mQ3^4+194 m3^12 mQ3^6+m3^10 mQ3^8-13 m3^8 mQ3^10-788 m3^16 mU3^2+1796 m3^14 mQ3^2 mU3^2-786 m3^12 mQ3^4 mU3^2-452 m3^10 mQ3^6 mU3^2+190 m3^8 mQ3^8 mU3^2+64 m3^6 mQ3^10 mU3^2+2556 m3^14 mU3^4-7366 m3^12 mQ3^2 mU3^4+6232 m3^10 mQ3^4 mU3^4-936 m3^8 mQ3^6 mU3^4-479 m3^6 mQ3^8 mU3^4-37 m3^4 mQ3^10 mU3^4-2618 m3^12 mU3^6+8472 m3^10 mQ3^2 mU3^6-8712 m3^8 mQ3^4 mU3^6+2648 m3^6 mQ3^6 mU3^6+234 m3^4 mQ3^8 mU3^6-24 m3^2 mQ3^10 mU3^6+835 m3^10 mU3^8-3031 m3^8 mQ3^2 mU3^8+3742 m3^6 mQ3^4 mU3^8-1612 m3^4 mQ3^6 mU3^8+90 m3^2 mQ3^8 mU3^8+6 mQ3^10 mU3^8-58 m3^8 mU3^10+116 m3^6 mQ3^2 mU3^10-238 m3^4 mQ3^4 mU3^10+180 m3^2 mQ3^6 mU3^10-24 mQ3^8 mU3^10+69 m3^6 mU3^12-123 m3^4 mQ3^2 mU3^12+90 m3^2 mQ3^4 mU3^12-30 mQ3^6 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)+(32 (100 m3^11 mQ3^2-169 m3^9 mQ3^4+75 m3^7 mQ3^6+36 m3^11 mU3^2-392 m3^9 mQ3^2 mU3^2+508 m3^7 mQ3^4 mU3^2-148 m3^5 mQ3^6 mU3^2-159 m3^9 mU3^4+585 m3^7 mQ3^2 mU3^4-535 m3^5 mQ3^4 mU3^4+93 m3^3 mQ3^6 mU3^4+216 m3^7 mU3^6-384 m3^5 mQ3^2 mU3^6+152 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6-61 m3^5 mU3^8+59 m3^3 mQ3^2 mU3^8+12 m3 mQ3^4 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)-(2560 m3^4 mU3^2 (mQ3^2+mU3^2) Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Log[mU3^2/M^2]^2+Log[mQ3^2/M^2] (-((8 (4 m3^16 mQ3^4-80 m3^14 mQ3^6+184 m3^12 mQ3^8-144 m3^10 mQ3^10+36 m3^8 mQ3^12+128 m3^16 mQ3^2 mU3^2-1456 m3^14 mQ3^4 mU3^2+3318 m3^12 mQ3^6 mU3^2-3764 m3^10 mQ3^8 mU3^2+2055 m3^8 mQ3^10 mU3^2-426 m3^6 mQ3^12 mU3^2+9 m3^4 mQ3^14 mU3^2+180 m3^12 mQ3^4 msq^2 mU3^2-420 m3^10 mQ3^6 msq^2 mU3^2+510 m3^8 mQ3^8 msq^2 mU3^2-360 m3^6 mQ3^10 msq^2 mU3^2+90 m3^4 mQ3^12 msq^2 mU3^2-4 m3^16 mU3^4+756 m3^14 mQ3^2 mU3^4+616 m3^12 mQ3^4 mU3^4-3622 m3^10 mQ3^6 mU3^4+5488 m3^8 mQ3^8 mU3^4-3331 m3^6 mQ3^10 mU3^4+638 m3^4 mQ3^12 mU3^4+3 m3^2 mQ3^14 mU3^4-180 m3^12 mQ3^2 msq^2 mU3^4+180 m3^10 mQ3^4 msq^2 mU3^4-450 m3^8 mQ3^6 msq^2 mU3^4+630 m3^6 mQ3^8 msq^2 mU3^4-210 m3^4 mQ3^10 msq^2 mU3^4+30 m3^2 mQ3^12 msq^2 mU3^4+12 m3^14 mU3^6-2186 m3^12 mQ3^2 mU3^6+3060 m3^10 mQ3^4 mU3^6-1590 m3^8 mQ3^6 mU3^6-1684 m3^6 mQ3^8 mU3^6+2028 m3^4 mQ3^10 mU3^6-456 m3^2 mQ3^12 mU3^6+240 m3^10 mQ3^2 msq^2 mU3^6+210 m3^8 mQ3^4 msq^2 mU3^6-450 m3^6 mQ3^6 msq^2 mU3^6+30 m3^4 mQ3^8 msq^2 mU3^6-30 m3^2 mQ3^10 msq^2 mU3^6-12 m3^12 mU3^8+1906 m3^10 mQ3^2 mU3^8-3520 m3^8 mQ3^4 mU3^8+3694 m3^6 mQ3^6 mU3^8-1524 m3^4 mQ3^8 mU3^8-84 m3^2 mQ3^10 mU3^8+84 mQ3^12 mU3^8-270 m3^8 mQ3^2 msq^2 mU3^8+90 m3^6 mQ3^4 msq^2 mU3^8+150 m3^4 mQ3^6 msq^2 mU3^8+30 m3^2 mQ3^8 msq^2 mU3^8+4 m3^10 mU3^10-549 m3^8 mQ3^2 mU3^10+970 m3^6 mQ3^4 mU3^10-1017 m3^4 mQ3^6 mU3^10+540 m3^2 mQ3^8 mU3^10-84 mQ3^10 mU3^10+90 m3^6 mQ3^2 msq^2 mU3^10-60 m3^4 mQ3^4 msq^2 mU3^10-30 m3^2 mQ3^6 msq^2 mU3^10+9 m3^6 mQ3^2 mU3^12-6 m3^4 mQ3^4 mU3^12-3 m3^2 mQ3^6 mU3^12))/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))-(32 (32 m3^11 mQ3^2-64 m3^9 mQ3^4+32 m3^7 mQ3^6+8 m3^11 mU3^2+431 m3^9 mQ3^2 mU3^2-444 m3^7 mQ3^4 mU3^2+171 m3^5 mQ3^6 mU3^2-6 m3^3 mQ3^8 mU3^2-120 m3^7 mQ3^2 msq^2 mU3^2+180 m3^5 mQ3^4 msq^2 mU3^2-60 m3^3 mQ3^6 msq^2 mU3^2-259 m3^9 mU3^4-213 m3^7 mQ3^2 mU3^4+221 m3^5 mQ3^4 mU3^4-69 m3^3 mQ3^6 mU3^4+120 m3^7 msq^2 mU3^4-120 m3^5 mQ3^2 msq^2 mU3^4+245 m3^7 mU3^6-134 m3^5 mQ3^2 mU3^6+73 m3^3 mQ3^4 mU3^6-24 m3 mQ3^6 mU3^6-60 m3^5 msq^2 mU3^6+60 m3^3 mQ3^2 msq^2 mU3^6-6 m3^5 mU3^8+6 m3^3 mQ3^2 mU3^8) Xt)/(3 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(16 (4 m3^14 mQ3^2-76 m3^12 mQ3^4+108 m3^10 mQ3^6-36 m3^8 mQ3^8-4 m3^14 mU3^2+1472 m3^12 mQ3^2 mU3^2-2650 m3^10 mQ3^4 mU3^2+1680 m3^8 mQ3^6 mU3^2-391 m3^6 mQ3^8 mU3^2+9 m3^4 mQ3^10 mU3^2-180 m3^10 mQ3^2 msq^2 mU3^2+240 m3^8 mQ3^4 msq^2 mU3^2-270 m3^6 mQ3^6 msq^2 mU3^2+90 m3^4 mQ3^8 msq^2 mU3^2+12 m3^12 mU3^4-3290 m3^10 mQ3^2 mU3^4+5696 m3^8 mQ3^4 mU3^4-3513 m3^6 mQ3^6 mU3^4+732 m3^4 mQ3^8 mU3^4+3 m3^2 mQ3^10 mU3^4+240 m3^8 mQ3^2 msq^2 mU3^4+180 m3^6 mQ3^4 msq^2 mU3^4-90 m3^4 mQ3^6 msq^2 mU3^4+30 m3^2 mQ3^8 msq^2 mU3^4-12 m3^10 mU3^6+2448 m3^8 mQ3^2 mU3^6-3993 m3^6 mQ3^4 mU3^6+2430 m3^4 mQ3^6 mU3^6-489 m3^2 mQ3^8 mU3^6-270 m3^6 mQ3^2 msq^2 mU3^6-90 m3^4 mQ3^4 msq^2 mU3^6+4 m3^8 mU3^8-591 m3^6 mQ3^2 mU3^8+852 m3^4 mQ3^4 mU3^8-513 m3^2 mQ3^6 mU3^8+96 mQ3^8 mU3^8+90 m3^4 mQ3^2 msq^2 mU3^8+30 m3^2 mQ3^4 msq^2 mU3^8+9 m3^4 mQ3^2 mU3^10+3 m3^2 mQ3^4 mU3^10) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(128 (16 m3^9 mQ3^4-16 m3^7 mQ3^6+22 m3^11 mU3^2+6 m3^9 mQ3^2 mU3^2-217 m3^7 mQ3^4 mU3^2+165 m3^5 mQ3^6 mU3^2-3 m3^3 mQ3^8 mU3^2+60 m3^5 mQ3^4 msq^2 mU3^2-30 m3^3 mQ3^6 msq^2 mU3^2+10 m3^9 mU3^4+238 m3^7 mQ3^2 mU3^4-17 m3^5 mQ3^4 mU3^4-115 m3^3 mQ3^6 mU3^4-120 m3^5 mQ3^2 msq^2 mU3^4+30 m3^3 mQ3^4 msq^2 mU3^4-193 m3^7 mU3^6-85 m3^5 mQ3^2 mU3^6+106 m3^3 mQ3^4 mU3^6+18 m3 mQ3^6 mU3^6+60 m3^5 msq^2 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6+153 m3^5 mU3^8-91 m3^3 mQ3^2 mU3^8+6 m3 mQ3^4 mU3^8-30 m3^3 msq^2 mU3^8-3 m3^3 mU3^10) Xt^3)/(3 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(8 (4 m3^16 mQ3^4-80 m3^14 mQ3^6+184 m3^12 mQ3^8-144 m3^10 mQ3^10+36 m3^8 mQ3^12-80 m3^16 mQ3^2 mU3^2+3172 m3^14 mQ3^4 mU3^2-9086 m3^12 mQ3^6 mU3^2+9890 m3^10 mQ3^8 mU3^2-4599 m3^8 mQ3^10 mU3^2+696 m3^6 mQ3^12 mU3^2-9 m3^4 mQ3^14 mU3^2-180 m3^12 mQ3^4 msq^2 mU3^2+420 m3^10 mQ3^6 msq^2 mU3^2-510 m3^8 mQ3^8 msq^2 mU3^2+360 m3^6 mQ3^10 msq^2 mU3^2-90 m3^4 mQ3^12 msq^2 mU3^2-4 m3^16 mU3^4+2160 m3^14 mQ3^2 mU3^4-12968 m3^12 mQ3^4 mU3^4+25568 m3^10 mQ3^6 mU3^4-22956 m3^8 mQ3^8 mU3^4+9529 m3^6 mQ3^10 mU3^4-1262 m3^4 mQ3^12 mU3^4-3 m3^2 mQ3^14 mU3^4-180 m3^12 mQ3^2 msq^2 mU3^4+660 m3^10 mQ3^4 msq^2 mU3^4-570 m3^8 mQ3^6 msq^2 mU3^4+90 m3^6 mQ3^8 msq^2 mU3^4+30 m3^4 mQ3^10 msq^2 mU3^4-30 m3^2 mQ3^12 msq^2 mU3^4+12 m3^14 mU3^6-4430 m3^12 mQ3^2 mU3^6+17622 m3^10 mQ3^4 mU3^6-25654 m3^8 mQ3^6 mU3^6+18064 m3^6 mQ3^8 mU3^6-6592 m3^4 mQ3^10 mU3^6+882 m3^2 mQ3^12 mU3^6+240 m3^10 mQ3^2 msq^2 mU3^6-330 m3^8 mQ3^4 msq^2 mU3^6-90 m3^6 mQ3^6 msq^2 mU3^6+210 m3^4 mQ3^8 msq^2 mU3^6-30 m3^2 mQ3^10 msq^2 mU3^6-12 m3^12 mU3^8+3020 m3^10 mQ3^2 mU3^8-9712 m3^8 mQ3^4 mU3^8+10886 m3^6 mQ3^6 mU3^8-5480 m3^4 mQ3^8 mU3^8+1602 m3^2 mQ3^10 mU3^8-240 mQ3^12 mU3^8-270 m3^8 mQ3^2 msq^2 mU3^8+270 m3^6 mQ3^4 msq^2 mU3^8+30 m3^4 mQ3^6 msq^2 mU3^8-30 m3^2 mQ3^8 msq^2 mU3^8+4 m3^10 mU3^10-651 m3^8 mQ3^2 mU3^10+1840 m3^6 mQ3^4 mU3^10-1731 m3^4 mQ3^6 mU3^10+618 m3^2 mQ3^8 mU3^10-96 mQ3^10 mU3^10+90 m3^6 mQ3^2 msq^2 mU3^10-60 m3^4 mQ3^4 msq^2 mU3^10-30 m3^2 mQ3^6 msq^2 mU3^10+9 m3^6 mQ3^2 mU3^12-6 m3^4 mQ3^4 mU3^12-3 m3^2 mQ3^6 mU3^12) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^4 mU3^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(64 (16 m3^11 mQ3^2-32 m3^9 mQ3^4+16 m3^7 mQ3^6+52 m3^11 mU3^2-362 m3^9 mQ3^2 mU3^2+558 m3^7 mQ3^4 mU3^2-219 m3^5 mQ3^6 mU3^2+3 m3^3 mQ3^8 mU3^2+60 m3^7 mQ3^2 msq^2 mU3^2-90 m3^5 mQ3^4 msq^2 mU3^2+30 m3^3 mQ3^6 msq^2 mU3^2-276 m3^9 mU3^4+770 m3^7 mQ3^2 mU3^4-737 m3^5 mQ3^4 mU3^4+179 m3^3 mQ3^6 mU3^4+60 m3^7 msq^2 mU3^4-120 m3^5 mQ3^2 msq^2 mU3^4+60 m3^3 mQ3^4 msq^2 mU3^4+214 m3^7 mU3^6-407 m3^5 mQ3^2 mU3^6+213 m3^3 mQ3^4 mU3^6+12 m3 mQ3^6 mU3^6-30 m3^5 msq^2 mU3^6+30 m3^3 mQ3^2 msq^2 mU3^6-3 m3^5 mU3^8+3 m3^3 mQ3^2 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(5120 m3^4 mQ3^2 Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+((16 (29 m3^12-109 m3^10 mQ3^2+30 m3^8 mQ3^4-10 m3^6 mQ3^6-127 m3^10 mU3^2+447 m3^8 mQ3^2 mU3^2-210 m3^6 mQ3^4 mU3^2+70 m3^4 mQ3^6 mU3^2+57 m3^8 mU3^4-237 m3^6 mQ3^2 mU3^4-19 m3^6 mU3^6+79 m3^4 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+((96 m3^3 (-2 m3^2+mQ3^2+mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))+(32 (m3^7+59 m3^5 mQ3^2+59 m3^5 mU3^2-119 m3^3 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Xt+(320 (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+((64 (m3^7+59 m3^5 mQ3^2+59 m3^5 mU3^2-119 m3^3 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(192 m3^3 (2 m3^4-mQ3^4+4 mQ3^2 mU3^2-mU3^4-2 m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^3-(160 (mQ3^2+mU3^2) (2 m3^12-10 m3^10 mQ3^2+3 m3^8 mQ3^4-m3^6 mQ3^6-10 m3^10 mU3^2+42 m3^8 mQ3^2 mU3^2-21 m3^6 mQ3^4 mU3^2+7 m3^4 mQ3^6 mU3^2+3 m3^8 mU3^4-21 m3^6 mQ3^2 mU3^4-m3^6 mU3^6+7 m3^4 mQ3^2 mU3^6) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(320 (mQ3^2+mU3^2) (m3^7+5 m3^5 mQ3^2+5 m3^5 mU3^2-11 m3^3 mQ3^2 mU3^2) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Log[msq^2/M^2]+((16 (91 m3^16-334 m3^14 mQ3^2+361 m3^12 mQ3^4-132 m3^10 mQ3^6+18 m3^8 mQ3^8-332 m3^14 mU3^2+1204 m3^12 mQ3^2 mU3^2-1272 m3^10 mQ3^4 mU3^2+452 m3^8 mQ3^6 mU3^2-68 m3^6 mQ3^8 mU3^2+356 m3^12 mU3^4-1264 m3^10 mQ3^2 mU3^4+1166 m3^8 mQ3^4 mU3^4-252 m3^6 mQ3^6 mU3^4+18 m3^4 mQ3^8 mU3^4-128 m3^10 mU3^6+444 m3^8 mQ3^2 mU3^6-248 m3^6 mQ3^4 mU3^6-132 m3^4 mQ3^6 mU3^6+48 m3^2 mQ3^8 mU3^6+17 m3^8 mU3^8-66 m3^6 mQ3^2 mU3^8+17 m3^4 mQ3^4 mU3^8+48 m3^2 mQ3^6 mU3^8-12 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4)+(64 (m3^13-43 m3^11 mQ3^2+67 m3^9 mQ3^4-41 m3^7 mQ3^6+38 m3^11 mU3^2+6 m3^9 mQ3^2 mU3^2-94 m3^7 mQ3^4 mU3^2+98 m3^5 mQ3^6 mU3^2-63 m3^9 mU3^4+85 m3^7 mQ3^2 mU3^4+3 m3^5 mQ3^4 mU3^4-73 m3^3 mQ3^6 mU3^4+40 m3^7 mU3^6-96 m3^5 mQ3^2 mU3^6+72 m3^3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(64 (64 m3^18-310 m3^16 mQ3^2+516 m3^14 mQ3^4-405 m3^12 mQ3^6+164 m3^10 mQ3^8-27 m3^8 mQ3^10-310 m3^16 mU3^2+1368 m3^14 mQ3^2 mU3^2-2083 m3^12 mQ3^4 mU3^2+1488 m3^10 mQ3^6 mU3^2-561 m3^8 mQ3^8 mU3^2+88 m3^6 mQ3^10 mU3^2+516 m3^14 mU3^4-2083 m3^12 mQ3^2 mU3^4+2808 m3^10 mQ3^4 mU3^4-1704 m3^8 mQ3^6 mU3^4+564 m3^6 mQ3^8 mU3^4-81 m3^4 mQ3^10 mU3^4-405 m3^12 mU3^6+1488 m3^10 mQ3^2 mU3^6-1704 m3^8 mQ3^4 mU3^6+776 m3^6 mQ3^6 mU3^6-199 m3^4 mQ3^8 mU3^6+24 m3^2 mQ3^10 mU3^6+164 m3^10 mU3^8-561 m3^8 mQ3^2 mU3^8+564 m3^6 mQ3^4 mU3^8-199 m3^4 mQ3^6 mU3^8+48 m3^2 mQ3^8 mU3^8-6 mQ3^10 mU3^8-27 m3^8 mU3^10+88 m3^6 mQ3^2 mU3^10-81 m3^4 mQ3^4 mU3^10+24 m3^2 mQ3^6 mU3^10-6 mQ3^8 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)+(128 (m3^15+33 m3^13 mQ3^2-147 m3^11 mQ3^4+181 m3^9 mQ3^6-52 m3^7 mQ3^8-39 m3^13 mU3^2+9 m3^11 mQ3^2 mU3^2+273 m3^9 mQ3^4 mU3^2-409 m3^7 mQ3^6 mU3^2+102 m3^5 mQ3^8 mU3^2+153 m3^11 mU3^4-291 m3^9 mQ3^2 mU3^4+9 m3^7 mQ3^4 mU3^4+279 m3^5 mQ3^6 mU3^4-54 m3^3 mQ3^8 mU3^4-183 m3^9 mU3^6+415 m3^7 mQ3^2 mU3^6-285 m3^5 mQ3^4 mU3^6+m3^3 mQ3^6 mU3^6-12 m3 mQ3^8 mU3^6+52 m3^7 mU3^8-102 m3^5 mQ3^2 mU3^8+54 m3^3 mQ3^4 mU3^8+12 m3 mQ3^6 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)+(32 (320 m3^18 mQ3^2-1036 m3^16 mQ3^4+1044 m3^14 mQ3^6-321 m3^12 mQ3^8+36 m3^10 mQ3^10-41 m3^8 mQ3^12+320 m3^18 mU3^2-2584 m3^16 mQ3^2 mU3^2+6252 m3^14 mQ3^4 mU3^2-5712 m3^12 mQ3^6 mU3^2+1612 m3^10 mQ3^8 mU3^2-24 m3^8 mQ3^10 mU3^2+128 m3^6 mQ3^12 mU3^2-1036 m3^16 mU3^4+6252 m3^14 mQ3^2 mU3^4-13598 m3^12 mQ3^4 mU3^4+12176 m3^10 mQ3^6 mU3^4-3615 m3^8 mQ3^8 mU3^4-44 m3^6 mQ3^10 mU3^4-125 m3^4 mQ3^12 mU3^4+1044 m3^14 mU3^6-5712 m3^12 mQ3^2 mU3^6+12176 m3^10 mQ3^4 mU3^6-11360 m3^8 mQ3^6 mU3^6+3884 m3^6 mQ3^8 mU3^6-80 m3^4 mQ3^10 mU3^6+48 m3^2 mQ3^12 mU3^6-321 m3^12 mU3^8+1612 m3^10 mQ3^2 mU3^8-3615 m3^8 mQ3^4 mU3^8+3884 m3^6 mQ3^6 mU3^8-1702 m3^4 mQ3^8 mU3^8+144 m3^2 mQ3^10 mU3^8-12 mQ3^12 mU3^8+36 m3^10 mU3^10-24 m3^8 mQ3^2 mU3^10-44 m3^6 mQ3^4 mU3^10-80 m3^4 mQ3^6 mU3^10+144 m3^2 mQ3^8 mU3^10-24 mQ3^10 mU3^10-41 m3^8 mU3^12+128 m3^6 mQ3^2 mU3^12-125 m3^4 mQ3^4 mU3^12+48 m3^2 mQ3^6 mU3^12-12 mQ3^8 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)-(128 (34 m3^13 mQ3^2-107 m3^11 mQ3^4+115 m3^9 mQ3^6-34 m3^7 mQ3^8+34 m3^13 mU3^2-214 m3^11 mQ3^2 mU3^2+411 m3^9 mQ3^4 mU3^2-314 m3^7 mQ3^6 mU3^2+67 m3^5 mQ3^8 mU3^2-107 m3^11 mU3^4+411 m3^9 mQ3^2 mU3^4-560 m3^7 mQ3^4 mU3^4+291 m3^5 mQ3^6 mU3^4-35 m3^3 mQ3^8 mU3^4+115 m3^9 mU3^6-314 m3^7 mQ3^2 mU3^6+291 m3^5 mQ3^4 mU3^6-70 m3^3 mQ3^6 mU3^6-6 m3 mQ3^8 mU3^6-34 m3^7 mU3^8+67 m3^5 mQ3^2 mU3^8-35 m3^3 mQ3^4 mU3^8-6 m3 mQ3^6 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(2560 m3^4 (mQ3^2+mU3^2) (m3^2 mQ3^2+m3^2 mU3^2-2 mQ3^2 mU3^2) Xt^6)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Log[mU3^2/M^2])+(-((64 m3^4 (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(-m3^2+mQ3^2)^2)+(512 (11 m3^11-10 m3^9 mQ3^2+7 m3^7 mQ3^4-18 m3^9 mU3^2+8 m3^7 mQ3^2 mU3^2-6 m3^5 mQ3^4 mU3^2+11 m3^7 mU3^4-6 m3^5 mQ3^2 mU3^4+3 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(8192 m3^6 Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+(128 m3 (2 m3^2-mQ3^2-mU3^2) (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Xt^3)/(mQ3^2-mU3^2)^3-(64 (2 m3^2-mQ3^2-mU3^2) (53 m3^8-50 m3^6 mQ3^2-7 m3^4 mQ3^4-50 m3^6 mU3^2+52 m3^4 mQ3^2 mU3^2+6 m3^2 mQ3^4 mU3^2-7 m3^4 mU3^4+6 m3^2 mQ3^2 mU3^4-3 mQ3^4 mU3^4) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(1024 m3^3 (-2 m3^2+mQ3^2+mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mQ3^2]+(-((64 m3^4 (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)))/(m3^2-mU3^2)^2)-(512 (11 m3^11-18 m3^9 mQ3^2+11 m3^7 mQ3^4-10 m3^9 mU3^2+8 m3^7 mQ3^2 mU3^2-6 m3^5 mQ3^4 mU3^2+7 m3^7 mU3^4-6 m3^5 mQ3^2 mU3^4+3 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(8192 m3^6 Xt^2)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(128 m3 (-2 m3^2+mQ3^2+mU3^2) (2+(8 (2 m3^8-2 m3^6 mQ3^2+m3^4 mQ3^4-2 m3^6 mU3^2+m3^4 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Xt^3)/(mQ3^2-mU3^2)^3+(64 (2 m3^2-mQ3^2-mU3^2) (53 m3^8-50 m3^6 mQ3^2-7 m3^4 mQ3^4-50 m3^6 mU3^2+52 m3^4 mQ3^2 mU3^2+6 m3^2 mQ3^4 mU3^2-7 m3^4 mU3^4+6 m3^2 mQ3^2 mU3^4-3 mQ3^4 mU3^4) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(1024 m3^3 (-2 m3^2+mQ3^2+mU3^2) Xt^5)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/mU3^2])+((4 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3)-(32 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(16 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(128 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-mQ3^2/mU3^2]+(-((80 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4))/(m3^2-mU3^2)^3)+(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(160 (-msq^2+mU3^2) (2 m3^4-3 m3^2 msq^2+3 m3^2 mU3^2-msq^2 mU3^2-mU3^4) Xt^4)/((m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)-(1280 m3 (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)) PolyLog[2,1-msq^2/mU3^2]+Log[M^2/mt^2] (-((16 (188 m3^8-316 m3^6 mQ3^2+169 m3^4 mQ3^4-9 m3^2 mQ3^6-180 m3^6 msq^2+150 m3^4 mQ3^2 msq^2-90 m3^2 mQ3^4 msq^2-316 m3^6 mU3^2+454 m3^4 mQ3^2 mU3^2-211 m3^2 mQ3^4 mU3^2-3 mQ3^6 mU3^2+150 m3^4 msq^2 mU3^2+120 m3^2 mQ3^2 msq^2 mU3^2-30 mQ3^4 msq^2 mU3^2+169 m3^4 mU3^4-211 m3^2 mQ3^2 mU3^4+98 mQ3^4 mU3^4-90 m3^2 msq^2 mU3^4-30 mQ3^2 msq^2 mU3^4-9 m3^2 mU3^6-3 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+((1792 m3^3)/(3 (m3^2-mQ3^2) (m3^2-mU3^2))-(128 m3 mQ3^2)/((m3^2-mQ3^2) (m3^2-mU3^2))-(1280 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2))-(128 m3 mU3^2)/((m3^2-mQ3^2) (m3^2-mU3^2))) Xt-128 z3+(-((16 (198 m3^16-726 m3^14 mQ3^2+721 m3^12 mQ3^4-232 m3^10 mQ3^6+27 m3^8 mQ3^8-726 m3^14 mU3^2+2700 m3^12 mQ3^2 mU3^2-2700 m3^10 mQ3^4 mU3^2+908 m3^8 mQ3^6 mU3^2-134 m3^6 mQ3^8 mU3^2+721 m3^12 mU3^4-2700 m3^10 mQ3^2 mU3^4+2240 m3^8 mQ3^4 mU3^4-320 m3^6 mQ3^6 mU3^4-13 m3^4 mQ3^8 mU3^4-232 m3^10 mU3^6+908 m3^8 mQ3^2 mU3^6-320 m3^6 mQ3^4 mU3^6-452 m3^4 mQ3^6 mU3^6+144 m3^2 mQ3^8 mU3^6+27 m3^8 mU3^8-134 m3^6 mQ3^2 mU3^8-13 m3^4 mQ3^4 mU3^8+144 m3^2 mQ3^6 mU3^8-36 mQ3^8 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^4))+(128 (m3^11-50 m3^9 mQ3^2+41 m3^7 mQ3^4-50 m3^9 mU3^2+196 m3^7 mQ3^2 mU3^2-130 m3^5 mQ3^4 mU3^2+41 m3^7 mU3^4-130 m3^5 mQ3^2 mU3^4+81 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(512 m3^6 Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[m3^2/M^2]^2+(-((8 (128 m3^16-556 m3^14 mQ3^2+722 m3^12 mQ3^4-473 m3^10 mQ3^6+192 m3^8 mQ3^8-66 m3^6 mQ3^10+38 m3^4 mQ3^12-9 m3^2 mQ3^14+60 m3^12 mQ3^2 msq^2+120 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2-90 m3^10 mQ3^2 msq^4+60 m3^8 mQ3^4 msq^4+30 m3^6 mQ3^6 msq^4-340 m3^14 mU3^2+1704 m3^12 mQ3^2 mU3^2-2133 m3^10 mQ3^4 mU3^2+1137 m3^8 mQ3^6 mU3^2-158 m3^6 mQ3^8 mU3^2-158 m3^4 mQ3^10 mU3^2+47 m3^2 mQ3^12 mU3^2-3 mQ3^14 mU3^2-60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2-180 m3^8 mQ3^4 msq^2 mU3^2+540 m3^6 mQ3^6 msq^2 mU3^2+90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-210 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+262 m3^12 mU3^4-1879 m3^10 mQ3^2 mU3^4+2495 m3^8 mQ3^4 mU3^4-1518 m3^6 mQ3^6 mU3^4+532 m3^4 mQ3^8 mU3^4-35 m3^2 mQ3^10 mU3^4-mQ3^12 mU3^4+180 m3^10 msq^2 mU3^4+540 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4-540 m3^4 mQ3^6 msq^2 mU3^4-270 m3^8 msq^4 mU3^4-90 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4+90 m3^2 mQ3^6 msq^4 mU3^4+5 m3^10 mU3^6+699 m3^8 mQ3^2 mU3^6-894 m3^6 mQ3^4 mU3^6+422 m3^4 mQ3^6 mU3^6-143 m3^2 mQ3^8 mU3^6+7 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6-420 m3^6 mQ3^2 msq^2 mU3^6+420 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+270 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6-150 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-43 m3^8 mU3^8-52 m3^6 mQ3^2 mU3^8+62 m3^4 mQ3^4 mU3^8+12 m3^2 mQ3^6 mU3^8-3 mQ3^8 mU3^8+60 m3^6 msq^2 mU3^8+120 m3^4 mQ3^2 msq^2 mU3^8-180 m3^2 mQ3^4 msq^2 mU3^8-90 m3^4 msq^4 mU3^8+60 m3^2 mQ3^2 msq^4 mU3^8+30 mQ3^4 msq^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))+(64 (18 m3^11 mQ3^2-84 m3^9 mQ3^4+60 m3^7 mQ3^6-15 m3^5 mQ3^8+8 m3^3 mQ3^10-3 m3 mQ3^12+60 m3^7 mQ3^4 msq^2-60 m3^5 mQ3^6 msq^2-30 m3^7 mQ3^2 msq^4+30 m3^5 mQ3^4 msq^4-18 m3^11 mU3^2+31 m3^9 mQ3^2 mU3^2+138 m3^7 mQ3^4 mU3^2-103 m3^5 mQ3^6 mU3^2-10 m3^3 mQ3^8 mU3^2+10 m3 mQ3^10 mU3^2-60 m3^7 mQ3^2 msq^2 mU3^2-60 m3^5 mQ3^4 msq^2 mU3^2+120 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2+30 m3^5 mQ3^2 msq^4 mU3^2-60 m3^3 mQ3^4 msq^4 mU3^2+37 m3^9 mU3^4-112 m3^7 mQ3^2 mU3^4-59 m3^5 mQ3^4 mU3^4+106 m3^3 mQ3^6 mU3^4-20 m3 mQ3^8 mU3^4+120 m3^5 mQ3^2 msq^2 mU3^4-60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-60 m3^5 msq^4 mU3^4+30 m3^3 mQ3^2 msq^4 mU3^4+30 m3 mQ3^4 msq^4 mU3^4-22 m3^7 mU3^6+81 m3^5 mQ3^2 mU3^6-40 m3^3 mQ3^4 mU3^6-3 m3 mQ3^6 mU3^6-60 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+30 m3^3 msq^4 mU3^6-30 m3 mQ3^2 msq^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(512 m3^2 mQ3^4 Xt^2)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)) Log[mQ3^2/M^2]^2+((16 (102 m3^8-203 m3^6 mQ3^2+101 m3^4 mQ3^4+180 m3^6 msq^2-150 m3^4 mQ3^2 msq^2+90 m3^2 mQ3^4 msq^2-203 m3^6 mU3^2+404 m3^4 mQ3^2 mU3^2-201 m3^2 mQ3^4 mU3^2-150 m3^4 msq^2 mU3^2-120 m3^2 mQ3^2 msq^2 mU3^2+30 mQ3^4 msq^2 mU3^2+101 m3^4 mU3^4-201 m3^2 mQ3^2 mU3^4+100 mQ3^4 mU3^4+90 m3^2 msq^2 mU3^4+30 mQ3^2 msq^2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(1280 m3 msq^2 Xt)/((m3^2-mQ3^2) (m3^2-mU3^2))) Log[msq^2/M^2]+(-48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))-48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt) Log[msq^2/M^2]^2+((16 (54 m3^10 mQ3^2-108 m3^8 mQ3^4+54 m3^6 mQ3^6-310 m3^10 mU3^2+582 m3^8 mQ3^2 mU3^2-296 m3^6 mQ3^4 mU3^2+21 m3^4 mQ3^6 mU3^2-9 m3^2 mQ3^8 mU3^2-90 m3^6 mQ3^2 msq^2 mU3^2+180 m3^4 mQ3^4 msq^2 mU3^2-90 m3^2 mQ3^6 msq^2 mU3^2+550 m3^8 mU3^4-1078 m3^6 mQ3^2 mU3^4+679 m3^4 mQ3^4 mU3^4-100 m3^2 mQ3^6 mU3^4-3 mQ3^8 mU3^4+90 m3^6 msq^2 mU3^4-210 m3^4 mQ3^2 msq^2 mU3^4+150 m3^2 mQ3^4 msq^2 mU3^4-30 mQ3^6 msq^2 mU3^4-216 m3^6 mU3^6+359 m3^4 mQ3^2 mU3^6-296 m3^2 mQ3^4 mU3^6+81 mQ3^6 mU3^6+30 m3^4 msq^2 mU3^6-60 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6-35 m3^4 mU3^8+158 m3^2 mQ3^2 mU3^8-75 mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2))-(128 (16 m3^7-16 m3^5 mQ3^2+10 m3^5 mU3^2-18 m3^3 mQ3^2 mU3^2+3 m3 mQ3^4 mU3^2-30 m3^3 msq^2 mU3^2+30 m3 mQ3^2 msq^2 mU3^2-26 m3^3 mU3^4+28 m3 mQ3^2 mU3^4+3 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(1/(3 (m3^2-mU3^2)^3) 16 (60 m3^4 msq^2-90 m3^2 msq^4+2 m3^4 mU3^2+180 m3^2 msq^2 mU3^2-30 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6)-(64 (-60 m3 msq^4+m3^3 mU3^2+120 m3 msq^2 mU3^2-m3 mU3^4) Xt)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Log[msq^2/M^2]) Log[mU3^2/M^2]+(-((8 (128 m3^10+44 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-556 m3^8 mU3^2+64 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+706 m3^6 mU3^4-51 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-437 m3^4 mU3^6-38 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6+136 m3^2 mU3^8+5 mQ3^2 mU3^8-mU3^10))/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2)))+(64 (-18 m3^7 mQ3^2+m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^2 msq^4+18 m3^7 mU3^2+68 m3^5 mQ3^2 mU3^2-11 m3^3 mQ3^4 mU3^2-3 m3 mQ3^6 mU3^2-60 m3^3 mQ3^2 msq^2 mU3^2-30 m3^3 msq^4 mU3^2-30 m3 mQ3^2 msq^4 mU3^2-85 m3^5 mU3^4-17 m3^3 mQ3^2 mU3^4+10 m3 mQ3^4 mU3^4+60 m3^3 msq^2 mU3^4+60 m3 mQ3^2 msq^2 mU3^4+30 m3 msq^4 mU3^4+57 m3^3 mU3^6-17 m3 mQ3^2 mU3^6-60 m3 msq^2 mU3^6-6 m3 mU3^8) Xt)/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^2)+(512 m3^2 mU3^4 Xt^2)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)) Log[mU3^2/M^2]^2+Log[mQ3^2/M^2] ((16 (-312 m3^10 mQ3^2+555 m3^8 mQ3^4-220 m3^6 mQ3^6-34 m3^4 mQ3^8-9 m3^2 mQ3^10+90 m3^6 mQ3^4 msq^2+30 m3^4 mQ3^6 msq^2+56 m3^10 mU3^2+580 m3^8 mQ3^2 mU3^2-1081 m3^6 mQ3^4 mU3^2+363 m3^4 mQ3^6 mU3^2+157 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2-90 m3^6 mQ3^2 msq^2 mU3^2-210 m3^4 mQ3^4 msq^2 mU3^2-60 m3^2 mQ3^6 msq^2 mU3^2-111 m3^8 mU3^4-290 m3^6 mQ3^2 mU3^4+676 m3^4 mQ3^4 mU3^4-296 m3^2 mQ3^6 mU3^4-75 mQ3^8 mU3^4+180 m3^4 mQ3^2 msq^2 mU3^4+150 m3^2 mQ3^4 msq^2 mU3^4+30 mQ3^6 msq^2 mU3^4+55 m3^6 mU3^6+19 m3^4 mQ3^2 mU3^6-99 m3^2 mQ3^4 mU3^6+81 mQ3^6 mU3^6-90 m3^2 mQ3^2 msq^2 mU3^6-30 mQ3^4 msq^2 mU3^6-9 m3^2 mQ3^2 mU3^8-3 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(128 (16 m3^7+10 m3^5 mQ3^2-26 m3^3 mQ3^4+3 m3 mQ3^6-30 m3^3 mQ3^2 msq^2-16 m3^5 mU3^2-18 m3^3 mQ3^2 mU3^2+28 m3 mQ3^4 mU3^2+30 m3 mQ3^2 msq^2 mU3^2+3 m3 mQ3^2 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))+(1/(3 (m3^2-mQ3^2)^3) 16 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+60 m3^4 msq^2+180 m3^2 mQ3^2 msq^2-90 m3^2 msq^4-30 mQ3^2 msq^4)-(64 (m3^3 mQ3^2-m3 mQ3^4+120 m3 mQ3^2 msq^2-60 m3 msq^4) Xt)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) Log[msq^2/M^2]+((16 (12 m3^10 mQ3^2-14 m3^8 mQ3^4+34 m3^6 mQ3^6-29 m3^4 mQ3^8+9 m3^2 mQ3^10-2 m3^10 mU3^2+70 m3^8 mQ3^2 mU3^2-194 m3^6 mQ3^4 mU3^2+122 m3^4 mQ3^6 mU3^2-35 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2+3 m3^8 mU3^4-87 m3^6 mQ3^2 mU3^4+184 m3^4 mQ3^4 mU3^4-68 m3^2 mQ3^6 mU3^4+4 mQ3^8 mU3^4-m3^6 mU3^6+29 m3^4 mQ3^2 mU3^6-60 m3^2 mQ3^4 mU3^6+20 mQ3^6 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(64 (-10 m3^5 mQ3^6+10 m3^3 mQ3^8-6 m3 mQ3^10+31 m3^7 mQ3^2 mU3^2-26 m3^5 mQ3^4 mU3^2+9 m3^3 mQ3^6 mU3^2+20 m3 mQ3^8 mU3^2+m3^7 mU3^4-59 m3^5 mQ3^2 mU3^4+45 m3^3 mQ3^4 mU3^4-37 m3 mQ3^6 mU3^4-m3^5 mU3^6+32 m3^3 mQ3^2 mU3^6-9 m3 mQ3^4 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(1024 m3^2 mQ3^2 mU3^2 Xt^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)) Log[mU3^2/M^2])+Log[m3^2/M^2] ((16 (660 m3^12-1562 m3^10 mQ3^2+1182 m3^8 mQ3^4-269 m3^6 mQ3^6+9 m3^4 mQ3^8-180 m3^10 msq^2+240 m3^8 mQ3^2 msq^2-270 m3^6 mQ3^4 msq^2+90 m3^4 mQ3^6 msq^2-1562 m3^10 mU3^2+3564 m3^8 mQ3^2 mU3^2-2575 m3^6 mQ3^4 mU3^2+498 m3^4 mQ3^6 mU3^2+3 m3^2 mQ3^8 mU3^2+240 m3^8 msq^2 mU3^2+180 m3^6 mQ3^2 msq^2 mU3^2-90 m3^4 mQ3^4 msq^2 mU3^2+30 m3^2 mQ3^6 msq^2 mU3^2+1182 m3^8 mU3^4-2575 m3^6 mQ3^2 mU3^4+1822 m3^4 mQ3^4 mU3^4-333 m3^2 mQ3^6 mU3^4-270 m3^6 msq^2 mU3^4-90 m3^4 mQ3^2 msq^2 mU3^4-269 m3^6 mU3^6+498 m3^4 mQ3^2 mU3^6-333 m3^2 mQ3^4 mU3^6+48 mQ3^6 mU3^6+90 m3^4 msq^2 mU3^6+30 m3^2 mQ3^2 msq^2 mU3^6+9 m3^4 mU3^8+3 m3^2 mQ3^2 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(128 (40 m3^7-40 m3^5 mQ3^2+3 m3^3 mQ3^4-60 m3^5 msq^2+30 m3^3 mQ3^2 msq^2-40 m3^5 mU3^2+34 m3^3 mQ3^2 mU3^2+30 m3^3 msq^2 mU3^2+3 m3^3 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+((64 (32 m3^16-89 m3^14 mQ3^2+32 m3^12 mQ3^4+33 m3^10 mQ3^6-19 m3^8 mQ3^8+5 m3^6 mQ3^10-103 m3^14 mU3^2+329 m3^12 mQ3^2 mU3^2-170 m3^10 mQ3^4 mU3^2-49 m3^8 mQ3^6 mU3^2+25 m3^6 mQ3^8 mU3^2-8 m3^4 mQ3^10 mU3^2+119 m3^12 mU3^4-451 m3^10 mQ3^2 mU3^4+290 m3^8 mQ3^4 mU3^4-3 m3^6 mQ3^6 mU3^4+9 m3^4 mQ3^8 mU3^4-52 m3^10 mU3^6+251 m3^8 mQ3^2 mU3^6-164 m3^6 mQ3^4 mU3^6-11 m3^4 mQ3^6 mU3^6+7 m3^8 mU3^8-55 m3^6 mQ3^2 mU3^8+42 m3^4 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^4 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(64 (68 m3^11-233 m3^9 mQ3^2+140 m3^7 mQ3^4-7 m3^5 mQ3^6-139 m3^9 mU3^2+459 m3^7 mQ3^2 mU3^2-241 m3^5 mQ3^4 mU3^2-15 m3^3 mQ3^6 mU3^2+77 m3^7 mU3^4-260 m3^5 mQ3^2 mU3^4+151 m3^3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(1024 m3^4 mQ3^2 Xt^2)/((m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2))) Log[mQ3^2/M^2]+((16 (58 m3^12-236 m3^10 mQ3^2+87 m3^8 mQ3^4-29 m3^6 mQ3^6-236 m3^10 mU3^2+894 m3^8 mQ3^2 mU3^2-447 m3^6 mQ3^4 mU3^2+149 m3^4 mQ3^6 mU3^2+87 m3^8 mU3^4-447 m3^6 mQ3^2 mU3^4-29 m3^6 mU3^6+149 m3^4 mQ3^2 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(64 (m3^7+59 m3^5 mQ3^2+59 m3^5 mU3^2-119 m3^3 mQ3^2 mU3^2) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) Log[msq^2/M^2]+(-((16 (128 m3^16-410 m3^14 mQ3^2+472 m3^12 mQ3^4-205 m3^10 mQ3^6+27 m3^8 mQ3^8-358 m3^14 mU3^2+1314 m3^12 mQ3^2 mU3^2-1797 m3^10 mQ3^4 mU3^2+999 m3^8 mQ3^6 mU3^2-218 m3^6 mQ3^8 mU3^2+134 m3^12 mU3^4-683 m3^10 mQ3^2 mU3^4+1157 m3^8 mQ3^4 mU3^4-655 m3^6 mQ3^6 mU3^4+167 m3^4 mQ3^8 mU3^4+125 m3^10 mU3^6-191 m3^8 mQ3^2 mU3^6-11 m3^6 mQ3^4 mU3^6-43 m3^4 mQ3^6 mU3^6-72 m3^8 mU3^8+97 m3^6 mQ3^2 mU3^8+35 m3^4 mQ3^4 mU3^8+19 m3^6 mU3^10-31 m3^4 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)))+(128 (34 m3^11-69 m3^9 mQ3^2+38 m3^7 mQ3^4-117 m3^9 mU3^2+229 m3^7 mQ3^2 mU3^2-129 m3^5 mQ3^4 mU3^2+71 m3^7 mU3^4-121 m3^5 mQ3^2 mU3^4+75 m3^3 mQ3^4 mU3^4-4 m3^5 mU3^6-7 m3^3 mQ3^2 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(1024 m3^4 mU3^2 Xt^2)/((m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2))) Log[mU3^2/M^2])+((16 (-128 m3^12+294 m3^10 mQ3^2-98 m3^8 mQ3^4-9 m3^6 mQ3^6-20 m3^4 mQ3^8+9 m3^2 mQ3^10+346 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+251 m3^6 mQ3^4 mU3^2+151 m3^4 mQ3^6 mU3^2-41 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-280 m3^8 mU3^4+1025 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4-41 m3^2 mQ3^6 mU3^4+mQ3^8 mU3^4+13 m3^6 mU3^6-391 m3^4 mQ3^2 mU3^6+167 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6+37 m3^4 mU3^8+34 m3^2 mQ3^2 mU3^8-23 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))-(128 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4) Xt)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) PolyLog[2,1-m3^2/mQ3^2]+(-24 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2)))-(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-m3^2/msq^2]+((80 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(-m3^2+mQ3^2)^3-(1280 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-msq^2/mQ3^2]+(-((16 (-128 m3^12+346 m3^10 mQ3^2-280 m3^8 mQ3^4+13 m3^6 mQ3^6+37 m3^4 mQ3^8+294 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+1025 m3^6 mQ3^4 mU3^2-391 m3^4 mQ3^6 mU3^2+34 m3^2 mQ3^8 mU3^2-98 m3^8 mU3^4+251 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4+167 m3^2 mQ3^6 mU3^4-23 mQ3^8 mU3^4-9 m3^6 mU3^6+151 m3^4 mQ3^2 mU3^6-41 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6-20 m3^4 mU3^8-41 m3^2 mQ3^2 mU3^8+mQ3^4 mU3^8+9 m3^2 mU3^10+3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)))-(128 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) PolyLog[2,1-m3^2/mU3^2]+(-((8 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3))+(64 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)) PolyLog[2,1-mQ3^2/mU3^2]+((80 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(-m3^2+mU3^2)^3-(1280 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) PolyLog[2,1-msq^2/mU3^2])+Log[mQ3^2/M^2] ((4 (8 m3^12 mQ3^4-152 m3^10 mQ3^6+216 m3^8 mQ3^8-72 m3^6 mQ3^10+17 m3^10 mQ3^4 mU3^2+631 m3^8 mQ3^6 mU3^2-1307 m3^6 mQ3^8 mU3^2+603 m3^4 mQ3^10 mU3^2+360 m3^8 mQ3^4 msq^2 mU3^2-840 m3^6 mQ3^6 msq^2 mU3^2+420 m3^4 mQ3^8 msq^2 mU3^2-180 m3^2 mQ3^10 msq^2 mU3^2-8 m3^12 mU3^4+631 m3^10 mQ3^2 mU3^4-2143 m3^8 mQ3^4 mU3^4+1943 m3^6 mQ3^6 mU3^4+335 m3^4 mQ3^8 mU3^4-614 m3^2 mQ3^10 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+540 m3^6 mQ3^4 msq^2 mU3^4+600 m3^2 mQ3^8 msq^2 mU3^4-60 mQ3^10 msq^2 mU3^4+16 m3^10 mU3^6-744 m3^8 mQ3^2 mU3^6+2281 m3^6 mQ3^4 mU3^6-2391 m3^4 mQ3^6 mU3^6+563 m3^2 mQ3^8 mU3^6+179 mQ3^10 mU3^6+300 m3^6 mQ3^2 msq^2 mU3^6-240 m3^4 mQ3^4 msq^2 mU3^6-720 m3^2 mQ3^6 msq^2 mU3^6-60 mQ3^8 msq^2 mU3^6-8 m3^8 mU3^8+227 m3^6 mQ3^2 mU3^8-577 m3^4 mQ3^4 mU3^8+533 m3^2 mQ3^6 mU3^8-191 mQ3^8 mU3^8-180 m3^4 mQ3^2 msq^2 mU3^8+300 m3^2 mQ3^4 msq^2 mU3^8+120 mQ3^6 msq^2 mU3^8-18 m3^4 mQ3^2 mU3^10+30 m3^2 mQ3^4 mU3^10+12 mQ3^6 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(64 (16 m3^7 mQ3^2-16 m3^5 mQ3^4+22 m3^7 mU3^2-112 m3^5 mQ3^2 mU3^2+87 m3^3 mQ3^4 mU3^2+30 m3 mQ3^4 msq^2 mU3^2-20 m3^5 mU3^4+100 m3^3 mQ3^2 mU3^4-74 m3 mQ3^4 mU3^4+30 m3^3 msq^2 mU3^4-60 m3 mQ3^2 msq^2 mU3^4+3 m3^3 mU3^6-6 m3 mQ3^2 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2) (mQ3^2-mU3^2))-(8 (8 m3^10 mQ3^2-144 m3^8 mQ3^4+72 m3^6 mQ3^6-8 m3^10 mU3^2+695 m3^8 mQ3^2 mU3^2-314 m3^6 mQ3^4 mU3^2-3 m3^4 mQ3^6 mU3^2-18 m3^2 mQ3^8 mU3^2-360 m3^6 mQ3^2 msq^2 mU3^2+300 m3^4 mQ3^4 msq^2 mU3^2-180 m3^2 mQ3^6 msq^2 mU3^2+16 m3^8 mU3^4-898 m3^6 mQ3^2 mU3^4+40 m3^4 mQ3^4 mU3^4+312 m3^2 mQ3^6 mU3^4-6 mQ3^8 mU3^4+300 m3^4 mQ3^2 msq^2 mU3^4+240 m3^2 mQ3^4 msq^2 mU3^4-60 mQ3^6 msq^2 mU3^4-8 m3^6 mU3^6+309 m3^4 mQ3^2 mU3^6+208 m3^2 mQ3^4 mU3^6-237 mQ3^6 mU3^6-180 m3^2 mQ3^2 msq^2 mU3^6-60 mQ3^4 msq^2 mU3^6-18 m3^2 mQ3^2 mU3^8-6 mQ3^4 mU3^8) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+((128 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))+(128 m3 (3 mQ3^2+mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(512 (4 m3^5 mQ3^2-4 m3^3 mQ3^4-3 m3^5 mU3^2-14 m3^3 mQ3^2 mU3^2+9 m3 mQ3^4 mU3^2))/(3 (m3^2-mQ3^2)^2 mU3^2 (mQ3^2-mU3^2)^2)) Xt^3+(4 (8 m3^12 mQ3^4-152 m3^10 mQ3^6+216 m3^8 mQ3^8-72 m3^6 mQ3^10+176 m3^12 mQ3^2 mU3^2-4777 m3^10 mQ3^4 mU3^2+7235 m3^8 mQ3^6 mU3^2-1833 m3^6 mQ3^8 mU3^2-991 m3^4 mQ3^10 mU3^2-18 m3^2 mQ3^12 mU3^2-360 m3^8 mQ3^4 msq^2 mU3^2+1020 m3^6 mQ3^6 msq^2 mU3^2-360 m3^4 mQ3^8 msq^2 mU3^2+180 m3^2 mQ3^10 msq^2 mU3^2-8 m3^12 mU3^4+655 m3^10 mQ3^2 mU3^4+10117 m3^8 mQ3^4 mU3^4-17603 m3^6 mQ3^6 mU3^4+5965 m3^4 mQ3^8 mU3^4+1552 m3^2 mQ3^10 mU3^4-6 mQ3^12 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+600 m3^6 mQ3^4 msq^2 mU3^4-1380 m3^4 mQ3^6 msq^2 mU3^4-360 m3^2 mQ3^8 msq^2 mU3^4+60 mQ3^10 msq^2 mU3^4+16 m3^10 mU3^6-1198 m3^8 mQ3^2 mU3^6-7009 m3^6 mQ3^4 mU3^6+13499 m3^4 mQ3^6 mU3^6-5635 m3^2 mQ3^8 mU3^6-441 mQ3^10 mU3^6+300 m3^6 mQ3^2 msq^2 mU3^6+480 m3^4 mQ3^4 msq^2 mU3^6+420 m3^2 mQ3^6 msq^2 mU3^6+240 mQ3^8 msq^2 mU3^6-8 m3^8 mU3^8+481 m3^6 mQ3^2 mU3^8+1421 m3^4 mQ3^4 mU3^8-2733 m3^2 mQ3^6 mU3^8+1191 mQ3^8 mU3^8-180 m3^4 mQ3^2 msq^2 mU3^8-240 m3^2 mQ3^4 msq^2 mU3^8-60 mQ3^6 msq^2 mU3^8-18 m3^4 mQ3^2 mU3^10-24 m3^2 mQ3^4 mU3^10-6 mQ3^6 mU3^10) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^3 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(64 (16 m3^7 mQ3^2-16 m3^5 mQ3^4+48 m3^7 mU3^2+172 m3^5 mQ3^2 mU3^2-185 m3^3 mQ3^4 mU3^2+3 m3 mQ3^6 mU3^2-30 m3^3 mQ3^2 msq^2 mU3^2-30 m3 mQ3^4 msq^2 mU3^2-62 m3^5 mU3^4-166 m3^3 mQ3^2 mU3^4+184 m3 mQ3^4 mU3^4+30 m3^3 msq^2 mU3^4+30 m3 mQ3^2 msq^2 mU3^4+3 m3^3 mU3^6+3 m3 mQ3^2 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+((16 (87 m3^10-230 m3^8 mQ3^2+212 m3^6 mQ3^4-69 m3^4 mQ3^6+60 m3^8 msq^2-255 m3^6 mQ3^2 msq^2+120 m3^4 mQ3^4 msq^2-45 m3^2 mQ3^6 msq^2+45 m3^6 msq^4+15 m3^4 mQ3^2 msq^4-169 m3^8 mU3^2+445 m3^6 mQ3^2 mU3^2-409 m3^4 mQ3^4 mU3^2+133 m3^2 mQ3^6 mU3^2-15 m3^6 msq^2 mU3^2+195 m3^4 mQ3^2 msq^2 mU3^2+75 m3^2 mQ3^4 msq^2 mU3^2-15 mQ3^6 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-30 m3^2 mQ3^2 msq^4 mU3^2+82 m3^6 mU3^4-215 m3^4 mQ3^2 mU3^4+197 m3^2 mQ3^4 mU3^4-64 mQ3^6 mU3^4+15 m3^4 msq^2 mU3^4-120 m3^2 mQ3^2 msq^2 mU3^4-15 mQ3^4 msq^2 mU3^4+45 m3^2 msq^4 mU3^4+15 mQ3^2 msq^4 mU3^4))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2)+(-((640 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2)))-(192 m3)/(mQ3^2-mU3^2)-(96 m3 mQ3^2)/((m3^2-mQ3^2) (mQ3^2-mU3^2))+(320 (m3^3 mQ3^2-m3 mQ3^4+12 m3 mQ3^2 msq^2-6 m3 msq^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))) Xt+(32 (87 m3^8-160 m3^6 mQ3^2+73 m3^4 mQ3^4+90 m3^6 msq^2-75 m3^4 mQ3^2 msq^2+45 m3^2 mQ3^4 msq^2-169 m3^6 mU3^2+310 m3^4 mQ3^2 mU3^2-141 m3^2 mQ3^4 mU3^2-75 m3^4 msq^2 mU3^2-60 m3^2 mQ3^2 msq^2 mU3^2+15 mQ3^4 msq^2 mU3^2+82 m3^4 mU3^4-150 m3^2 mQ3^2 mU3^4+68 mQ3^4 mU3^4+45 m3^2 msq^2 mU3^4+15 mQ3^2 msq^2 mU3^4) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((1280 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)))+(192 m3 (3 mQ3^2+mU3^2))/(mQ3^2-mU3^2)^3) Xt^3+((16 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+60 m3^4 msq^2+180 m3^2 mQ3^2 msq^2-90 m3^2 msq^4-30 mQ3^2 msq^4))/(3 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^2)-(8 (mQ3^2+mU3^2) (111 m3^8-221 m3^6 mQ3^2+110 m3^4 mQ3^4+180 m3^6 msq^2-150 m3^4 mQ3^2 msq^2+90 m3^2 mQ3^4 msq^2-221 m3^6 mU3^2+440 m3^4 mQ3^2 mU3^2-219 m3^2 mQ3^4 mU3^2-150 m3^4 msq^2 mU3^2-120 m3^2 mQ3^2 msq^2 mU3^2+30 mQ3^4 msq^2 mU3^2+110 m3^4 mU3^4-219 m3^2 mQ3^2 mU3^4+109 mQ3^4 mU3^4+90 m3^2 msq^2 mU3^4+30 mQ3^2 msq^2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)-(24 (4 m3^8-3 mQ3^4 mU3^2 (5 mQ3^2+mU3^2)+m3^6 (7 mQ3^2+3 mU3^2)-m3^4 (29 mQ3^4+15 mQ3^2 mU3^2+6 mU3^4)+m3^2 (16 mQ3^6+31 mQ3^4 mU3^2+7 mQ3^2 mU3^4)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^4+(-((192 m3 mQ3^2)/((m3^2-mQ3^2) (mQ3^2-mU3^2)^3))-(64 (m3^3 mQ3^2-m3 mQ3^4+120 m3 mQ3^2 msq^2-60 m3 msq^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)+(640 m3 msq^2 (mQ3^2+mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)) Xt^5) Log[msq^2/M^2]+(-24 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))-24 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-1/(mQ3^2-mU3^2) 48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^2-(48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^3)/(mQ3^2-mU3^2)+1/(mQ3^2-mU3^2)^3 24 (mQ3^2+mU3^2) (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^4+(24 (mQ3^2+mU3^2) ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^5)/(mQ3^2-mU3^2)^3) Log[msq^2/M^2]^2+(-((8 (342 m3^12-246 m3^10 mQ3^2-219 m3^8 mQ3^4+198 m3^6 mQ3^6-20 m3^4 mQ3^8+9 m3^2 mQ3^10-90 m3^8 mQ3^2 msq^2-30 m3^6 mQ3^4 msq^2-266 m3^10 mU3^2-734 m3^8 mQ3^2 mU3^2+1528 m3^6 mQ3^4 mU3^2-703 m3^4 mQ3^6 mU3^2-32 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-90 m3^8 msq^2 mU3^2+540 m3^6 mQ3^2 msq^2 mU3^2-180 m3^4 mQ3^4 msq^2 mU3^2+90 m3^2 mQ3^6 msq^2 mU3^2-193 m3^8 mU3^4+1626 m3^6 mQ3^2 mU3^4-2035 m3^4 mQ3^4 mU3^4+826 m3^2 mQ3^6 mU3^4+4 mQ3^8 mU3^4-30 m3^6 msq^2 mU3^4-180 m3^4 mQ3^2 msq^2 mU3^4-180 m3^2 mQ3^4 msq^2 mU3^4+30 mQ3^6 msq^2 mU3^4+160 m3^6 mU3^6-793 m3^4 mQ3^2 mU3^6+834 m3^2 mQ3^4 mU3^6-301 mQ3^6 mU3^6+90 m3^2 mQ3^2 msq^2 mU3^6+30 mQ3^4 msq^2 mU3^6+9 m3^4 mU3^8+3 m3^2 mQ3^2 mU3^8))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3))+(32 (2 m3^9 mQ3^2+214 m3^7 mQ3^4-206 m3^5 mQ3^6-4 m3^3 mQ3^8+6 m3 mQ3^10-60 m3^5 mQ3^4 msq^2-2 m3^9 mU3^2-467 m3^7 mQ3^2 mU3^2+250 m3^5 mQ3^4 mU3^2+117 m3^3 mQ3^6 mU3^2-20 m3 mQ3^8 mU3^2+120 m3^5 mQ3^2 msq^2 mU3^2+60 m3 mQ3^6 msq^2 mU3^2+221 m3^7 mU3^4+271 m3^5 mQ3^2 mU3^4-317 m3^3 mQ3^4 mU3^4+39 m3 mQ3^6 mU3^4-60 m3^5 msq^2 mU3^4-120 m3 mQ3^4 msq^2 mU3^4-219 m3^5 mU3^6+102 m3^3 mQ3^2 mU3^6+7 m3 mQ3^4 mU3^6+60 m3 mQ3^2 msq^2 mU3^6+6 m3^3 mU3^8) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)+(16 (462 m3^12 mQ3^2-384 m3^10 mQ3^4-310 m3^8 mQ3^6+275 m3^6 mQ3^8+9 m3^4 mQ3^10-90 m3^8 mQ3^4 msq^2-30 m3^6 mQ3^6 msq^2+474 m3^12 mU3^2-3694 m3^10 mQ3^2 mU3^2+4594 m3^8 mQ3^4 mU3^2-1226 m3^6 mQ3^6 mU3^2-353 m3^4 mQ3^8 mU3^2-15 m3^2 mQ3^10 mU3^2+180 m3^8 mQ3^2 msq^2 mU3^2+30 m3^6 mQ3^4 msq^2 mU3^2+360 m3^4 mQ3^6 msq^2 mU3^2-90 m3^2 mQ3^8 msq^2 mU3^2-418 m3^10 mU3^4+4658 m3^8 mQ3^2 mU3^4-6424 m3^6 mQ3^4 mU3^4+2394 m3^4 mQ3^6 mU3^4+156 m3^2 mQ3^8 mU3^4-6 mQ3^10 mU3^4-90 m3^8 msq^2 mU3^4+30 m3^6 mQ3^2 msq^2 mU3^4-720 m3^4 mQ3^4 msq^2 mU3^4+90 m3^2 mQ3^6 msq^2 mU3^4-30 mQ3^8 msq^2 mU3^4-278 m3^8 mU3^6-1306 m3^6 mQ3^2 mU3^6+2450 m3^4 mQ3^4 mU3^6-1168 m3^2 mQ3^6 mU3^6+22 mQ3^8 mU3^6-30 m3^6 msq^2 mU3^6+360 m3^4 mQ3^2 msq^2 mU3^6+90 m3^2 mQ3^4 msq^2 mU3^6+60 mQ3^6 msq^2 mU3^6+265 m3^6 mU3^8-325 m3^4 mQ3^2 mU3^8+130 m3^2 mQ3^4 mU3^8+30 mQ3^6 mU3^8-90 m3^2 mQ3^2 msq^2 mU3^8-30 mQ3^4 msq^2 mU3^8+9 m3^4 mU3^10-15 m3^2 mQ3^2 mU3^10-6 mQ3^4 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(64 (63 m3^9 mQ3^2+124 m3^7 mQ3^4-171 m3^5 mQ3^6+6 m3^3 mQ3^8-60 m3^5 mQ3^4 msq^2-59 m3^9 mU3^2-8 m3^7 mQ3^2 mU3^2-669 m3^5 mQ3^4 mU3^2+612 m3^3 mQ3^6 mU3^2-12 m3 mQ3^8 mU3^2+240 m3^3 mQ3^4 msq^2 mU3^2-60 m3 mQ3^6 msq^2 mU3^2-132 m3^7 mU3^4+689 m3^5 mQ3^2 mU3^4-8 m3^3 mQ3^4 mU3^4-337 m3 mQ3^6 mU3^4+60 m3^5 msq^2 mU3^4-240 m3^3 mQ3^2 msq^2 mU3^4+175 m3^5 mU3^6-620 m3^3 mQ3^2 mU3^6+341 m3 mQ3^4 mU3^6+60 m3 mQ3^2 msq^2 mU3^6-6 m3^3 mU3^8+12 m3 mQ3^2 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(8 (4 m3^14 mQ3^2-822 m3^12 mQ3^4+1054 m3^10 mQ3^6-8 m3^8 mQ3^8-207 m3^6 mQ3^10-67 m3^4 mQ3^12+18 m3^2 mQ3^14+90 m3^8 mQ3^6 msq^2+30 m3^6 mQ3^8 msq^2-4 m3^14 mU3^2-3128 m3^12 mQ3^2 mU3^2+10898 m3^10 mQ3^4 mU3^2-11670 m3^8 mQ3^6 mU3^2+3785 m3^6 mQ3^8 mU3^2+252 m3^4 mQ3^10 mU3^2-91 m3^2 mQ3^12 mU3^2+6 mQ3^14 mU3^2-90 m3^8 mQ3^4 msq^2 mU3^2-360 m3^4 mQ3^8 msq^2 mU3^2+90 m3^2 mQ3^10 msq^2 mU3^2-818 m3^12 mU3^4+10982 m3^10 mQ3^2 mU3^4-24928 m3^8 mQ3^4 mU3^4+21138 m3^6 mQ3^6 mU3^4-6833 m3^4 mQ3^8 mU3^4+557 m3^2 mQ3^10 mU3^4+2 mQ3^12 mU3^4-90 m3^8 mQ3^2 msq^2 mU3^4-60 m3^6 mQ3^4 msq^2 mU3^4+360 m3^4 mQ3^6 msq^2 mU3^4+30 mQ3^10 msq^2 mU3^4+1066 m3^10 mU3^6-11842 m3^8 mQ3^2 mU3^6+20522 m3^6 mQ3^4 mU3^6-12116 m3^4 mQ3^6 mU3^6+2260 m3^2 mQ3^8 mU3^6-210 mQ3^10 mU3^6+90 m3^8 msq^2 mU3^6+360 m3^4 mQ3^4 msq^2 mU3^6-180 m3^2 mQ3^6 msq^2 mU3^6-30 mQ3^8 msq^2 mU3^6-32 m3^8 mU3^8+4297 m3^6 mQ3^2 mU3^8-6555 m3^4 mQ3^4 mU3^8+2334 m3^2 mQ3^6 mU3^8+256 mQ3^8 mU3^8+30 m3^6 msq^2 mU3^8-360 m3^4 mQ3^2 msq^2 mU3^8-30 mQ3^6 msq^2 mU3^8-255 m3^6 mU3^10-112 m3^4 mQ3^2 mU3^10+475 m3^2 mQ3^4 mU3^10-220 mQ3^6 mU3^10+90 m3^2 mQ3^2 msq^2 mU3^10+30 mQ3^4 msq^2 mU3^10-9 m3^4 mU3^12+15 m3^2 mQ3^2 mU3^12+6 mQ3^4 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^4)+(64 (64 m3^9 mQ3^2+44 m3^7 mQ3^4-102 m3^5 mQ3^6+13 m3^3 mQ3^8-6 m3 mQ3^10-30 m3^5 mQ3^4 msq^2+64 m3^9 mU3^2+188 m3^7 mQ3^2 mU3^2-638 m3^5 mQ3^4 mU3^2+352 m3^3 mQ3^6 mU3^2+14 m3 mQ3^8 mU3^2-60 m3^5 mQ3^2 msq^2 mU3^2+120 m3^3 mQ3^4 msq^2 mU3^2-30 m3 mQ3^6 msq^2 mU3^2+44 m3^7 mU3^4-668 m3^5 mQ3^2 mU3^4+920 m3^3 mQ3^4 mU3^4-298 m3 mQ3^6 mU3^4-30 m3^5 msq^2 mU3^4+120 m3^3 mQ3^2 msq^2 mU3^4-60 m3 mQ3^4 msq^2 mU3^4-92 m3^5 mU3^6+372 m3^3 mQ3^2 mU3^6-268 m3 mQ3^4 mU3^6-30 m3 mQ3^2 msq^2 mU3^6+3 m3^3 mU3^8-6 m3 mQ3^2 mU3^8) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)+(5120 m3^2 mQ3^2 mU3^2 Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^4)+(1/(3 (m3^2-mQ3^2)^3) 8 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+60 m3^4 msq^2+180 m3^2 mQ3^2 msq^2-90 m3^2 msq^4-30 mQ3^2 msq^4)+1/(3 (m3^2-mU3^2)^3) 8 (60 m3^4 msq^2-90 m3^2 msq^4+2 m3^4 mU3^2+180 m3^2 msq^2 mU3^2-30 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6)+(24 (-2 mQ3^4 mU3^4+2 m3^6 (mQ3^2+mU3^2)+4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)-m3^4 (mQ3^4+8 mQ3^2 mU3^2+mU3^4)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(320 (m3^7-m3^5 mQ3^2+12 m3^5 msq^2-12 m3^3 msq^4+6 m3 mQ3^2 msq^4-m3^5 mU3^2+m3^3 mQ3^2 mU3^2-12 m3 mQ3^2 msq^2 mU3^2+6 m3 msq^4 mU3^2) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(-((16 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+60 m3^4 msq^2+180 m3^2 mQ3^2 msq^2-90 m3^2 msq^4-30 mQ3^2 msq^4))/(3 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2)))+(16 (60 m3^4 msq^2-90 m3^2 msq^4+2 m3^4 mU3^2+180 m3^2 msq^2 mU3^2-30 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6))/(3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+(48 (2 m3^8 (mQ3^2+mU3^2)+2 mQ3^4 mU3^4 (mQ3^2+mU3^2)-4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)^2+3 m3^4 (mQ3^2+mU3^2)^3-2 m3^6 (3 mQ3^4+2 mQ3^2 mU3^2+3 mU3^4)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Xt^2+((64 (m3^3 mQ3^2-m3 mQ3^4+120 m3 mQ3^2 msq^2-60 m3 msq^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)+(64 (-60 m3 msq^4+m3^3 mU3^2+120 m3 msq^2 mU3^2-m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)+(192 (-2 m3 mQ3^2 mU3^2+m3^3 (mQ3^2+mU3^2)))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^2)) Xt^3+((8 (2 m3^4 mQ3^2-3 m3^2 mQ3^4+mQ3^6+60 m3^4 msq^2+180 m3^2 mQ3^2 msq^2-90 m3^2 msq^4-30 mQ3^2 msq^4) (mQ3^2+mU3^2))/(3 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2)^3)-(8 (mQ3^2+mU3^2) (60 m3^4 msq^2-90 m3^2 msq^4+2 m3^4 mU3^2+180 m3^2 msq^2 mU3^2-30 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6))/(3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(24 (mQ3^2+mU3^2) (4 m3^8 (mQ3^2+mU3^2)+4 mQ3^4 mU3^4 (mQ3^2+mU3^2)-8 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)^2-2 m3^6 (5 mQ3^4+6 mQ3^2 mU3^2+5 mU3^4)+m3^4 (5 mQ3^6+19 mQ3^4 mU3^2+19 mQ3^2 mU3^4+5 mU3^6)))/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)) Xt^4+(-((32 (m3^3 mQ3^2-m3 mQ3^4+120 m3 mQ3^2 msq^2-60 m3 msq^4) (mQ3^2+mU3^2))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4))-(32 (mQ3^2+mU3^2) (-60 m3 msq^4+m3^3 mU3^2+120 m3 msq^2 mU3^2-m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)-(96 m3 (mQ3^2+mU3^2) (-2 mQ3^2 mU3^2+m3^2 (mQ3^2+mU3^2)))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^4)) Xt^5) Log[msq^2/M^2]) Log[mU3^2/M^2]+(-((16 (m3^4-18 m3^2 mQ3^2+9 mQ3^4) (3 m3^4-4 m3^2 mU3^2+2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2))+(8 (12 m3^10 mQ3^2-14 m3^8 mQ3^4+34 m3^6 mQ3^6-29 m3^4 mQ3^8+9 m3^2 mQ3^10-2 m3^10 mU3^2+70 m3^8 mQ3^2 mU3^2-194 m3^6 mQ3^4 mU3^2+122 m3^4 mQ3^6 mU3^2-35 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2+3 m3^8 mU3^4-87 m3^6 mQ3^2 mU3^4+184 m3^4 mQ3^4 mU3^4-68 m3^2 mQ3^6 mU3^4+4 mQ3^8 mU3^4-m3^6 mU3^6+29 m3^4 mQ3^2 mU3^6-60 m3^2 mQ3^4 mU3^6+20 mQ3^6 mU3^6))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(4 (128 m3^10+44 m3^8 mQ3^2-2 m3^6 mQ3^4+9 m3^4 mQ3^6-60 m3^6 mQ3^2 msq^2+90 m3^4 mQ3^2 msq^4-556 m3^8 mU3^2+64 m3^6 mQ3^2 mU3^2-33 m3^4 mQ3^4 mU3^2-6 m3^2 mQ3^6 mU3^2+60 m3^6 msq^2 mU3^2-120 m3^4 mQ3^2 msq^2 mU3^2-90 m3^4 msq^4 mU3^2-60 m3^2 mQ3^2 msq^4 mU3^2+706 m3^6 mU3^4-51 m3^4 mQ3^2 mU3^4+36 m3^2 mQ3^4 mU3^4-3 mQ3^6 mU3^4+120 m3^4 msq^2 mU3^4+180 m3^2 mQ3^2 msq^2 mU3^4+60 m3^2 msq^4 mU3^4-30 mQ3^2 msq^4 mU3^4-437 m3^4 mU3^6-38 m3^2 mQ3^2 mU3^6-mQ3^4 mU3^6-180 m3^2 msq^2 mU3^6+30 msq^4 mU3^6+136 m3^2 mU3^8+5 mQ3^2 mU3^8-mU3^10))/(3 (m3^2-mU3^2)^4 (-mQ3^2+mU3^2))+(8 (m3^4-34 m3^2 mU3^2+17 mU3^4) (-2 mQ3^4 mU3^4+2 m3^6 (mQ3^2+mU3^2)+4 m3^2 mQ3^2 mU3^2 (mQ3^2+mU3^2)-m3^4 (mQ3^4+8 mQ3^2 mU3^2+mU3^4)))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^4)-(32 (22 m3^11 mQ3^2-60 m3^9 mQ3^4+14 m3^7 mQ3^6-5 m3^5 mQ3^8+3 m3^3 mQ3^10-30 m3^7 mQ3^2 msq^4+60 m3^5 mQ3^4 msq^4-30 m3^3 mQ3^6 msq^4-22 m3^11 mU3^2-51 m3^9 mQ3^2 mU3^2+235 m3^7 mQ3^4 mU3^2-66 m3^5 mQ3^6 mU3^2-5 m3^3 mQ3^8 mU3^2-3 m3 mQ3^10 mU3^2+60 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+60 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2-30 m3^5 mQ3^2 msq^4 mU3^2-30 m3^3 mQ3^4 msq^4 mU3^2+30 m3 mQ3^6 msq^4 mU3^2+95 m3^9 mU3^4-56 m3^7 mQ3^2 mU3^4-263 m3^5 mQ3^4 mU3^4+106 m3^3 mQ3^6 mU3^4+10 m3 mQ3^8 mU3^4-60 m3^7 msq^2 mU3^4+60 m3^5 mQ3^2 msq^2 mU3^4+60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-30 m3^5 msq^4 mU3^4+60 m3^3 mQ3^2 msq^4 mU3^4-30 m3 mQ3^4 msq^4 mU3^4-129 m3^7 mU3^6+198 m3^5 mQ3^2 mU3^6+25 m3^3 mQ3^4 mU3^6-38 m3 mQ3^6 mU3^6+60 m3^5 msq^2 mU3^6-120 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+40 m3^5 mU3^8-65 m3^3 mQ3^2 mU3^8+15 m3 mQ3^4 mU3^8) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(8 (128 m3^16-844 m3^14 mQ3^2+1158 m3^12 mQ3^4-339 m3^10 mQ3^6-187 m3^8 mQ3^8+87 m3^6 mQ3^10-27 m3^4 mQ3^12-60 m3^12 mQ3^2 msq^2+180 m3^10 mQ3^4 msq^2-180 m3^8 mQ3^6 msq^2+60 m3^6 mQ3^8 msq^2+90 m3^10 mQ3^2 msq^4-270 m3^8 mQ3^4 msq^4+270 m3^6 mQ3^6 msq^4-90 m3^4 mQ3^8 msq^4-564 m3^14 mU3^2+3672 m3^12 mQ3^2 mU3^2-5047 m3^10 mQ3^4 mU3^2+1947 m3^8 mQ3^6 mU3^2+49 m3^6 mQ3^8 mU3^2+45 m3^4 mQ3^10 mU3^2+18 m3^2 mQ3^12 mU3^2+60 m3^12 msq^2 mU3^2-300 m3^10 mQ3^2 msq^2 mU3^2+540 m3^8 mQ3^4 msq^2 mU3^2-420 m3^6 mQ3^6 msq^2 mU3^2+120 m3^4 mQ3^8 msq^2 mU3^2-90 m3^10 msq^4 mU3^2+210 m3^8 mQ3^2 msq^4 mU3^2-90 m3^6 mQ3^4 msq^4 mU3^2-90 m3^4 mQ3^6 msq^4 mU3^2+60 m3^2 mQ3^8 msq^4 mU3^2+450 m3^12 mU3^4-4593 m3^10 mQ3^2 mU3^4+6259 m3^8 mQ3^4 mU3^4-1992 m3^6 mQ3^6 mU3^4-238 m3^4 mQ3^8 mU3^4-135 m3^2 mQ3^10 mU3^4+9 mQ3^12 mU3^4+120 m3^10 msq^2 mU3^4-180 m3^8 mQ3^2 msq^2 mU3^4-180 m3^6 mQ3^4 msq^2 mU3^4+420 m3^4 mQ3^6 msq^2 mU3^4-180 m3^2 mQ3^8 msq^2 mU3^4+60 m3^8 msq^4 mU3^4-210 m3^6 mQ3^2 msq^4 mU3^4+270 m3^4 mQ3^4 msq^4 mU3^4-150 m3^2 mQ3^6 msq^4 mU3^4+30 mQ3^8 msq^4 mU3^4+283 m3^10 mU3^6+1957 m3^8 mQ3^2 mU3^6-2924 m3^6 mQ3^4 mU3^6+424 m3^4 mQ3^6 mU3^6+497 m3^2 mQ3^8 mU3^6+3 mQ3^10 mU3^6-180 m3^8 msq^2 mU3^6+540 m3^6 mQ3^2 msq^2 mU3^6-540 m3^4 mQ3^4 msq^2 mU3^6+180 m3^2 mQ3^6 msq^2 mU3^6+30 m3^6 msq^4 mU3^6-90 m3^4 mQ3^2 msq^4 mU3^6+90 m3^2 mQ3^4 msq^4 mU3^6-30 mQ3^6 msq^4 mU3^6-376 m3^8 mU3^8-235 m3^6 mQ3^2 mU3^8+713 m3^4 mQ3^4 mU3^8-109 m3^2 mQ3^6 mU3^8-113 mQ3^8 mU3^8+87 m3^6 mU3^10+11 m3^4 mQ3^2 mU3^10-111 m3^2 mQ3^4 mU3^10+37 mQ3^6 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^2)-(64 (2 m3^13-19 m3^11 mQ3^2-18 m3^9 mQ3^4+15 m3^7 mQ3^6+15 m3^5 mQ3^8-9 m3^3 mQ3^10-30 m3^7 mQ3^2 msq^4+60 m3^5 mQ3^4 msq^4-30 m3^3 mQ3^6 msq^4-25 m3^11 mU3^2+210 m3^9 mQ3^2 mU3^2-78 m3^7 mQ3^4 mU3^2-91 m3^5 mQ3^6 mU3^2+15 m3^3 mQ3^8 mU3^2+9 m3 mQ3^10 mU3^2+60 m3^7 mQ3^2 msq^2 mU3^2-120 m3^5 mQ3^4 msq^2 mU3^2+60 m3^3 mQ3^6 msq^2 mU3^2+30 m3^7 msq^4 mU3^2-30 m3^5 mQ3^2 msq^4 mU3^2-30 m3^3 mQ3^4 msq^4 mU3^2+30 m3 mQ3^6 msq^4 mU3^2+126 m3^9 mU3^4-619 m3^7 mQ3^2 mU3^4+442 m3^5 mQ3^4 mU3^4+45 m3^3 mQ3^6 mU3^4-30 m3 mQ3^8 mU3^4-60 m3^7 msq^2 mU3^4+60 m3^5 mQ3^2 msq^2 mU3^4+60 m3^3 mQ3^4 msq^2 mU3^4-60 m3 mQ3^6 msq^2 mU3^4-30 m3^5 msq^4 mU3^4+60 m3^3 mQ3^2 msq^4 mU3^4-30 m3 mQ3^4 msq^4 mU3^4-190 m3^7 mU3^6+681 m3^5 mQ3^2 mU3^6-530 m3^3 mQ3^4 mU3^6+47 m3 mQ3^6 mU3^6+60 m3^5 msq^2 mU3^6-120 m3^3 mQ3^2 msq^2 mU3^6+60 m3 mQ3^4 msq^2 mU3^6+71 m3^5 mU3^8-205 m3^3 mQ3^2 mU3^8+136 m3 mQ3^4 mU3^8) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^3)-(4 (1160 m3^16 mQ3^2-3044 m3^14 mQ3^4+2462 m3^12 mQ3^6-507 m3^10 mQ3^8-155 m3^8 mQ3^10+87 m3^6 mQ3^12-27 m3^4 mQ3^14-60 m3^12 mQ3^4 msq^2+180 m3^10 mQ3^6 msq^2-180 m3^8 mQ3^8 msq^2+60 m3^6 mQ3^10 msq^2+90 m3^10 mQ3^4 msq^4-270 m3^8 mQ3^6 msq^4+270 m3^6 mQ3^8 msq^4-90 m3^4 mQ3^10 msq^4+120 m3^16 mU3^2-5384 m3^14 mQ3^2 mU3^2+14886 m3^12 mQ3^4 mU3^2-13282 m3^10 mQ3^6 mU3^2+3584 m3^8 mQ3^8 mU3^2+136 m3^6 mQ3^10 mU3^2+18 m3^4 mQ3^12 mU3^2+18 m3^2 mQ3^14 mU3^2-120 m3^10 mQ3^4 msq^2 mU3^2+360 m3^8 mQ3^6 msq^2 mU3^2-360 m3^6 mQ3^8 msq^2 mU3^2+120 m3^4 mQ3^10 msq^2 mU3^2-60 m3^8 mQ3^4 msq^4 mU3^2+180 m3^6 mQ3^6 msq^4 mU3^2-180 m3^4 mQ3^8 msq^4 mU3^2+60 m3^2 mQ3^10 msq^4 mU3^2-532 m3^14 mU3^4+9226 m3^12 mQ3^2 mU3^4-26136 m3^10 mQ3^4 mU3^4+24958 m3^8 mQ3^6 mU3^4-7207 m3^6 mQ3^8 mU3^4-321 m3^4 mQ3^10 mU3^4-117 m3^2 mQ3^12 mU3^4+9 mQ3^14 mU3^4+60 m3^12 msq^2 mU3^4-180 m3^10 mQ3^2 msq^2 mU3^4+360 m3^8 mQ3^4 msq^2 mU3^4-600 m3^6 mQ3^6 msq^2 mU3^4+540 m3^4 mQ3^8 msq^2 mU3^4-180 m3^2 mQ3^10 msq^2 mU3^4-90 m3^10 msq^4 mU3^4+270 m3^8 mQ3^2 msq^4 mU3^4-300 m3^6 mQ3^4 msq^4 mU3^4+180 m3^4 mQ3^6 msq^4 mU3^4-90 m3^2 mQ3^8 msq^4 mU3^4+30 mQ3^10 msq^4 mU3^4-206 m3^12 mU3^6-3654 m3^10 mQ3^2 mU3^6+15560 m3^8 mQ3^4 mU3^6-18052 m3^6 mQ3^6 mU3^6+5850 m3^4 mQ3^8 mU3^6+490 m3^2 mQ3^10 mU3^6+12 mQ3^12 mU3^6+120 m3^10 msq^2 mU3^6-360 m3^8 mQ3^2 msq^2 mU3^6+360 m3^6 mQ3^4 msq^2 mU3^6-120 m3^4 mQ3^6 msq^2 mU3^6+60 m3^8 msq^4 mU3^6-180 m3^6 mQ3^2 msq^4 mU3^6+180 m3^4 mQ3^4 msq^4 mU3^6-60 m3^2 mQ3^6 msq^4 mU3^6+1595 m3^10 mU3^8-4427 m3^8 mQ3^2 mU3^8+2513 m3^6 mQ3^4 mU3^8+2185 m3^4 mQ3^6 mU3^8-1604 m3^2 mQ3^8 mU3^8-142 mQ3^10 mU3^8-180 m3^8 msq^2 mU3^8+540 m3^6 mQ3^2 msq^2 mU3^8-540 m3^4 mQ3^4 msq^2 mU3^8+180 m3^2 mQ3^6 msq^2 mU3^8+30 m3^6 msq^4 mU3^8-90 m3^4 mQ3^2 msq^4 mU3^8+90 m3^2 mQ3^4 msq^4 mU3^8-30 mQ3^6 msq^4 mU3^8-1120 m3^8 mU3^10+3172 m3^6 mQ3^2 mU3^10-3748 m3^4 mQ3^4 mU3^10+1740 m3^2 mQ3^6 mU3^10-140 mQ3^8 mU3^10+151 m3^6 mU3^12-117 m3^4 mQ3^2 mU3^12-15 m3^2 mQ3^4 mU3^12+5 mQ3^6 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^4 (mQ3^2-mU3^2)^4)-(32 (14 m3^11 mQ3^4+6 m3^9 mQ3^6-8 m3^7 mQ3^8-15 m3^5 mQ3^10+9 m3^3 mQ3^12+30 m3^7 mQ3^4 msq^4-60 m3^5 mQ3^6 msq^4+30 m3^3 mQ3^8 msq^4-32 m3^11 mQ3^2 mU3^2-155 m3^9 mQ3^4 mU3^2+177 m3^7 mQ3^6 mU3^2+23 m3^5 mQ3^8 mU3^2-6 m3^3 mQ3^10 mU3^2-9 m3 mQ3^12 mU3^2-60 m3^7 mQ3^4 msq^2 mU3^2+120 m3^5 mQ3^6 msq^2 mU3^2-60 m3^3 mQ3^8 msq^2 mU3^2-30 m3^5 mQ3^4 msq^4 mU3^2+60 m3^3 mQ3^6 msq^4 mU3^2-30 m3 mQ3^8 msq^4 mU3^2+18 m3^11 mU3^4-182 m3^9 mQ3^2 mU3^4+719 m3^7 mQ3^4 mU3^4-621 m3^5 mQ3^6 mU3^4+25 m3^3 mQ3^8 mU3^4+21 m3 mQ3^10 mU3^4+60 m3^5 mQ3^4 msq^2 mU3^4-120 m3^3 mQ3^6 msq^2 mU3^4+60 m3 mQ3^8 msq^2 mU3^4-30 m3^7 msq^4 mU3^4+60 m3^5 mQ3^2 msq^4 mU3^4-30 m3^3 mQ3^4 msq^4 mU3^4-149 m3^9 mU3^6+783 m3^7 mQ3^2 mU3^6-1297 m3^5 mQ3^4 mU3^6+731 m3^3 mQ3^6 mU3^6-56 m3 mQ3^8 mU3^6+60 m3^7 msq^2 mU3^6-120 m3^5 mQ3^2 msq^2 mU3^6+60 m3^3 mQ3^4 msq^2 mU3^6+30 m3^5 msq^4 mU3^6-60 m3^3 mQ3^2 msq^4 mU3^6+30 m3 mQ3^4 msq^4 mU3^6+249 m3^7 mU3^8-868 m3^5 mQ3^2 mU3^8+894 m3^3 mQ3^4 mU3^8-261 m3 mQ3^6 mU3^8-60 m3^5 msq^2 mU3^8+120 m3^3 mQ3^2 msq^2 mU3^8-60 m3 mQ3^4 msq^2 mU3^8-102 m3^5 mU3^10+267 m3^3 mQ3^2 mU3^10-175 m3 mQ3^4 mU3^10) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^5)+(1280 m3^2 (mQ3^2+mU3^2) (2 m3^2 mQ3^2 mU3^2+m3^2 mU3^4-3 mQ3^2 mU3^4) Xt^6)/(3 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^5)) Log[mU3^2/M^2]^2+(-((64 m3^4 (m3^4-18 m3^2 mQ3^2+9 mQ3^4))/(3 (m3^2-mQ3^2)^4))+(8 (-128 m3^12+294 m3^10 mQ3^2-98 m3^8 mQ3^4-9 m3^6 mQ3^6-20 m3^4 mQ3^8+9 m3^2 mQ3^10+346 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+251 m3^6 mQ3^4 mU3^2+151 m3^4 mQ3^6 mU3^2-41 m3^2 mQ3^8 mU3^2+3 mQ3^10 mU3^2-280 m3^8 mU3^4+1025 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4-41 m3^2 mQ3^6 mU3^4+mQ3^8 mU3^4+13 m3^6 mU3^6-391 m3^4 mQ3^2 mU3^6+167 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6+37 m3^4 mU3^8+34 m3^2 mQ3^2 mU3^8-23 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2))+((256 (m3^7-26 m3^5 mQ3^2+9 m3^3 mQ3^4))/(3 (m3^2-mQ3^2)^3 (mQ3^2-mU3^2))-(64 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-(16 (128 m3^12-806 m3^10 mQ3^2+98 m3^8 mQ3^4+9 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-346 m3^10 mU3^2+2438 m3^8 mQ3^2 mU3^2-251 m3^6 mQ3^4 mU3^2-151 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+280 m3^8 mU3^4-2561 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4+41 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-13 m3^6 mU3^6+903 m3^4 mQ3^2 mU3^6-167 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6-37 m3^4 mU3^8-34 m3^2 mQ3^2 mU3^8+23 mQ3^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+((128 m3 (m3^4-18 m3^2 mQ3^2+9 mQ3^4) (2 m3^2-mQ3^2-mU3^2))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)+(128 (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)) Xt^3+(8 (1160 m3^12 mQ3^2-1978 m3^10 mQ3^4+754 m3^8 mQ3^6-27 m3^6 mQ3^8+20 m3^4 mQ3^10-9 m3^2 mQ3^12+120 m3^12 mU3^2-4104 m3^10 mQ3^2 mU3^2+6492 m3^8 mQ3^4 mU3^2-2210 m3^6 mQ3^6 mU3^2-23 m3^4 mQ3^8 mU3^2+32 m3^2 mQ3^10 mU3^2-3 mQ3^12 mU3^2-318 m3^10 mU3^4+5310 m3^8 mQ3^2 mU3^4-7612 m3^6 mQ3^4 mU3^4+2234 m3^4 mQ3^6 mU3^4-26 m3^2 mQ3^8 mU3^4-4 mQ3^10 mU3^4+244 m3^8 mU3^6-2958 m3^6 mQ3^2 mU3^6+3704 m3^4 mQ3^4 mU3^6-782 m3^2 mQ3^6 mU3^6+16 mQ3^8 mU3^6+7 m3^6 mU3^8+506 m3^4 mQ3^2 mU3^8-533 m3^2 mQ3^4 mU3^8+4 mQ3^6 mU3^8-41 m3^4 mU3^10+38 m3^2 mQ3^2 mU3^10-13 mQ3^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^4)+(-((1024 m3 mQ3^2 (-2 m3^2+mQ3^2+mU3^2))/(3 (m3^2-mQ3^2) (mQ3^2-mU3^2)^4))-(64 (mQ3^2+mU3^2) (18 m3^5+m3^3 mQ3^2+3 m3 mQ3^4-37 m3^3 mU3^2-7 m3 mQ3^2 mU3^2+22 m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) Xt^5) PolyLog[2,1-m3^2/mQ3^2]+(-12 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2)))-(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-1/(mQ3^2-mU3^2) 24 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^2-(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^3)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+1/(mQ3^2-mU3^2)^3 12 (2 m3^2-2 msq^2) (mQ3^2+mU3^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^4+(640 (m3^2-msq^2) (mQ3^2+mU3^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/msq^2]+((40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(-m3^2+mQ3^2)^3-(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(80 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) Xt^2)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2))-(1280 m3 (mQ3^2-msq^2)^2 Xt^3)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)-(40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) (mQ3^2+mU3^2) Xt^4)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2)^3)+(640 m3 (mQ3^2-msq^2)^2 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-msq^2/mQ3^2]+((8 (128 m3^12-354 m3^10 mQ3^2+432 m3^8 mQ3^4-229 m3^6 mQ3^6+35 m3^4 mQ3^8-286 m3^10 mU3^2+750 m3^8 mQ3^2 mU3^2-809 m3^6 mQ3^4 mU3^2+319 m3^4 mQ3^6 mU3^2-34 m3^2 mQ3^8 mU3^2+98 m3^8 mU3^4-251 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4-167 m3^2 mQ3^6 mU3^4+23 mQ3^8 mU3^4+9 m3^6 mU3^6-151 m3^4 mQ3^2 mU3^6+41 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-(64 (22 m3^9-77 m3^7 mQ3^2+26 m3^5 mQ3^4-39 m3^7 mU3^2+139 m3^5 mQ3^2 mU3^2-80 m3^3 mQ3^4 mU3^2+19 m3^5 mU3^4-23 m3^3 mQ3^2 mU3^4+22 m3 mQ3^4 mU3^4-5 m3^3 mU3^6-7 m3 mQ3^2 mU3^6+3 m3 mU3^8) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(16 (128 m3^12-858 m3^10 mQ3^2+1304 m3^8 mQ3^4-525 m3^6 mQ3^6-37 m3^4 mQ3^8-294 m3^10 mU3^2+1414 m3^8 mQ3^2 mU3^2-2049 m3^6 mQ3^4 mU3^2+903 m3^4 mQ3^6 mU3^2-34 m3^2 mQ3^8 mU3^2+98 m3^8 mU3^4-251 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4-167 m3^2 mQ3^6 mU3^4+23 mQ3^8 mU3^4+9 m3^6 mU3^6-151 m3^4 mQ3^2 mU3^6+41 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)-(128 (2 m3^7-19 m3^5 mQ3^2-m3^3 mQ3^4+13 m3 mQ3^6-19 m3^5 mU3^2+56 m3^3 mQ3^2 mU3^2-38 m3 mQ3^4 mU3^2-m3^3 mU3^4+10 m3 mQ3^2 mU3^4-3 m3 mU3^6) Xt^3)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^3)-(8 (1160 m3^12 mQ3^2-3062 m3^10 mQ3^4+2620 m3^8 mQ3^6-705 m3^6 mQ3^8-m3^4 mQ3^10+120 m3^12 mU3^2-3064 m3^10 mQ3^2 mU3^2+7422 m3^8 mQ3^4 mU3^2-6158 m3^6 mQ3^6 mU3^2+1738 m3^4 mQ3^8 mU3^2-106 m3^2 mQ3^10 mU3^2-274 m3^10 mU3^4+2676 m3^8 mQ3^2 mU3^4-5500 m3^6 mQ3^4 mU3^4+3992 m3^4 mQ3^6 mU3^4-893 m3^2 mQ3^8 mU3^4+59 mQ3^10 mU3^4+82 m3^8 mU3^6-450 m3^6 mQ3^2 mU3^6+858 m3^4 mQ3^4 mU3^6-494 m3^2 mQ3^6 mU3^6+4 mQ3^8 mU3^6+13 m3^6 mU3^8-207 m3^4 mQ3^2 mU3^8+190 m3^2 mQ3^4 mU3^8-56 mQ3^6 mU3^8+20 m3^4 mU3^10+32 m3^2 mQ3^2 mU3^10-4 mQ3^4 mU3^10-9 m3^2 mU3^12-3 mQ3^2 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^4)-(64 (14 m3^5 mQ3^2-11 m3^3 mQ3^4-6 m3 mQ3^6-18 m3^5 mU3^2+20 m3^3 mQ3^2 mU3^2+m3 mQ3^4 mU3^2-m3^3 mU3^4+4 m3 mQ3^2 mU3^4-3 m3 mU3^6) Xt^5)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-m3^2/mU3^2]+(-((4 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3))+(32 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(16 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)+(128 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(8 (mQ3^2+mU3^2) (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(64 (mQ3^2+mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-mQ3^2/mU3^2]+((40 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(-m3^2+mU3^2)^3-(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(160 (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^2)/((mQ3^2-mU3^2) (-m3^2+mU3^2)^3)+(1280 m3 (-msq^2+mU3^2)^2 Xt^3)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)-(80 (mQ3^2+mU3^2) (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^4)/((mQ3^2-mU3^2)^3 (-m3^2+mU3^2)^3)-(640 m3 (mQ3^2+mU3^2) (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) PolyLog[2,1-msq^2/mU3^2])+Log[mU3^2/M^2] ((4 (-4 m3^12 mQ3^4+8 m3^10 mQ3^6-4 m3^8 mQ3^8+611 m3^10 mQ3^4 mU3^2-704 m3^8 mQ3^6 mU3^2+207 m3^6 mQ3^8 mU3^2-18 m3^4 mQ3^10 mU3^2-360 m3^8 mQ3^4 msq^2 mU3^2+300 m3^6 mQ3^6 msq^2 mU3^2-180 m3^4 mQ3^8 msq^2 mU3^2+4 m3^12 mU3^4+33 m3^10 mQ3^2 mU3^4-2127 m3^8 mQ3^4 mU3^4+2217 m3^6 mQ3^6 mU3^4-541 m3^4 mQ3^8 mU3^4+30 m3^2 mQ3^10 mU3^4+360 m3^8 mQ3^2 msq^2 mU3^4+540 m3^6 mQ3^4 msq^2 mU3^4-240 m3^4 mQ3^6 msq^2 mU3^4+300 m3^2 mQ3^8 msq^2 mU3^4-140 m3^10 mU3^6+583 m3^8 mQ3^2 mU3^6+1975 m3^6 mQ3^4 mU3^6-2359 m3^4 mQ3^6 mU3^6+505 m3^2 mQ3^8 mU3^6+12 mQ3^10 mU3^6-840 m3^6 mQ3^2 msq^2 mU3^6-720 m3^2 mQ3^6 msq^2 mU3^6+120 mQ3^8 msq^2 mU3^6+204 m3^8 mU3^8-1259 m3^6 mQ3^2 mU3^8+283 m3^4 mQ3^4 mU3^8+571 m3^2 mQ3^6 mU3^8-183 mQ3^8 mU3^8+420 m3^4 mQ3^2 msq^2 mU3^8+600 m3^2 mQ3^4 msq^2 mU3^8-60 mQ3^6 msq^2 mU3^8-68 m3^6 mU3^10+587 m3^4 mQ3^2 mU3^10-594 m3^2 mQ3^4 mU3^10+171 mQ3^6 mU3^10-180 m3^2 mQ3^2 msq^2 mU3^10-60 mQ3^4 msq^2 mU3^10))/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2))+(64 (22 m3^7 mQ3^2-20 m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^4 msq^2+16 m3^7 mU3^2-112 m3^5 mQ3^2 mU3^2+100 m3^3 mQ3^4 mU3^2-6 m3 mQ3^6 mU3^2-60 m3 mQ3^4 msq^2 mU3^2-16 m3^5 mU3^4+87 m3^3 mQ3^2 mU3^4-74 m3 mQ3^4 mU3^4+30 m3 mQ3^2 msq^2 mU3^4) Xt)/(3 mQ3^2 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))-(8 (4 m3^10 mQ3^2-8 m3^8 mQ3^4+4 m3^6 mQ3^6-4 m3^10 mU3^2-695 m3^8 mQ3^2 mU3^2+886 m3^6 mQ3^4 mU3^2-301 m3^4 mQ3^6 mU3^2+18 m3^2 mQ3^8 mU3^2+360 m3^6 mQ3^2 msq^2 mU3^2-300 m3^4 mQ3^4 msq^2 mU3^2+180 m3^2 mQ3^6 msq^2 mU3^2+136 m3^8 mU3^4+326 m3^6 mQ3^2 mU3^4-40 m3^4 mQ3^4 mU3^4-212 m3^2 mQ3^6 mU3^4+6 mQ3^8 mU3^4-300 m3^4 mQ3^2 msq^2 mU3^4-240 m3^2 mQ3^4 msq^2 mU3^4+60 mQ3^6 msq^2 mU3^4-68 m3^6 mU3^6-5 m3^4 mQ3^2 mU3^6-308 m3^2 mQ3^4 mU3^6+237 mQ3^6 mU3^6+180 m3^2 mQ3^2 msq^2 mU3^6+60 mQ3^4 msq^2 mU3^6+18 m3^2 mQ3^2 mU3^8+6 mQ3^4 mU3^8) Xt^2)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((128 (14 m3^3-3 m3 mQ3^2-30 m3 msq^2-3 m3 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)))-(128 m3 (mQ3^2+3 mU3^2) (-11 m3^4+3 m3^2 mQ3^2+3 m3^2 mU3^2+5 mQ3^2 mU3^2))/(3 (m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)+(256 (7 m3^5 mQ3^2-8 m3^5 mU3^2+26 m3^3 mQ3^2 mU3^2+8 m3^3 mU3^4-17 m3 mQ3^2 mU3^4))/(3 mQ3^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) Xt^3+(4 (-4 m3^12 mQ3^4+8 m3^10 mQ3^6-4 m3^8 mQ3^8+176 m3^12 mQ3^2 mU3^2+667 m3^10 mQ3^4 mU3^2-1222 m3^8 mQ3^6 mU3^2+493 m3^6 mQ3^8 mU3^2-18 m3^4 mQ3^10 mU3^2-360 m3^8 mQ3^4 msq^2 mU3^2+300 m3^6 mQ3^6 msq^2 mU3^2-180 m3^4 mQ3^8 msq^2 mU3^2+4 m3^12 mU3^4-4793 m3^10 mQ3^2 mU3^4+10101 m3^8 mQ3^4 mU3^4-6913 m3^6 mQ3^6 mU3^4+1361 m3^4 mQ3^8 mU3^4-24 m3^2 mQ3^10 mU3^4-360 m3^8 mQ3^2 msq^2 mU3^4+600 m3^6 mQ3^4 msq^2 mU3^4+480 m3^4 mQ3^6 msq^2 mU3^4-240 m3^2 mQ3^8 msq^2 mU3^4-140 m3^10 mU3^6+7283 m3^8 mQ3^2 mU3^6-17667 m3^6 mQ3^4 mU3^6+13435 m3^4 mQ3^6 mU3^6-2665 m3^2 mQ3^8 mU3^6-6 mQ3^10 mU3^6+1020 m3^6 mQ3^2 msq^2 mU3^6-1380 m3^4 mQ3^4 msq^2 mU3^6+420 m3^2 mQ3^6 msq^2 mU3^6-60 mQ3^8 msq^2 mU3^6+204 m3^8 mU3^8-1881 m3^6 mQ3^2 mU3^8+6073 m3^4 mQ3^4 mU3^8-5659 m3^2 mQ3^6 mU3^8+1167 mQ3^8 mU3^8-360 m3^4 mQ3^2 msq^2 mU3^8-360 m3^2 mQ3^4 msq^2 mU3^8+240 mQ3^6 msq^2 mU3^8-68 m3^6 mU3^10-975 m3^4 mQ3^2 mU3^10+1508 m3^2 mQ3^4 mU3^10-417 mQ3^6 mU3^10+180 m3^2 mQ3^2 msq^2 mU3^10+60 mQ3^4 msq^2 mU3^10-18 m3^2 mQ3^2 mU3^12-6 mQ3^4 mU3^12) Xt^4)/(3 mQ3^2 (m3^2-mQ3^2)^2 mU3^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^3)-(64 (48 m3^7 mQ3^2-62 m3^5 mQ3^4+3 m3^3 mQ3^6+30 m3^3 mQ3^4 msq^2+16 m3^7 mU3^2+172 m3^5 mQ3^2 mU3^2-166 m3^3 mQ3^4 mU3^2+3 m3 mQ3^6 mU3^2-30 m3^3 mQ3^2 msq^2 mU3^2+30 m3 mQ3^4 msq^2 mU3^2-16 m3^5 mU3^4-185 m3^3 mQ3^2 mU3^4+184 m3 mQ3^4 mU3^4-30 m3 mQ3^2 msq^2 mU3^4+3 m3 mQ3^2 mU3^6) Xt^5)/(3 mQ3^2 (m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+((16 (87 m3^10-169 m3^8 mQ3^2+82 m3^6 mQ3^4+60 m3^8 msq^2-15 m3^6 mQ3^2 msq^2+15 m3^4 mQ3^4 msq^2+45 m3^6 msq^4-90 m3^4 mQ3^2 msq^4+45 m3^2 mQ3^4 msq^4-230 m3^8 mU3^2+445 m3^6 mQ3^2 mU3^2-215 m3^4 mQ3^4 mU3^2-255 m3^6 msq^2 mU3^2+195 m3^4 mQ3^2 msq^2 mU3^2-120 m3^2 mQ3^4 msq^2 mU3^2+15 m3^4 msq^4 mU3^2-30 m3^2 mQ3^2 msq^4 mU3^2+15 mQ3^4 msq^4 mU3^2+212 m3^6 mU3^4-409 m3^4 mQ3^2 mU3^4+197 m3^2 mQ3^4 mU3^4+120 m3^4 msq^2 mU3^4+75 m3^2 mQ3^2 msq^2 mU3^4-15 mQ3^4 msq^2 mU3^4-69 m3^4 mU3^6+133 m3^2 mQ3^2 mU3^6-64 mQ3^4 mU3^6-45 m3^2 msq^2 mU3^6-15 mQ3^2 msq^2 mU3^6))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3)+(-((640 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2)))+(192 m3)/(mQ3^2-mU3^2)-(96 m3 mU3^2)/((m3^2-mU3^2) (-mQ3^2+mU3^2))+(320 (-6 m3 msq^4+m3^3 mU3^2+12 m3 msq^2 mU3^2-m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt-(32 (87 m3^8-169 m3^6 mQ3^2+82 m3^4 mQ3^4+90 m3^6 msq^2-75 m3^4 mQ3^2 msq^2+45 m3^2 mQ3^4 msq^2-160 m3^6 mU3^2+310 m3^4 mQ3^2 mU3^2-150 m3^2 mQ3^4 mU3^2-75 m3^4 msq^2 mU3^2-60 m3^2 mQ3^2 msq^2 mU3^2+15 mQ3^4 msq^2 mU3^2+73 m3^4 mU3^4-141 m3^2 mQ3^2 mU3^4+68 mQ3^4 mU3^4+45 m3^2 msq^2 mU3^4+15 mQ3^2 msq^2 mU3^4) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+((1280 m3 msq^2)/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2))-(192 m3 (mQ3^2+3 mU3^2))/(mQ3^2-mU3^2)^3) Xt^3+((8 (mQ3^2+mU3^2) (111 m3^8-221 m3^6 mQ3^2+110 m3^4 mQ3^4+180 m3^6 msq^2-150 m3^4 mQ3^2 msq^2+90 m3^2 mQ3^4 msq^2-221 m3^6 mU3^2+440 m3^4 mQ3^2 mU3^2-219 m3^2 mQ3^4 mU3^2-150 m3^4 msq^2 mU3^2-120 m3^2 mQ3^2 msq^2 mU3^2+30 mQ3^4 msq^2 mU3^2+110 m3^4 mU3^4-219 m3^2 mQ3^2 mU3^4+109 mQ3^4 mU3^4+90 m3^2 msq^2 mU3^4+30 mQ3^2 msq^2 mU3^4))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(16 (60 m3^4 msq^2-90 m3^2 msq^4+2 m3^4 mU3^2+180 m3^2 msq^2 mU3^2-30 msq^4 mU3^2-3 m3^2 mU3^4+mU3^6))/(3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(24 (4 m3^8-3 mQ3^2 mU3^4 (mQ3^2+5 mU3^2)+m3^6 (3 mQ3^2+7 mU3^2)-m3^4 (6 mQ3^4+15 mQ3^2 mU3^2+29 mU3^4)+m3^2 (7 mQ3^4 mU3^2+31 mQ3^2 mU3^4+16 mU3^6)))/((m3^2-mQ3^2) (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) Xt^4+(-((192 m3 mU3^2)/((m3^2-mU3^2) (-mQ3^2+mU3^2)^3))-(640 m3 msq^2 (mQ3^2+mU3^2))/((m3^2-mQ3^2) (m3^2-mU3^2) (mQ3^2-mU3^2)^3)-(64 (-60 m3 msq^4+m3^3 mU3^2+120 m3 msq^2 mU3^2-m3 mU3^4))/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)) Xt^5) Log[msq^2/M^2]+(-24 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3))-24 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt+1/(mQ3^2-mU3^2) 48 (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^2+(48 ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^3)/(mQ3^2-mU3^2)-1/(mQ3^2-mU3^2)^3 24 (mQ3^2+mU3^2) (-((5 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(6 (-m3^2+mQ3^2)^3))-(5 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(6 (-m3^2+mU3^2)^3)) Xt^4-(24 (mQ3^2+mU3^2) ((40 m3 (mQ3^2-msq^2)^2)/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2))+(40 m3 (-msq^2+mU3^2)^2)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))) Xt^5)/(mQ3^2-mU3^2)^3) Log[msq^2/M^2]^2+(-((8 (128 m3^12-290 m3^10 mQ3^2+98 m3^8 mQ3^4+9 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-350 m3^10 mU3^2+762 m3^8 mQ3^2 mU3^2-251 m3^6 mQ3^4 mU3^2-151 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+420 m3^8 mU3^4-821 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4+41 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-217 m3^6 mU3^6+323 m3^4 mQ3^2 mU3^6-167 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+31 m3^4 mU3^8-34 m3^2 mQ3^2 mU3^8+23 mQ3^4 mU3^8))/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)))-(64 (20 m3^9-37 m3^7 mQ3^2+19 m3^5 mQ3^4-5 m3^3 mQ3^6+3 m3 mQ3^8-73 m3^7 mU3^2+135 m3^5 mQ3^2 mU3^2-23 m3^3 mQ3^4 mU3^2-7 m3 mQ3^6 mU3^2+24 m3^5 mU3^4-78 m3^3 mQ3^2 mU3^4+22 m3 mQ3^4 mU3^4) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2))+(16 (128 m3^12-294 m3^10 mQ3^2+98 m3^8 mQ3^4+9 m3^6 mQ3^6+20 m3^4 mQ3^8-9 m3^2 mQ3^10-858 m3^10 mU3^2+1414 m3^8 mQ3^2 mU3^2-251 m3^6 mQ3^4 mU3^2-151 m3^4 mQ3^6 mU3^2+41 m3^2 mQ3^8 mU3^2-3 mQ3^10 mU3^2+1304 m3^8 mU3^4-2049 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4+41 m3^2 mQ3^6 mU3^4-mQ3^8 mU3^4-525 m3^6 mU3^6+903 m3^4 mQ3^2 mU3^6-167 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6-37 m3^4 mU3^8-34 m3^2 mQ3^2 mU3^8+23 mQ3^4 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)-(64 (2 m3^7-37 m3^5 mQ3^2-2 m3^3 mQ3^4-6 m3 mQ3^6-33 m3^5 mU3^2+110 m3^3 mQ3^2 mU3^2+20 m3 mQ3^4 mU3^2-6 m3^3 mU3^4-75 m3 mQ3^2 mU3^4+27 m3 mU3^6) Xt^3)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^3)-(8 (124 m3^12 mQ3^2-284 m3^10 mQ3^4+90 m3^8 mQ3^6+11 m3^6 mQ3^8+20 m3^4 mQ3^10-9 m3^2 mQ3^12+1156 m3^12 mU3^2-3068 m3^10 mQ3^2 mU3^2+2702 m3^8 mQ3^4 mU3^2-474 m3^6 mQ3^6 mU3^2-201 m3^4 mQ3^8 mU3^2+32 m3^2 mQ3^10 mU3^2-3 mQ3^12 mU3^2-3048 m3^10 mU3^4+7406 m3^8 mQ3^2 mU3^4-5516 m3^6 mQ3^4 mU3^4+882 m3^4 mQ3^6 mU3^4+184 m3^2 mQ3^8 mU3^4-4 mQ3^10 mU3^4+2602 m3^8 mU3^6-6126 m3^6 mQ3^2 mU3^6+3984 m3^4 mQ3^4 mU3^6-502 m3^2 mQ3^6 mU3^6-54 mQ3^8 mU3^6-695 m3^6 mU3^8+1718 m3^4 mQ3^2 mU3^8-883 m3^2 mQ3^4 mU3^8+4 mQ3^6 mU3^8-3 m3^4 mU3^10-102 m3^2 mQ3^2 mU3^10+57 mQ3^4 mU3^10) Xt^4)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2)^4)-(64 (-18 m3^5 mQ3^2-m3^3 mQ3^4-3 m3 mQ3^6+14 m3^5 mU3^2+20 m3^3 mQ3^2 mU3^2+4 m3 mQ3^4 mU3^2-11 m3^3 mU3^4+m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt^5)/(3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) PolyLog[2,1-m3^2/mQ3^2]+(-12 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2)))-(640 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+1/(mQ3^2-mU3^2) 24 (2 m3^2-2 msq^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^2+(1280 (m3^2-msq^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^3)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))-1/(mQ3^2-mU3^2)^3 12 (2 m3^2-2 msq^2) (mQ3^2+mU3^2) ((40 mQ3^4)/(3 (-m3^2+mQ3^2)^3)-(10 mQ3^2)/(-m3^2+mQ3^2)^2-10/(3 (-m3^2+mQ3^2))-(40 mQ3^2 msq^2)/(3 (-m3^2+mQ3^2)^3)+(10 msq^2)/(-m3^2+mQ3^2)^2-(40 msq^2 mU3^2)/(3 (-m3^2+mU3^2)^3)+(40 mU3^4)/(3 (-m3^2+mU3^2)^3)+(10 msq^2)/(-m3^2+mU3^2)^2-(10 mU3^2)/(-m3^2+mU3^2)^2-10/(3 (-m3^2+mU3^2))) Xt^4-(640 (m3^2-msq^2) (mQ3^2+mU3^2) (m3^3 mQ3^2-2 m3^3 msq^2+m3 mQ3^2 msq^2+m3^3 mU3^2-2 m3 mQ3^2 mU3^2+m3 msq^2 mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)) PolyLog[2,1-m3^2/msq^2]+((40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2))/(-m3^2+mQ3^2)^3-(640 m3 (mQ3^2-msq^2)^2 Xt)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2))-(80 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) Xt^2)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2))+(1280 m3 (mQ3^2-msq^2)^2 Xt^3)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)+(40 (2 mQ3^2-2 msq^2) (-2 m3^4-3 m3^2 mQ3^2+mQ3^4+3 m3^2 msq^2+mQ3^2 msq^2) (mQ3^2+mU3^2) Xt^4)/((-m3^2+mQ3^2)^3 (mQ3^2-mU3^2)^3)-(640 m3 (mQ3^2-msq^2)^2 (mQ3^2+mU3^2) Xt^5)/((m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) PolyLog[2,1-msq^2/mQ3^2]+(-((32 m3^4 (m3^4-34 m3^2 mU3^2+17 mU3^4))/(3 (m3^2-mU3^2)^4))-(8 (-128 m3^12+346 m3^10 mQ3^2-280 m3^8 mQ3^4+13 m3^6 mQ3^6+37 m3^4 mQ3^8+294 m3^10 mU3^2-902 m3^8 mQ3^2 mU3^2+1025 m3^6 mQ3^4 mU3^2-391 m3^4 mQ3^6 mU3^2+34 m3^2 mQ3^8 mU3^2-98 m3^8 mU3^4+251 m3^6 mQ3^2 mU3^4-417 m3^4 mQ3^4 mU3^4+167 m3^2 mQ3^6 mU3^4-23 mQ3^8 mU3^4-9 m3^6 mU3^6+151 m3^4 mQ3^2 mU3^6-41 m3^2 mQ3^4 mU3^6+19 mQ3^6 mU3^6-20 m3^4 mU3^8-41 m3^2 mQ3^2 mU3^8+mQ3^4 mU3^8+9 m3^2 mU3^10+3 mQ3^2 mU3^10))/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2))+(-((64 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)))+(128 (m3^7-50 m3^5 mU3^2+17 m3^3 mU3^4))/(3 (m3^2-mU3^2)^3 (-mQ3^2+mU3^2))) Xt-(16 (128 m3^12-346 m3^10 mQ3^2+280 m3^8 mQ3^4-13 m3^6 mQ3^6-37 m3^4 mQ3^8-806 m3^10 mU3^2+2438 m3^8 mQ3^2 mU3^2-2561 m3^6 mQ3^4 mU3^2+903 m3^4 mQ3^6 mU3^2-34 m3^2 mQ3^8 mU3^2+98 m3^8 mU3^4-251 m3^6 mQ3^2 mU3^4+417 m3^4 mQ3^4 mU3^4-167 m3^2 mQ3^6 mU3^4+23 mQ3^8 mU3^4+9 m3^6 mU3^6-151 m3^4 mQ3^2 mU3^6+41 m3^2 mQ3^4 mU3^6-19 mQ3^6 mU3^6+20 m3^4 mU3^8+41 m3^2 mQ3^2 mU3^8-mQ3^4 mU3^8-9 m3^2 mU3^10-3 mQ3^2 mU3^10) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)+((64 m3 (-2 m3^2+mQ3^2+mU3^2) (m3^4-34 m3^2 mU3^2+17 mU3^4))/(3 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^3)+(128 (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^2)) Xt^3+(8 (124 m3^12 mQ3^2-332 m3^10 mQ3^4+262 m3^8 mQ3^6-3 m3^6 mQ3^8-39 m3^4 mQ3^10+1156 m3^12 mU3^2-4100 m3^10 mQ3^2 mU3^2+5326 m3^8 mQ3^4 mU3^2-2990 m3^6 mQ3^6 mU3^2+526 m3^4 mQ3^8 mU3^2+34 m3^2 mQ3^10 mU3^2-1968 m3^10 mU3^4+6466 m3^8 mQ3^2 mU3^4-7596 m3^6 mQ3^4 mU3^4+3712 m3^4 mQ3^6 mU3^4-543 m3^2 mQ3^8 mU3^4-11 mQ3^10 mU3^4+746 m3^8 mU3^6-2186 m3^6 mQ3^2 mU3^6+2210 m3^4 mQ3^4 mU3^6-774 m3^2 mQ3^6 mU3^6+4 mQ3^8 mU3^6-25 m3^6 mU3^8-29 m3^4 mQ3^2 mU3^8-20 m3^2 mQ3^4 mU3^8+14 mQ3^6 mU3^8+20 m3^4 mU3^10+32 m3^2 mQ3^2 mU3^10-4 mQ3^4 mU3^10-9 m3^2 mU3^12-3 mQ3^2 mU3^12) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)+((1024 m3 mU3^2 (2 m3^2-mQ3^2-mU3^2))/(3 (m3^2-mU3^2) (-mQ3^2+mU3^2)^4)-(64 (mQ3^2+mU3^2) (18 m3^5-37 m3^3 mQ3^2+22 m3 mQ3^4+m3^3 mU3^2-7 m3 mQ3^2 mU3^2+3 m3 mU3^4))/(3 (m3^2-mQ3^2)^2 (mQ3^2-mU3^2)^4)) Xt^5) PolyLog[2,1-m3^2/mU3^2]+(-((4 (-24 m3^10 mQ3^2+28 m3^8 mQ3^4-68 m3^6 mQ3^6+58 m3^4 mQ3^8-18 m3^2 mQ3^10+24 m3^10 mU3^2+220 m3^6 mQ3^4 mU3^2-188 m3^4 mQ3^6 mU3^2+70 m3^2 mQ3^8 mU3^2-6 mQ3^10 mU3^2-28 m3^8 mU3^4-220 m3^6 mQ3^2 mU3^4+16 m3^2 mQ3^6 mU3^4-8 mQ3^8 mU3^4+68 m3^6 mU3^6+188 m3^4 mQ3^2 mU3^6-16 m3^2 mQ3^4 mU3^6-58 m3^4 mU3^8-70 m3^2 mQ3^2 mU3^8+8 mQ3^4 mU3^8+18 m3^2 mU3^10+6 mQ3^2 mU3^10))/(3 (-m3^2+mQ3^2)^3 (m3^2-mU3^2)^3))+(32 (10 m3^5 mQ3^2-10 m3^3 mQ3^4+6 m3 mQ3^6-10 m3^5 mU3^2-8 m3 mQ3^4 mU3^2+10 m3^3 mU3^4+8 m3 mQ3^2 mU3^4-6 m3 mU3^6) Xt)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)+(16 (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^2)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3)-(128 (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^3)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2)-(8 (mQ3^2+mU3^2) (12 m3^10-14 m3^8 mQ3^2+34 m3^6 mQ3^4-29 m3^4 mQ3^6+9 m3^2 mQ3^8-14 m3^8 mU3^2-76 m3^6 mQ3^2 mU3^2+65 m3^4 mQ3^4 mU3^2-26 m3^2 mQ3^6 mU3^2+3 mQ3^8 mU3^2+34 m3^6 mU3^4+65 m3^4 mQ3^2 mU3^4-34 m3^2 mQ3^4 mU3^4+7 mQ3^6 mU3^4-29 m3^4 mU3^6-26 m3^2 mQ3^2 mU3^6+7 mQ3^4 mU3^6+9 m3^2 mU3^8+3 mQ3^2 mU3^8) Xt^4)/(3 (m3^2-mQ3^2)^3 (m3^2-mU3^2)^3 (mQ3^2-mU3^2)^2)+(64 (mQ3^2+mU3^2) (5 m3^5-5 m3^3 mQ3^2+3 m3 mQ3^4-5 m3^3 mU3^2-m3 mQ3^2 mU3^2+3 m3 mU3^4) Xt^5)/(3 (m3^2-mQ3^2)^2 (m3^2-mU3^2)^2 (mQ3^2-mU3^2)^2)) PolyLog[2,1-mQ3^2/mU3^2]+((40 (-2 msq^2+2 mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4))/(-m3^2+mU3^2)^3-(640 m3 (-msq^2+mU3^2)^2 Xt)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2))-(160 (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^2)/((mQ3^2-mU3^2) (-m3^2+mU3^2)^3)-(1280 m3 (-msq^2+mU3^2)^2 Xt^3)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^2)+(80 (mQ3^2+mU3^2) (-msq^2+mU3^2) (-2 m3^4+3 m3^2 msq^2-3 m3^2 mU3^2+msq^2 mU3^2+mU3^4) Xt^4)/((mQ3^2-mU3^2)^3 (-m3^2+mU3^2)^3)+(640 m3 (mQ3^2+mU3^2) (-msq^2+mU3^2)^2 Xt^5)/((m3^2-mU3^2)^2 (-mQ3^2+mU3^2)^4)) PolyLog[2,1-msq^2/mU3^2])))
);
